; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TText structure and handling library.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: TText is so called "buffer gap" method of editing text information.
;         it allows fast insert and delete operations.
;_________________________________________________________________________________________
module "Buffer gap engine"


gapDefaultSize = 1024  ; 256

; There are 3 coordinate systems in the TText structure:
;
;   1. "Offset"   - this is simple offset inside the memory, relative to the begining of the structure.
;
;   2. "Position" - this is the coordinate of the text characters in the buffer, excluding the gap (which can be placed anywhere in the text).
;
;   3. "Index"    - character position - the number of the character in the text buffer. In the case the buffer contains only ANSI characters index=pos;
;                   In the case when the buffer contains UTF-8 characters, the relation is not so direct and has to be computed for every character.
;                   This coordinate is not easy to be calculated, but it is rarely needed.
;
; The structure TText is backward structure. All members has negative offsets and the
; pointer to the structure points to its end, not to the start.
;
; The structure has fixed header and variable, dynamically changed length.
; All the text in the structure has positive offsets and all the header fields, negative.

rstruct TText
  .Length     dd  ?       ; The total length of the buffer in bytes
  .GapBegin   dd  ?       ; In offset units.
  .GapEnd     dd  ?       ; In offset units.
  .struc_size dd  ?       ; the size of the structure (it can be extended up).
  label .     byte        ; the text in UTF-8 encoding.
ends



proc TextIndexToPos, .pText, .index
begin
        pushad

        mov     esi, [.pText]
        xor     edx, edx
        mov     ebx, [.index]

.loop:
        dec     ebx
        js      .finish

        mov     eax, [esi+TText.GapBegin]
        cmp     edx, eax
        cmovae  eax, [esi+TText.GapEnd]
        sub     eax, [esi+TText.GapBegin]
        add     eax, edx

        cmp     eax, [esi+TText.Length]
        jae     .error

        movzx   eax, byte [esi+eax]

                                ; compute the length of the UTF-8 character by its first byte (version without branches) Is it fast?
        not     al              ; utf-8 starts with one of: 0xxxxxxx, 110xxxxx, 1110xxxx, 11110xxx, 111110xx, 1111110x
        bsr     eax, eax        ; eax =                 7,  5,  4,  3,  2,  1
        sub     eax, 7          ; eax =                 0, -2, -3, -4, -5, -6
                                ; CF  =                 0,  1,  1,  1,  1,  1
        not     eax             ; eax =                -1, +1, +2, +3, +4, +5
        sbb     eax, -2         ; eax = edx + 2 - CF = +1, +2, +3, +4, +5, +6

        add     edx, eax
        jmp     .loop

.finish:
        mov     [esp+4*regEAX], edx
        clc
        popad
        return

.error:
        mov     eax, [esi+TText.Length]
        sub     eax, [esi+TText.GapEnd]
        add     eax, [esi+TText.GapBegin]
        mov     [esp+4*regEAX], eax
        stc
        popad
        return
endp






; Converts from text position (in bytes) to buffer position (in bytes).
; returns CF=1 if the position is outside the text.
; the returned position is relative to the text start.
;
; esi == .buffer; eax = .pos
;
; If eax is above the size of the text, the position of the buffer end is returned.
; All positions are unsigned, so passing eax = $ffffffff will return the end of the
; data in the buffer.
;
; See also the pos2ofs macro. Because of the simple operation it is better to use it
; or even explicit processing.

proc TextPosToOffset, .pText, .pos
begin
        push    ebx
        mov     eax, [.pos]
        mov     ebx, [.pText]

        cmp     eax, [ebx+TText.GapBegin]
        jb      @f
        add     eax, [ebx+TText.GapEnd]
        sub     eax, [ebx+TText.GapBegin]
@@:
        cmp     eax, [ebx+TText.Length]
        cmova   eax, [ebx+TText.Length]
        pop     ebx
        return
endp


macro pos2ofs  pos, text {
local ..skip

  if pos eqtype eax & text eqtype eax
    cmp    pos, [text+TText.GapBegin]
    jb     ..skip
    add    pos, [text+TText.GapEnd]
    sub    pos, [text+TText.GapBegin]
..skip:
    cmp    pos, [text+TText.Length]    ; ?????????
    cmova  pos, [text+TText.Length]    ; ?????????
  else
    error "Error! Arguments must be registers!"
    err
  end if
}


; Converts from buffer offset to the text position.
; Arguments:
;   .pText: points to the TText strucure, eax is the offset
;   .offs: offset in bytes.
;
; if the buffer offset is placed inside the gap, the start of the gap is returned.
;
; See also ofs2pos macro. Very often it is better to use the macro or explicit processing,
; instead of procedure call, because the operation is very simple.

proc TextOffsetToPos, .pText, .offs
begin
        push    ebx
        mov     eax, [.offs]
        mov     ebx, [.pText]

        cmp     eax, [ebx+TText.Length]
        cmova   eax, [ebx+TText.Length]

        cmp     eax, [ebx+TText.GapBegin]
        jb      .finish

        add     eax, [ebx+TText.GapEnd]
        sub     eax, [ebx+TText.GapBegin]

        cmp     eax, [ebx+TText.GapBegin]
        cmovb   eax, [ebx+TText.GapBegin]

.finish:
        pop     ebx
        return
endp


macro ofs2pos ofs, text {
local ..skip

  if ofs eqtype eax & text eqtype eax
        cmp     ofs, [text+TText.Length]
        cmova   ofs, [text+TText.Length]

        cmp     ofs, [text+TText.GapBegin]
        jb      ..skip

        add     ofs, [text+TText.GapBegin]
        sub     ofs, [text+TText.GapEnd]

        cmp     ofs, [text+TText.GapBegin]
        cmovb   ofs, [text+TText.GapBegin]
..skip:
  else
    error "Error! Arguments must be registers!"
    err
  end if
}


; Moves the gap to the desired position.
; Arguments:
;   .pText        - pointer to the TText
;   .position     - position of the gap in position units.
; Returns:
;   Nothing
;
; This operation can't fail. Preserves all registers.

proc TextMoveGap, .pText, .position
begin
        pushad

;        int3

        mov     edx, [.pText]
        mov     ecx, [.position]

        mov     eax, [edx+TText.Length]
        mov     edi, [edx+TText.GapBegin]
        mov     esi, [edx+TText.GapEnd]

        cmp     esi, edi
        je      .finish         ; the gap length is 0 nothing to move

        sub     eax, esi
        add     eax, edi

        cmp     ecx, eax
        cmova   ecx, eax        ; the maximal possible position for the gap.

        sub     ecx, edi        ; [edx+TText.GapBegin]  ; ecx is the distance to move the gap.
        jz      .finish                                 ; source and destination are the same - no move.

        add     edi, edx
        add     esi, edx

        add     [edx+TText.GapBegin], ecx
        add     [edx+TText.GapEnd], ecx

        test    ecx, ecx
        js      .move_back

        push    ecx
        shr     ecx, 2
        rep movsd
        pop     ecx
        and     ecx, 3
        rep movsb

.finish:

; fill the gap with placeholder char in order to make debug easy! It lowers the performance a lot!!!
if defined options.DebugMode & options.DebugMode
        call    __fill_the_gap
end if

        popad
        return

.move_back:
        std

        neg     ecx
        xchg    esi, edi

        dec     esi
        dec     edi

        push    ecx
        and     ecx, 3
        rep movsb

        sub     esi, 3
        sub     edi, 3

        pop     ecx
        shr     ecx, 2
        rep movsd

        cld
        jmp     .finish
endp


proc __fill_the_gap
begin
        push    edi ecx eax

        mov     edi, [edx+TText.GapBegin]
        mov     ecx, [edx+TText.GapEnd]
        sub     ecx, edi
        add     edi, edx
        mov     al, '*'
        rep stosb

        pop     eax ecx edi
        return
endp



proc TextCreate, .struc_size
begin
        push    ecx

        mov     ecx, [.struc_size]
        mov     eax, ecx
        add     eax, gapDefaultSize

        stdcall GetMem, eax
        jc      .finish

        lea     eax, [eax+ecx]
        mov     [eax+TText.struc_size], ecx
        xor     ecx, ecx
        mov     [eax+TText.GapBegin], ecx
        mov     [eax+TText.GapEnd], gapDefaultSize
        mov     [eax+TText.Length], gapDefaultSize

.finish:
        pop     ecx
        return
endp


;---------------------------------------------------------------------------------
; Duplicates TText structure of any size. Returns pointer to the structure in EAX
; The returned pointer have to be freed with TextFree when not needed anymore.
;
; Arguments:
;   .pText - pointer to the source structure
;
; Returns:
;   CF = 0 if everithing is OK. In this case EAX is a pointer to the new TText structure.
;   CF = 1 if there is an error. The TText is not duplicated and EAX contains the value
;          before the call.
;---------------------------------------------------------------------------------

proc TextDup, .pText
begin
        pushad

        mov     esi, [.pText]
        test    esi, esi
        stc
        jz      .finish

        mov     ecx, [esi+TText.Length]
        add     ecx, [esi+TText.struc_size]

        stdcall GetMem, ecx
        jc      .finish
        mov     edi, eax

        mov     edx, [esi+TText.struc_size]

        add     eax, edx
        mov     [esp+4*regEAX], eax

        sub     esi, edx

        mov     ebx, ecx
        shr     ecx, 2
        and     ebx, 3

        rep movsd

        mov     ecx, ebx
        rep movsb
        clc

.finish:
        popad
        return
endp



proc TextFree, .pText
begin
        push    eax
        mov     eax, [.pText]
        test    eax, eax
        jz      .exit

        sub     eax, [eax+TText.struc_size]
        stdcall FreeMem, eax

.exit:
        pop     eax
        return
endp



; Ensures that the gap size is at least [.desired_size] bytes long.
; If needed, reallocates the TText
;
; In order to provide less frequent reallocations, the procedure
; always allocates bigger gap than requested.
; max([.desired_size]*2, gapDefaultSize) memory is allocated,
; if the buffer is to be reallocated.
;
; Arguments:
;   .pText        - pointer to TText
;   .desired_size - desired size of the gap.
;
; Returns
;   edx - pointer to the same TText. Can be changed because of memory reallocation.
;   CF = 0 if OK
;   CF = 1 if the memory reallocation failed.
;

proc TextSetGapSize, .pText, .desired_size
begin
        pushad

        mov     edx, [.pText]
        mov     ecx, [.desired_size]
        mov     eax, [edx+TText.GapEnd]
        sub     eax, [edx+TText.GapBegin]

        cmp     eax, ecx
        jae     .finish         ; CF==0

        lea     ecx, [2*ecx]
        mov     eax, gapDefaultSize
        cmp     ecx, eax
        cmovl   ecx, eax                          ; ecx is the new gap size here.

        mov     ebx, [edx+TText.Length]           ; current data size. text+gap
        add     ebx, [edx+TText.GapBegin]
        sub     ebx, [edx+TText.GapEnd]
        add     ebx, ecx
        add     ebx, [edx+TText.struc_size]       ; the new buffer size. it will always be higher than the current.

        push    [edx+TText.struc_size]
        sub     edx, [edx+TText.struc_size]
        stdcall ResizeMem, edx, ebx
        pop     edx
        lea     edx, [edx+eax]
        jc      .finish                           ; well, it is just an error, but how to proceed?

        sub     ebx, [edx+TText.struc_size]       ; total size of the buffer

        lea     edi, [edx+ebx]
        xchg    ebx,[edx+TText.Length]            ; ebx is now the old buffer length.

        lea     esi, [edx+ebx]

; the space is allocated and there are 2 gaps. So, merge them now. The new gap is always at the end.
; the size of the new gap is

        mov     ecx, ebx
        sub     ecx, [edx+TText.GapEnd]

        std

        dec     esi
        dec     edi

        push    ecx
        and     ecx, 3

        rep movsb

        sub     esi, 3
        sub     edi, 3

        pop     ecx
        shr     ecx, 2

        rep movsd

        cld

        sub     edi, edx
        add     edi, 4
        mov     [edx+TText.GapEnd], edi
        clc

.finish:
        mov     [esp+4*regEDX], edx

; fill the gap with placeholder char in order to make debug easy! It lowers the performance a lot!!!
if defined options.DebugMode & options.DebugMode
        call    __fill_the_gap
end if
        popad
        return
endp





; Inserts on the current gap position up to 4 bytes of character data.
; They can be 4 ANSI characters, or one 4 byte encoded UTF-8 or
; two 2-byte encoded UTF-8, one ASCII and one 2 byte UTF-8 and so on.
;
; .pText - pointer to the TText,
; .char -  from 0 to 4 bytes character data.

proc TextAddChar, .pText, .char
begin
        push    eax ecx

        mov     edx, [.pText]

        stdcall TextSetGapSize, edx, 4
        jc      .finish

        mov     eax, [edx+TText.GapBegin]

        pushd   [.char]
        popd    [edx+eax]

        mov     ecx, 4

.len_loop:
        cmp     byte [edx+eax], 0
        je      .len_ok
        inc     eax
        loop    .len_loop

.len_ok:
        mov     [edx+TText.GapBegin], eax
        clc

.finish:
        pop     ecx eax
        return
endp




; Deletes one UTF-8 character at the current gap position.
; esi - pointer to TText

;proc TextDelChar
;begin
;        push    eax
;
;        mov     eax, [esi+TText.GapStart]
;        add     eax, [esi+TText.GapLen]
;
;        stdcall DecodeUtf8, [esi+eax]
;        add     [esi+TText.GapLen], edx
;
;.len_ok:
;        pop     eax
;        return
;endp




; Inserts a string in the TText structure
; which can be pointer to NULL terminated string or
; string handle.
;
; Arguments:
;   .pText - pointer to the TText structure
;   .position  - position where the string to be inserted.
;   .hString - pointer or handle of string to be inserted.
; Returns:
;   edx: pointer to TText structure. Can be changed because of reallocations.
;   eax: length of the string added.

proc TextAddString, .pText, .position, .hString
begin
        pushad

        mov     edx, [.pText]

        stdcall StrLen, [.hString]
        mov     ecx, eax
        mov     ebx, eax

        stdcall StrPtr, [.hString]
        mov     esi, eax

        stdcall TextSetGapSize, edx, ecx
        jc      .finish

        stdcall TextMoveGap, edx, [.position]

        mov     edi, [edx+TText.GapBegin]
        add     edi, edx

        shr     ecx, 2
        rep movsd
        mov     ecx, ebx
        and     ecx, 3
        rep movsb

        add     [edx+TText.GapBegin], ebx

.finish:
        mov     [esp+4*regEAX], ebx
        mov     [esp+4*regEDX], edx
        popad
        return
endp





proc TextAddStr2, .pText, .position, .hString, .maxLength
begin
        pushad
        mov     edx, [.pText]

        mov     eax, [.hString]
        cmp     eax, $c0000000
        jb      .proceed

        stdcall StrLen, [.hString]
        mov     [.maxLength], eax

.proceed:
        stdcall StrPtr, [.hString]
        mov     esi, eax

        add     [.maxLength], 4  ; some reserve
        stdcall TextSetGapSize, edx, [.maxLength]
        jc      .finish

        stdcall TextMoveGap, edx, [.position]

        mov     edi, [edx+TText.GapBegin]
        add     edi, edx

.copy_loop:
        lodsd
        stosd

        mov     ebx, eax
        not     eax
        sub     ebx, $01010101
        and     eax, ebx
        and     eax, $80808080
        jz      .copy_loop

        xor     eax, eax
        sub     edi, 5

.final_loop:
        inc     edi
        cmp     [edi], al
        jne     .final_loop

        sub     edi, edx
        sub     edi, [edx+TText.GapBegin]
        add     [edx+TText.GapBegin], edi

        clc
        mov     [esp+4*regEAX], edi

.finish:
        mov     [esp+4*regEDX], edx
        popad
        return
endp



proc TextAddBytes, .pText, .position, .pData, .DataLen
begin
        pushad
        mov     edx, [.pText]
        mov     esi, [.pData]

        stdcall TextSetGapSize, edx, [.DataLen]
        jc      .finish

        stdcall TextMoveGap, edx, [.position]

        mov     edi, [edx+TText.GapBegin]
        add     edi, edx
        mov     ebx, [.DataLen]

        mov     ecx, ebx
        shr     ecx, 2
        rep movsd

        mov     ecx, ebx
        and     ecx, 3
        rep movsb

        add     [edx+TText.GapBegin], ebx

.finish:
        mov     [esp+4*regEDX], edx
        popad
        return
endp




proc TextAddText, .pText, .position, .pTextSrc
begin
        pushad

        mov     edx, [.pText]
        mov     ebx, [.pTextSrc]

        mov     ecx, [ebx+TText.Length]
        sub     ecx, [ebx+TText.GapEnd]
        add     ecx, [ebx+TText.GapBegin]       ; total length
        mov     [esp+4*regEAX], ecx

        stdcall TextSetGapSize, edx, ecx
        stdcall TextMoveGap, edx, [.position]

        mov     edi, edx
        mov     esi, ebx

        add     edi, [edx+TText.GapBegin]
        mov     eax, [ebx+TText.GapBegin]

        mov     ecx, eax
        and     eax, 3
        shr     ecx, 2

        rep movsd
        mov     ecx, eax
        rep movsb

        mov     eax, [ebx+TText.Length]
        mov     esi, ebx

        sub     eax, [ebx+TText.GapEnd]
        add     esi, [ebx+TText.GapEnd]

        mov     ecx, eax
        and     eax, 3
        shr     ecx, 2

        rep movsd
        mov     ecx, eax
        rep movsb

        sub     edi, edx
        mov     [edx+TText.GapBegin], edi

        mov     [esp+4*regEDX], edx
        popad
        return
endp



; Moves the gap to the end of the text and if possible inserts byte 0 after it, so
; the buffer can be used as an ASCIIZ string.
; returns the length of the string in eax and the pointer to the TText object in edx (it can be changed!)

proc TextCompact, .pText
begin
        mov     edx, [.pText]
        stdcall TextMoveGap, edx, -1
        stdcall TextSetGapSize, edx, 4

        mov     eax, [edx+TText.GapBegin]   ; the length of the text.
        and     dword [edx+eax], 0
        return
endp



endmodule