module "Hash tree"

; The hash tree is a TArray of THashTreeNode elements.
; .bit0 and .bit1 elements contain the offset to the next element of the array
; for the hash bit respectively 0 and 1.
;
; The procedure of searching in the tree is by separating the every next bit of the
; hash and jumping to the offset written in the [.bit0] or [.bit1] of the current node.
;
; if on some loop, the [.bitX] is 0 then, the string is not inside the tree.
;
; The offset on the 32th bit points to the THashTreeLeaf element (it takes 2 elements of
; the array). Here, direct comparison with [THashTreeLeaf.hString] element has to be done
;
; If the strings match, then the element has been found. If not, then a hash collision occured.
; In this case, the strings with equal hashes are added in a linear list linked through the
; [THashTreeLeaf.next] field. Follow the links and compare the strings until 0 (string not found)
; or string match.
;
; The searching in the tree is slower than the classic hash table, but is still O(1)
; and has the great advantage, that allows using of 32 bit hashes where the collisions
; are very rare, and still keeps the hash tree pretty small. Something like O(n.log n)
;
; This structure worth to be used when very big amount of strings need to be searched.


struct THashTreeNode
  .bit0 dd ?
  .bit1 dd ?
ends

struct THashTreeLeaf
  .hString dd ?         ; a handle of the string stored.
  .next    dd ?         ; the next element in the cases of collisions. (well, it is very case, but still happens sometimes).
  .lparam  dd ?         ; user defined parameter.
  .wparam  dd ?         ; user defined parameter.
ends


; Searches the hash tree for the given string.
; if [.fAdd] is TRUE and if the string is not in the tree it will be added.
;
; Arguments:
;   .pHashTree - pointer to TArray with elements size sizeof.THashTreeNode (8 bytes)
;   .hString - handle/pointer to the string to be searched/added. Notice, that exactly this
;              string will be added to the THashTreeLeaf element. The caller is responsible
;              to not free this string until the hash tree is needed.
;   .fAdd - flag indicating whether to add the string to the tree if is not included.
;
; Returns:
;   edx - pointer to the array. Can be reallocated!
;   eax - index in the TArray of the THashTreeLeaf of the string.
;   CF = 1: the string is already in the tree.
;   CF = 0: the string is not in the tree.
;     [.fAdd] <> 0: eax contains a pointer to the new added leaf.
;                   eax == 0 means there is an out of memory error on
;                   the attempt to add the string.
;     [.fAdd] = 0: eax == 0

proc SearchHashTree, .pHashTree, .hString, .fAdd
.hash dd ?
begin
        pushad

        mov     edx, [.pHashTree]

        stdcall StrHash, [.hString]
        mov     [.hash], eax

        mov     ecx, 32                          ; the depth of the tree
        mov     ebx, TArray.array

        cmp     [edx+TArray.count], 0
        jne     .treeseek

        stdcall AddArrayItems, edx, 1
        and     dword [eax], 0
        and     dword [eax], 0

.treeseek:
        xor     edi, edi
        ror     [.hash], 1
        adc     edi, edi
        lea     edi, [ebx+4*edi]

        cmp     dword [edx + edi], 0             ; edi is offset from the begining of the array.
        je      .notfound

        mov     ebx, [edx + edi]
        loop    .treeseek

; The hash exists in the hash tree. Process the possible collisions.
; here ebx is the offset to the THashTreeLeaf !

.cmp_loop:
        stdcall StrCompCase, [.hString], [edx+ebx+THashTreeLeaf.hString]
        jc      .finish      ; the string is already in the hash tree!!!

;        OutputValue "Collision at offset:", ebx, 10, -1

        lea     edi, [ebx + THashTreeLeaf.next]

        cmp     dword [edx + edi], 0
        je      .add_in_list

        mov     ebx, [edx + edi]
        jmp     .cmp_loop

.add_in_list:
        cmp     [.fAdd], 0
        je      .finish_zero

        stdcall AddArrayItems, edx, 2
        jc      .finish_zero

        sub     eax, edx
        jmp     .do_add


.notfound:
        cmp     [.fAdd], 0
        je      .finish_zero

; edx - pointer to the tree array,  ebx - offset of the last found element.
; eax - 0 or 1 depending of the last bit checked. ecx - remaining bits count.
; edi - the offset to the [.bitX] field of the last node.

.add_remaining:

        lea     eax, [ecx+1]
        stdcall AddArrayItems, edx, eax  ; add all needed (THashTreeNode) + THashTreeLeaf
        jc      .finish_zero

        sub     eax, edx

        dec     ecx
        jz      .do_add

.addloop:
        mov     [edx + edi], eax
        and     dword [edx + eax + THashTreeNode.bit0], 0
        and     dword [edx + eax + THashTreeNode.bit1], 0

        xor     edi, edi
        ror     [.hash], 1
        adc     edi, edi

        lea     edi, [eax + 4*edi]
        add     eax, sizeof.THashTreeNode
        loop    .addloop

.do_add:
        mov     [edx + edi], eax
        lea     ebx, [edx+eax]

        mov     ecx, [.hString]
        xor     eax, eax

        mov     [ebx+THashTreeLeaf.hString], ecx
        mov     [ebx+THashTreeLeaf.next], eax
        mov     [ebx+THashTreeLeaf.lparam], eax
        mov     [ebx+THashTreeLeaf.wparam], eax

        clc

.finish:
        mov     [esp+4*regEAX], ebx
        mov     [esp+4*regEDX], edx
        popad
        return


.finish_zero:
        xor     ebx, ebx
        clc
        jmp     .finish
endp




proc FreeHashTree, .pHashTree
begin
        pushad

        mov     edx, [.pHashTree]
        cmp     [edx+TArray.count], 32
        jb      .free_array

        pushd   0 0

        mov     ebx, TArray.array
        mov     ecx, 32

.scan:
        dec     ecx
        js      .free_leaf

        mov     esi, [edx+ebx+THashTreeNode.bit0]
        mov     ebx, [edx+ebx+THashTreeNode.bit1]

        test    ebx, ebx
        jz      .bit0

        test    esi, esi
        jz      .scan

        push    esi ecx
        jmp     .scan

.bit0:
        mov     ebx, esi
        jmp     .scan

.free_leaf:

.free_loop:

        mov     eax, [edx+ebx+THashTreeLeaf.hString]
        cmp     eax, $c0000000
        jb      .free_ok
        stdcall StrDel, eax
.free_ok:
        mov     ebx, [edx+ebx+THashTreeLeaf.next]
        test    ebx, ebx
        jnz     .free_loop

        pop     ecx ebx
        test    ebx, ebx
        jnz     .scan

.free_array:
        stdcall FreeMem, edx

        popad
        return
endp










endmodule