; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Memory based images manipulation library.
;
;  Target OS: Linux
;
;  Dependencies: memory.asm
;
;  Notes:
;_________________________________________________________________________________________

uses libX11, xshm



struct TLinuxImage
  . TImage

; os dependent data
  .ximage  dd ?
  .shminfo XShmSegmentInfo
ends



body CreateImage
begin
        pushad

        stdcall GetMem, sizeof.TLinuxImage
        jc      .finish
        mov     esi, eax

        xor     eax, eax
        xor     ebx, ebx
        inc     eax

        mov     ecx, [.width]
        mov     edx, [.height]

        cmp     ecx, ebx
        cmovle  ecx, eax

        cmp     edx, ebx
        cmovle  edx, eax

        mov     [esi+TLinuxImage.width], ecx
        mov     [esi+TLinuxImage.height], edx

        mov     [esi+TLinuxImage.orgX], ebx
        mov     [esi+TLinuxImage.orgY], ebx
        mov     [esi+TLinuxImage.wrapW], ecx
        mov     [esi+TLinuxImage.wrapH], edx

        lea     eax, [4*ecx]
        imul    eax, edx

        cmp     [flagSharedMemory], 0
        je      .fall_back

        cinvoke shmget, IPC_PRIVATE, eax, IPC_CREAT or 777o
        test    eax, eax
        js      .error

        mov     [esi+TLinuxImage.shminfo.ShmID], eax

        cinvoke shmat, eax, 0, 0
        cmp     eax, -1
        je      .error_free

        mov     [esi+TLinuxImage.shminfo.Addr], eax
        mov     [esi+TLinuxImage.pPixels], eax
        mov     [esi+TLinuxImage.shminfo.fReadOnly], 0

        lea     ebx, [esi+TLinuxImage.shminfo]
        cinvoke XShmCreateImage, [hApplicationDisplay], 0, $20, ZPixmap, eax, ebx, [esi+TLinuxImage.width], [esi+TLinuxImage.height]
        mov     [esi+TLinuxImage.ximage], eax


;     XShmAttach will return the shared segment ID that later will be returned in the
; XShmCompletionEvent.shmseg field on completion events.
;
;     This value is a server side ID and can not be easy associated with the client side image
; in order to provide some locking mechanism.
;
;     The only way I know is to collect these IDs in some array (hash table, etc.) and later to
; search in it in order to unlock the image.

        cinvoke XShmAttach, [hApplicationDisplay], ebx

;        OutputValue "Create image SegID2: $", [esi+TLinuxImage.shminfo.SegID], 16, 8

.exit_ok:
        mov     [esp+4*regEAX], esi

        lea     esi, [esi+TLinuxImage.lock]
        stdcall MutexCreate, 0, esi
        stdcall MutexRelease, esi

        clc

.finish:
        popad
        return

.error_free:
;        OutputValue "Error create image ShmID: $", [esi+TLinuxImage.shminfo.ShmID], 16, 8
        cinvoke shmctl, [esi+TLinuxImage.shminfo.ShmID], IPC_RMID, 0

.error:
        stdcall FreeMem, esi
        stc
        jmp     .finish


.fall_back:
        stdcall GetMem, eax
        mov     [esi+TLinuxImage.pPixels], eax

        cinvoke XCreateImage, [hApplicationDisplay], 0, $20, ZPixmap, 0, eax, [esi+TLinuxImage.width], [esi+TLinuxImage.height], 32, 0
        test    eax, eax
        jz      .error

        mov     [esi+TLinuxImage.ximage], eax
        jmp     .exit_ok
endp




body DestroyImage
begin
        pushad

        mov     esi, [.ptrImage]
        test    esi, esi
        jz      .finish

.lock_wait:
;        stdcall LockImg, esi
;        jc      .lock_wait

        cmp     [flagSharedMemory], 0
        je      .fall_back

;        OutputValue "Destroy image SegID: $", [esi+TLinuxImage.shminfo.SegID], 16, 8

        lea     eax, [esi+TLinuxImage.shminfo]
        cinvoke XShmDetach, [hApplicationDisplay], eax

        cinvoke XSync, [hApplicationDisplay], FALSE

        cinvoke XDestroyImage, [esi+TLinuxImage.ximage]

        cinvoke shmdt, [esi+TLinuxImage.shminfo.Addr]
        cinvoke shmctl, [esi+TLinuxImage.shminfo.ShmID], IPC_RMID, 0

.exit_ok:
        lea     edi, [esi+TLinuxImage.lock]
        stdcall MutexDestroy, edi
        stdcall FreeMem, esi

.finish:
        popad
        return


.fall_back:
        mov     eax, [esi+TLinuxImage.ximage]
        stdcall FreeMem, [eax+XImage.pData]
        mov     [eax+XImage.pData], 0

        cinvoke XDestroyImage, eax
        jmp     .exit_ok

endp




if used ___CheckCompletionEvent

___CheckCompletionEvent:

virtual at esp+4
  .display dd ?
  .pEvent  dd ?
  .user    dd ?
end virtual

; timeout
        mov     eax, [.pEvent]
        mov     eax, [eax+XEvent.type]
        cmp     eax, [ShmCompletionEvent]
        sete    al
        movzx   eax, al
        retn
end if


body DrawImageRect

.maskUL RECT ; upper left
.maskUR RECT ; upper right
.maskLL RECT ; lower left
.maskLR RECT ; lower right

.sRect  RECT ; the rectangle used for intersecting the source rectangle.

.event XEvent
       rb 256

.gc         dd ?
.draw_proc  dd ?
.fDraw   rd 1

begin
        pushad

        mov     esi, [.pImage]
        test    esi, esi
        jz      .finish

        stdcall LockImg, esi
        jc      .finish

        xor     eax, eax

        cmp     [flagSharedMemory], eax
        cmovz   edx, [XPutImage]
        cmovnz  edx, [XShmPutImage]
        mov     [.draw_proc], edx

        mov     [.fDraw], eax

        cmp     [esi+TLinuxImage.ximage], eax
        je      .exit

        mov     ecx, [esi+TImage.wrapW]
        mov     edx, [esi+TImage.wrapH]

        mov     [.maskUL.left], eax     ; UL
        mov     [.maskUL.top], eax      ; UL
        mov     [.maskLL.left], eax             ; LL
        mov     [.maskUR.top], eax                      ; UR

        mov     [.maskUL.right], ecx    ; UL
        mov     [.maskUL.bottom], edx   ; UL
        mov     [.maskLR.left], ecx                             ; LR
        mov     [.maskLR.top], edx                              ; LR
        mov     [.maskLL.top], edx              ; LL
        mov     [.maskLL.right], ecx            ; LL
        mov     [.maskUR.left], ecx                     ; UR
        mov     [.maskUR.bottom], edx                   ; UR

        shl     ecx, 1
        shl     edx, 1

        mov     [.maskUR.right], ecx                    ; UR
        mov     [.maskLL.bottom], edx                   ; LL
        mov     [.maskLR.right], ecx                            ; LR
        mov     [.maskLR.bottom], edx                           ; LR

        mov     ecx, [.xSrc]
        mov     edx, [.ySrc]
        mov     [.sRect.left], ecx
        mov     [.sRect.top], edx
        add     ecx, [.width]
        add     edx, [.height]
        mov     [.sRect.right], ecx
        mov     [.sRect.bottom], edx

        lea     eax, [.sRect]
        lea     edx, [.maskUL]
        stdcall RectIntersect2, eax, edx
        jc      .exit

; at least some part of the image, has to be drawn.

        cinvoke XCreateGC, [hApplicationDisplay], [.where], 0, 0
        mov     [.gc], eax

        mov     ebx, [esi+TImage.orgX]
        mov     edx, [esi+TImage.orgY]

        add     [.sRect.left], ebx
        add     [.sRect.top], edx
        add     [.sRect.right], ebx
        add     [.sRect.bottom], edx

        mov     ebx, [esi+TImage.wrapW]
        mov     edx, [esi+TImage.wrapH]

        lea     edi, [.maskUL]
        lea     ecx, [.sRect]

        mov     eax, 3

.loop_rect:

        stdcall RectIntersect2, edi, ecx
        cmc
        rcl     [.fDraw], 1

        test    eax, 1  ; horizontal
        jnz     .h_ok

        sub     [edi+RECT.left], ebx
        sub     [edi+RECT.right], ebx

.h_ok:
        test    eax, 2 ; vertical
        jnz     .v_ok

        sub     [edi+RECT.top], edx
        sub     [edi+RECT.bottom], edx

.v_ok:
        stdcall RectToBounds, edi, edi

        add     edi, sizeof.RECT
        dec     eax
        jns     .loop_rect

; Now draw the rectangles.


        lea     ebx, [.maskLR]
        mov     edi, 3

.draw_loop:
        shr     [.fDraw], 1
        jnc     .next

        setz    cl              ; is it the last rectangle to draw?
        movzx   ecx, cl

; arguments from here!

        push    ecx     ; events flag

        mov     ecx, [.yDst]
        mov     eax, [.xDst]

        test    edi, 2
        jz      .y_ok
        add     ecx, [.maskUR.bottom]
.y_ok:
        test    edi, 1
        jz      .x_ok
        add     eax, [.maskUL.right]
.x_ok:

        push    [ebx+RECT.bottom]               ; .height
        push    [ebx+RECT.right]                ; .width

        push    ecx                             ; destination Y
        push    eax                             ; destination X

        push    [ebx+RECT.top]                  ; .ySrc
        push    [ebx+RECT.left]                 ; .xSrc
        push    [esi+TLinuxImage.ximage]        ; .ximage
        push    [.gc]                           ; .context
        push    [.where]                        ; .drawable
        push    [hApplicationDisplay]           ; .display

        call    [.draw_proc]     ; XShmPutImage or XPutImage depending on SHM available.
        add     esp, 11*4

.next:
        cmp     [.fDraw], 0
        je      .end_draw

        sub     ebx, sizeof.RECT
        dec     edi
        jns     .draw_loop

.end_draw:

        cmp     [flagSharedMemory], 0
        je      .events_ok

; now, wait for the events:
        lea     eax, [.event]
        cinvoke XIfEvent, [hApplicationDisplay], eax, ___CheckCompletionEvent, eax

.events_ok:

        cinvoke XFreeGC, [hApplicationDisplay], [.gc]

.exit:
        stdcall UnlockImg, esi

.finish:
        popad
        return

endp
