; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Memory based images manipulation library.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: Include respective OS dependent files.
;_________________________________________________________________________________________
module "Images library"

struct TImage
  .width     dd ?  ; width in pixels.
  .height    dd ?  ; height in pixels.
  .pPixels   dd ?  ; pointer to the pixel memory.

  .orgX      dd ?
  .orgY      dd ?
  .wrapW     dd ?
  .wrapH     dd ?

  .lock      TMutex
ends



; Creates memory image with parameters specified.
interface CreateImage, .width, .height


interface DestroyImage, .ptrImage


; Draws the image on a OS provided window surface.
interface DrawImageRect, .where, .pImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height



proc LockImg, .pImage
begin
        push    eax

        mov     eax, [.pImage]
        lea     eax, [eax+TImage.lock]
        stdcall WaitForMutex, eax, 100
;        jc      .error

        pop     eax
        return

;.error:
;;        OutputValue "Unable to lock image. Error=", eax, 10, -1
;;
;;        OutputCallStack 2
;;        OutputLn
;
;;        mov     eax, [ebp+4]
;;        OutputValue "Unable to lock image on: ", eax, 16, 8
;;        stc
;        pop      eax
;        return
endp


proc UnlockImg, .pImage
begin
        push    eax
        mov     eax, [.pImage]
        lea     eax, [eax+TImage.lock]
        stdcall MutexRelease, eax
        pop     eax
        return
endp


; Sets the wrap parameters of the image.
; Arguments:
;    .pImage - pointer to the TImage
;    .newWrapW - new wrap width
;    .newWrapH - new wrap height
;
; After this operation, the orgX and orgY of the image are set both to 0
; without affecting the image itself.
;
; As long as on moving orgX and orgY, the image need to be rearanged, the operation is relatively slow.
; So, use it carefully.

proc SetImageWrap, .pImage, .newWrapW, .newWrapH
begin
        pushad
        mov     ebx, [.pImage]
        stdcall LockImg, ebx
        jc      .finish

        stdcall ZeroOrgImage, ebx

        mov     ecx, [.newWrapW]
        mov     edx, [.newWrapH]
        cmp     ecx, [ebx+TImage.width]
        cmova   ecx, [ebx+TImage.width]
        cmp     edx, [ebx+TImage.height]
        cmova   edx, [ebx+TImage.height]

        mov     eax, 1
        test    ecx, ecx
        cmovz   ecx, eax
        test    edx, edx
        cmovz   edx, eax

        mov     [ebx+TImage.wrapW], ecx
        mov     [ebx+TImage.wrapH], edx

        stdcall UnlockImg, ebx
        clc

.finish:
        popad
        return
endp




proc ZeroOrgImage, .pImage
.imgL dd ?
.imgU dd ?
begin
        pushad

        mov     ebx, [.pImage]

        xor     edx, edx
        mov     [.imgL], edx
        mov     [.imgU], edx

        cmp     [ebx+TImage.orgX], edx
        je      .imgL_ok

        stdcall CreateImage, [ebx+TImage.orgX], [ebx+TImage.wrapH]
        mov     [.imgL], eax
        stdcall CopyImageRect, eax, 0, 0, ebx, 0, 0, [ebx+TImage.orgX], [ebx+TImage.wrapH]

.imgL_ok:
        cmp     [ebx+TImage.orgY], edx
        je      .imgU_ok

        stdcall CreateImage, [ebx+TImage.wrapW], [ebx+TImage.orgY]
        mov     [.imgU], eax
        mov     ecx, [ebx+TImage.wrapW]
        sub     ecx, [ebx+TImage.orgX]
        stdcall CopyImageRect, eax, 0, 0, ebx, [ebx+TImage.orgX], 0, ecx, [ebx+TImage.orgY]

.imgU_ok:
        mov     ecx, [ebx+TImage.wrapW]
        mov     edi, [ebx+TImage.wrapH]
        sub     ecx, [ebx+TImage.orgX]
        sub     edi, [ebx+TImage.orgY]
        xor     esi, esi

        cmp     [ebx+TImage.orgX], edx
        je      .scrollx_ok

        inc     esi
        stdcall ScrollRectLeft2, ebx, 0, [ebx+TImage.orgY], [ebx+TImage.wrapW], edi, [ebx+TImage.orgX]
        stdcall CopyImageRect, ebx, ecx, [ebx+TImage.orgY], [.imgL], 0, [ebx+TImage.orgY], [ebx+TImage.orgX], edi

.scrollx_ok:

        cmp     [ebx+TImage.orgY], edx
        je      .scrolly_ok

        inc     esi
        stdcall ScrollRectUp2, ebx, 0, 0, [ebx+TImage.wrapW], [ebx+TImage.wrapH], [ebx+TImage.orgY]
        stdcall CopyImageRect, ebx, 0, edi, [.imgU], 0, 0, ecx, [ebx+TImage.orgY]

.scrolly_ok:
        cmp     esi, 2
        jne     .lower_right_ok

        stdcall CopyImageRect, ebx, ecx, edi, [.imgL], 0, 0, [ebx+TImage.orgX], [ebx+TImage.orgY]

.lower_right_ok:

        mov     [ebx+TImage.orgX], edx
        mov     [ebx+TImage.orgY], edx


        stdcall DestroyImage, [.imgL]
        stdcall DestroyImage, [.imgU]

        popad
        return

.moveit:



endp



proc BlendImage, .pDstImage, .xDst, .yDst, .pSrcImage, .xSrc, .ySrc, .width, .height
.srect RECT
.drect RECT
.rect  RECT

.srcYstep dd ?
.dstYstep dd ?

.srcWrapX dd ?
.dstWrapX dd ?

.srcWrapY dd ?
.dstWrapY dd ?

.srcStartX dd ?
.dstStartX dd ?

begin
        pushad

        mov     esi, [.pSrcImage]
        mov     edi, [.pDstImage]

        xor     eax, eax
        mov     ecx, [esi+TImage.wrapW]
        mov     edx, [esi+TImage.wrapH]

        mov     [.srect.left], eax
        mov     [.srect.top], eax
        mov     [.srect.right], ecx
        mov     [.srect.bottom], edx

        mov     ecx, [edi+TImage.wrapW]
        mov     edx, [edi+TImage.wrapH]

        mov     [.drect.left], eax
        mov     [.drect.top], eax
        mov     [.drect.right], ecx
        mov     [.drect.bottom], edx

        mov     ecx, [.xSrc]
        mov     edx, [.ySrc]

        mov     [.rect.left], ecx
        mov     [.rect.top], edx

        add     ecx, [.width]
        add     edx, [.height]

        mov     [.rect.right], ecx
        mov     [.rect.bottom], edx

        lea     ecx, [.srect]
        lea     edx, [.rect]
        stdcall RectIntersect2, ecx, edx
        jc      .finish

        mov     ecx, [.xSrc]
        mov     edx, [.ySrc]
        sub     ecx, [.xDst]
        sub     edx, [.yDst]

; convert the rect to the destination coordinates.
        sub     [.srect.left], ecx
        sub     [.srect.top], edx
        sub     [.srect.right], ecx
        sub     [.srect.bottom], edx

        lea     eax, [.srect]
        lea     ebx, [.drect]
        stdcall RectIntersect2, ebx, eax
        jc      .finish

        stdcall RectCopy, eax, ebx      ; srect = drect

;convert back to source coordinates.
        add     [.srect.left], ecx
        add     [.srect.top], edx
        add     [.srect.right], ecx
        add     [.srect.bottom], edx

; here [.srect] contains the source rectange
; [.drect] contains the destination rectangle
; the width and height of the rectangles  is the same.

        mov     eax, [.srect.top]
        sub     [.srect.bottom], eax    ; only the rect height is important.

        mov     ecx, [esi+TImage.width]
        mov     edx, [edi+TImage.width]
        lea     ecx, [4*ecx]
        lea     edx, [4*edx]

        mov     [.srcYstep], ecx
        mov     [.dstYstep], edx

        mov     ecx, [esi+TImage.wrapH]
        mov     edx, [edi+TImage.wrapH]
        imul    ecx, [.srcYstep]
        imul    edx, [.dstYstep]
        add     ecx, [esi+TImage.pPixels]
        add     edx, [edi+TImage.pPixels]
        mov     [.srcWrapY], ecx
        mov     [.dstWrapY], edx


        mov     eax, [.srect.right]
        mov     ebx, [.drect.right]

        add     eax, [esi+TImage.orgX]
        add     ebx, [edi+TImage.orgY]

        cmp     eax, [esi+TImage.wrapW]
        jb      @f
        sub     eax, [esi+TImage.wrapW]
@@:

        cmp     ebx, [edi+TImage.wrapW]
        jb      @f
        sub     ebx, [edi+TImage.wrapW]
@@:
        mov     [.srcStartX], eax
        mov     [.dstStartX], ebx

        mov     eax, [esi+TImage.wrapW]
        mov     ebx, [edi+TImage.wrapW]
        dec     eax
        dec     ebx
        mov     [.srcWrapX], eax
        mov     [.dstWrapX], ebx

        mov     ecx, [.srect.top]
        mov     edx, [.drect.top]
        add     ecx, [esi+TImage.orgY]
        add     edx, [edi+TImage.orgY]

        cmp     ecx, [esi+TImage.wrapH]
        jb      @f
        sub     ecx, [esi+TImage.wrapH]
@@:

        cmp     edx, [edi+TImage.wrapH]
        jb      @f
        sub     edx, [edi+TImage.wrapH]
@@:
        imul    ecx, [.srcYstep]
        imul    edx, [.dstYstep]
        add     ecx, [esi+TImage.pPixels]
        add     edx, [edi+TImage.pPixels]

; lock the images.
        stdcall LockImg, esi
        jc      .finish

        stdcall LockImg, edi
        jc      .finish2

; prepare MMX constants

        pxor    mm0, mm0        ; mm0 = 0000 0000 0000 0000
        pcmpeqw mm1, mm1                                        ; mm1 = ffff ffff ffff ffff
        psrlw   mm1, 8                                         ; mm1 = 00ff 00ff 00ff 00ff

;        psrlw   mm1, 15                                         ; mm1 = 0001 0001 0001 0001
;        psllw   mm1, 8          ; mm1 = 0100 0100 0100 0100

.loopY:
        mov     eax, [.srect.right]
        sub     eax, [.srect.left]
        mov     [.width], eax

        mov     eax, [.srcStartX]
        mov     ebx, [.dstStartX]

.loopX:
        dec     eax
        cmovs   eax, [.srcWrapX]

        dec     ebx
        cmovs   ebx, [.dstWrapX]

        movd    mm7, [ecx+4*eax]        ; source pixel.
        movd    mm6, [edx+4*ebx]        ; destination pixel.

        movq    mm5, mm7
        psrlq   mm5, 24                 ; alpha source
        punpckldq mm5, mm5
        packssdw  mm5, mm5              ; mm5 = As, As, As, As

        movq    mm4, mm1
        psubw   mm4, mm5                ; mm4 = Bs Bs Bs Bs     (Bs = $100 - As)

        punpcklbw mm6, mm0              ; byte to word destination pixel
        punpcklbw mm7, mm0              ; byte to word source pixel
        psllw     mm7, 8                ; Cs = Cs*256

        pmullw  mm6, mm4                ; mm6 = Cd = Cd * (256-As) = Cd * Bs   (For alpha Ad = Ad*Bs)
        paddusw mm6, mm7                ; Cd = Cs + Cd
        psrlw   mm6, 8                  ; Cd = Cd/256

        packuswb mm6, mm0               ; pack words to bytes 0A 0R 0G 0B -> 0000ARGB
        movd     [edx+4*ebx], mm6

        dec     [.width]
        jnz     .loopX

.nextY:
        add     ecx, [.srcYstep]
        cmp     ecx, [.srcWrapY]
        cmove   ecx, [esi+TImage.pPixels]

        add     edx, [.dstYstep]
        cmp     edx, [.dstWrapY]
        cmove   edx, [edi+TImage.pPixels]

        dec     [.srect.bottom]
        jnz     .loopY

        emms

; unlock the images.
        stdcall UnlockImg, edi

.finish2:
        stdcall UnlockImg, esi

.finish:
        popad
        return
endp



proc BlendImageTileWrap, .pDstImage, .xDst, .yDst, .wDst, .hDst, .pSrcImage

.srcXOffs dd ?
.srcXwrap dd ?

.srcYoffs dd ?
.srcYwrap dd ?

.dstXbegn dd ?
.dstXwrap dd ?
.dstYoffs dd ?
.dstYwrap dd ?

.stepYSrc dd ?
.stepYDst dd ?

.srcY0    dd ?
.dstY0    dd ?

begin
        pushad

        mov     esi, [.pSrcImage]
        mov     edi, [.pDstImage]

; lock the images.
        stdcall LockImg, esi
        jc      .finish

        stdcall LockImg, edi
        jc      .finish2

; step Y, source, destination

        mov     ecx, [esi+TImage.width]
        mov     edx, [edi+TImage.width]
        shl     ecx, 2
        shl     edx, 2
        mov     [.stepYSrc], ecx
        mov     [.stepYDst], edx

; Destination Y offset + wrap:

        mov     eax, [.yDst]
        mov     ebx, [edi+TImage.wrapH]

        add     eax, [edi+TImage.orgY]
        cmp     eax, ebx
        jb      @f
        sub     eax, ebx
@@:
        mov     ecx, [edi+TImage.pPixels]

        imul    eax, edx
        imul    ebx, edx
        add     eax, ecx
        add     ebx, ecx

        mov     [.dstYoffs], eax
        mov     [.dstYwrap], ebx
        mov     [.dstY0], ecx

        mov     ecx, [edi+TImage.wrapW]
        mov     [.dstXwrap], ecx

; Destination X start:

        mov     eax, [.xDst]
        add     eax, [edi+TImage.orgX]
        cmp     eax, ecx
        jb      @f
        sub     eax, ecx
@@:
        mov     [.dstXbegn], eax

; Source offset X,Y

        mov     eax, [.xDst]
        mov     ecx, [esi+TImage.wrapW]
        cdq
        div     ecx
        add     edx, [esi+TImage.orgX]
        cmp     edx, ecx
        jb      @f
        sub     edx, ecx
@@:
        mov     [.srcXOffs], edx
        mov     [.srcXwrap], ecx

        mov     eax, [.yDst]
        mov     ecx, [esi+TImage.wrapH]
        mov     ebx, [esi+TImage.pPixels]
        cdq
        div     ecx
        add     edx, [esi+TImage.orgY]
        cmp     edx, ecx
        jb      @f
        sub     edx, ecx
@@:
        imul    edx, [.stepYSrc]
        imul    ecx, [.stepYSrc]
        add     edx, ebx
        add     ecx, ebx
        mov     [.srcYoffs], edx
        mov     [.srcYwrap], ecx
        mov     [.srcY0],    ebx

; prepare MMX constants

        pxor    mm0, mm0        ; mm0 = 0000 0000 0000 0000
        pcmpeqw mm1, mm1                                        ; mm1 = ffff ffff ffff ffff
        psrlw   mm1, 8                                         ; mm1 = 00ff 00ff 00ff 00ff

;        psrlw   mm1, 15                                         ; mm1 = 0001 0001 0001 0001
;        psllw   mm1, 8                                          ; mm1 = 0100 0100 0100 0100

        xor     ebx, ebx
        mov     esi, [.srcYoffs]
        mov     edi, [.dstYoffs]


.loopY:
        dec     [.hDst]
        js      .exit

        mov     ecx, [.srcXOffs]
        mov     edx, [.dstXbegn]
        mov     eax, [.wDst]

.loopX:
        dec     eax
        js      .nextY

        movd    mm7, [esi+4*ecx]        ; source pixel.
        movd    mm6, [edi+4*edx]        ; destination pixel.

        movq    mm5, mm7
        psrlq   mm5, 24                 ; alpha source
        punpckldq mm5, mm5
        packssdw  mm5, mm5              ; mm5 = As, As, As, As

        movq    mm4, mm1
        psubw   mm4, mm5                ; mm4 = Bs Bs Bs Bs     (Bs = $100 - As)

        punpcklbw mm6, mm0              ; byte to word destination pixel
        punpcklbw mm7, mm0              ; byte to word source pixel
        psllw     mm7, 8                ; Cs = Cs*256

        pmullw  mm6, mm4                ; mm6 = Cd = Cd * (256-As) = Cd * Bs   (For alpha Ad = Ad*Bs)
        paddusw mm6, mm7                ; Cd = Cs + Cd
        psrlw   mm6, 8                  ; Cd = Cd/256

        packuswb mm6, mm0               ; pack words to bytes 0A 0R 0G 0B -> 0000ARGB
        movd     [edi+4*edx], mm6

        inc     ecx
        inc     edx

        cmp     ecx, [.srcXwrap]
        cmovae  ecx, ebx

        cmp     edx, [.dstXwrap]
        cmovae  edx, ebx

        jmp     .loopX

.nextY:
        add     esi, [.stepYSrc]
        cmp     esi, [.srcYwrap]
        cmovae  esi, [.srcY0]

        add     edi, [.stepYDst]
        cmp     edi, [.dstYwrap]
        cmovae  edi, [.dstY0]
        jmp     .loopY

.exit:
        emms

; unlock the images.
        stdcall UnlockImg, [.pDstImage]

.finish2:
        stdcall UnlockImg, [.pSrcImage]

.finish:
        popad
        return
endp






proc CopyImageRect, .pDstImage, .xDst, .yDst, .pSrcImage, .xSrc, .ySrc, .width, .height
.srect RECT
.drect RECT
.rect  RECT
begin
        pushad

        mov     esi, [.pSrcImage]
        mov     edi, [.pDstImage]

        xor     eax, eax
        mov     ecx, [esi+TImage.width]
        mov     edx, [esi+TImage.height]

        mov     [.srect.left], eax
        mov     [.srect.top], eax
        mov     [.srect.right], ecx
        mov     [.srect.bottom], edx

        mov     ecx, [edi+TImage.width]
        mov     edx, [edi+TImage.height]

        mov     [.drect.left], eax
        mov     [.drect.top], eax
        mov     [.drect.right], ecx
        mov     [.drect.bottom], edx

        mov     ecx, [.xSrc]
        mov     edx, [.ySrc]

        mov     [.rect.left], ecx
        mov     [.rect.top], edx

        add     ecx, [.width]
        add     edx, [.height]

        mov     [.rect.right], ecx
        mov     [.rect.bottom], edx

        lea     ecx, [.srect]
        lea     edx, [.rect]
        stdcall RectIntersect, ecx, ecx, edx
        jc      .finish

        mov     ecx, [.xSrc]
        mov     edx, [.ySrc]
        sub     ecx, [.xDst]
        sub     edx, [.yDst]

; convert the rect to the destination coordinates.
        sub     [.srect.left], ecx
        sub     [.srect.top], edx
        sub     [.srect.right], ecx
        sub     [.srect.bottom], edx

        lea     eax, [.srect]
        lea     ebx, [.drect]
        stdcall RectIntersect, ebx, ebx, eax
        jc      .finish

        stdcall RectCopy, eax, ebx      ; srect = drect

;convert back to source coordinates.
        add     [.srect.left], ecx
        add     [.srect.top], edx
        add     [.srect.right], ecx
        add     [.srect.bottom], edx

; here [.srect] contains the source rectange
; [.drect] contains the destination rectangle
; the width and height of the rectangles  is the same.

        mov     ecx, [.srect.left]
        mov     edx, [.srect.top]
        sub     [.srect.right], ecx
        sub     [.srect.bottom], edx

        mov     ecx, [.srect.top]
        mov     edx, [.drect.top]

        imul    ecx, [esi+TImage.width]
        imul    edx, [edi+TImage.width]

        add     ecx, [.srect.left]
        add     edx, [.drect.left]

        shl     ecx, 2
        shl     edx, 2

        add     ecx, [esi+TImage.pPixels]
        add     edx, [edi+TImage.pPixels]

        stdcall LockImg, esi
        jc      .finish

        stdcall LockImg, edi
        jc      .finish2

        mov     esi, [esi+TImage.width]
        mov     edi, [edi+TImage.width]
        lea     esi, [4*esi]
        lea     edi, [4*edi]


.loopY:
        mov     ebx, [.srect.right]     ; the width of the source rectangle
        test    ebx, 1
        jz      .loopX

; copy 1 dword

        dec     ebx
        mov     eax, [ecx+4*ebx]
        mov     [edx+4*ebx], eax
        jz      .nextY

.loopX:
        sub     ebx, 2

        movq    mm0, [ecx+4*ebx]
        movq    [edx+4*ebx], mm0

        jnz     .loopX

.nextY:
        add     ecx, esi
        add     edx, edi

        dec     [.srect.bottom]
        jnz     .loopY

        emms
        stdcall UnlockImg, [.pDstImage]

.finish2:
        stdcall UnlockImg, [.pSrcImage]

.finish:
        popad
        return
endp



;proc ImageMulAlpha2, .pImage, .x, .y, .width, .height
;.drect RECT
;.rect  RECT
;begin
;        pushad
;
;        mov     edi, [.pImage]
;
;        xor     eax, eax
;        mov     ecx, [edi+TImage.width]
;        mov     edx, [edi+TImage.height]
;
;        mov     [.drect.left], eax
;        mov     [.drect.top], eax
;        mov     [.drect.right], ecx
;        mov     [.drect.bottom], edx
;
;        mov     ecx, [.x]
;        mov     edx, [.y]
;
;        mov     [.rect.left], ecx
;        mov     [.rect.top], edx
;
;        add     ecx, [.width]
;        add     edx, [.height]
;
;        mov     [.rect.right], ecx
;        mov     [.rect.bottom], edx
;
;        lea     ecx, [.drect]
;        lea     edx, [.rect]
;        stdcall RectIntersect, ecx, ecx, edx
;        jc      .finish
;
;        mov     ecx, [.drect.left]
;        mov     edx, [.drect.top]
;        sub     [.drect.right], ecx
;        sub     [.drect.bottom], edx
;
;        mov     esi, [.drect.top]
;        imul    esi, [edi+TImage.width]
;
;        add     esi, [.drect.left]
;        lea     esi, [4*esi]
;        add     esi, [edi+TImage.pPixels]
;
;        mov     edi, [edi+TImage.width]
;        lea     edi, [4*edi]
;
;; prepare MMX constants
;
;        pxor    mm0, mm0        ; mm0 = 0000 0000 0000 0000
;        pcmpeqw mm1, mm1                                        ; mm1 = ffff ffff ffff ffff
;        psrlw   mm1, 15                                         ; mm1 = 0001 0001 0001 0001
;        psllw   mm1, 8          ; mm1 = 0100 0100 0100 0100
;        psllq   mm1, 48         ; mm1 = 0100 0000 0000 0000
;
;.loopY:
;        mov     ebx, [.drect.right]     ; the width of the source rectangle
;
;.loopX:
;        dec     ebx                     ; The ZF will not be changed later in the inner loop.
;
;        movd    mm6, [esi+4*ebx]        ; destination pixel.
;
;        movq    mm5, mm6
;        psrlq   mm5, 24                 ; alpha
;
;        punpckldq mm5, mm5              ;
;        punpcklbw mm6, mm0              ; byte to word destination pixel
;        packssdw  mm5, mm5              ; mm5 = Ad, Ad, Ad, Ad
;
;        psrlq   mm5, 16                 ; mm5 = 0  Ad Ad Ad
;        por     mm5, mm1                ; mm5 = $100 Ad Ad Ad
;
;        pmullw  mm6, mm5                ; mm6 = Cd = Cd * Ad   (For alpha Ad = Ad * 256)
;        psrlw   mm6, 8                  ; Cd = Cd/256          (For alpha Ad = Ad)
;
;        packuswb mm6, mm0               ; pack words to bytes 0A 0R 0G 0B -> 0000ARGB
;        movd     [esi+4*ebx], mm6
;
;        jnz     .loopX
;
;.nextY:
;        add     esi, edi
;
;        dec     [.drect.bottom]
;        jnz     .loopY
;
;.finish:
;        emms
;        popad
;        return
;endp


; the colors must be pre-multiplied ARGB.

proc _BlendPixel, .dstColor, .srcColor
begin
        pushad

        movzx   ecx, byte [.srcColor+3]      ; As
        mov     ebx, $ff
        mov     esi, ebx

        sub     ebx, ecx                     ; 1 - As

        movzx   eax, byte [.dstColor+3]      ; Ad
        imul    ecx, esi
        mul     ebx
        add     eax, ecx
        adc     edx, 0
        div     esi

        mov     byte [esp+4*regEAX+3], al


        movzx   eax, byte [.dstColor]       ; Bd
        movzx   ecx, byte [.srcColor]       ; Bs

        imul    ecx, esi
        mul     ebx
        add     eax, ecx
        adc     edx, 0
        div     esi

        mov     byte [esp+4*regEAX], al



        movzx   eax, byte [.dstColor+1]       ; Gd
        movzx   ecx, byte [.srcColor+1]       ; Gs

        imul    ecx, esi
        mul     ebx
        add     eax, ecx
        adc     edx, 0
        div     esi

        mov     byte [esp+4*regEAX+1], al



        movzx   eax, byte [.dstColor+2]       ; Rd
        movzx   ecx, byte [.srcColor+2]       ; Rs

        imul    ecx, esi
        mul     ebx
        add     eax, ecx
        adc     edx, 0
        div     esi

        mov     byte [esp+4*regEAX+2], al


        popad
        return
endp





include '%TargetOS%/images.asm'

endmodule