; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Main procedure of GUI application library.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: Organize the main message/event loop needed by every GUI engine.
;         This file contains only OS independent part and includes OS dependent files.
;_________________________________________________________________________________________
module "Main library"

uglobal
  var pApplication = ?

  var __LastPointedWindow = ?
  var __FocusedWindow = ?
  var __MouseTarget = ?

  var __ActiveMenu = ?
endg


;_________________________________________________________________________________________
;
; Process all the events queued from the system and exits.
; Arguments: none
; Returns: CF = 1 if the termination event has been received.
;_________________________________________________________________________________________

interface ProcessSystemEvents




;_________________________________________________________________________________________
;
; Waits for system events to arive.
; During the wait, the application
; does not use CPU resources.
;_________________________________________________________________________________________

interface WaitForSystemEvent





proc Run
begin
.mainloop:
        stdcall ProcessSystemEvents
        jc      .terminate

        mov     eax, [pApplication]
        test    eax, eax
        jz      .eventok

        get     ecx, eax, TApplication:OnIdle
        jecxz   .eventok

        stdcall ecx, eax

.eventok:
if defined Caret
        stdcall __CaretTimerProc
end if
        stdcall WaitForSystemEvent
        jc      .eventok
        jmp     .mainloop

.terminate:
        DebugMsg "Terminate GUI application!"
        return
endp



proc ServeMenuMouseMove, .window, .x, .y, .kbdState
begin
        cmp     [__ActiveMenu], 0
        jne     .serve_menu

        clc
        return

.serve_menu:
        pushad

        stdcall _ClientToScreen, [.window], [.x], [.y]
        exec    [__ActiveMenu], TWindow:EventMouseMove, ecx, edx, [.kbdState]

        popad
        stc
        return
endp



proc ServeMenuButtonPress, .window, .button, .kbdState, .x, .y
begin
        cmp     [__ActiveMenu], 0
        jne     .serve_menu

        clc
        return

.serve_menu:
        pushad

        stdcall _ClientToScreen, [.window], [.x], [.y]
        exec    [__ActiveMenu], TWindow:EventButtonPress, [.button], [.kbdState], ecx, edx

        popad
        stc
        return
endp



proc ShowModal, .pForm, .pParent
.prev dd ?
begin
        push    ebx ecx edx

        mov     ecx, [.pParent]
        mov     edx, [.pForm]
        test    ecx, ecx
        jz      .modal_parent_ok

; center the window on the parent.

        mov     eax, [ecx+TWindow._width]
        mov     ebx, [ecx+TWindow._height]
        sub     eax, [edx+TWindow._width]
        sub     ebx, [edx+TWindow._height]
        sar     eax, 1
        sar     ebx, 1
        add     eax, [ecx+TWindow._x]
        add     ebx, [ecx+TWindow._y]

;TODO: Min/Max coordinates range check.

        exec    edx, TWindow:Move, eax, ebx

.modal_parent_ok:

        set     edx, TForm:Visible, TRUE

        test    ecx, ecx
        jz      .parent_ok2

        stdcall _SetModalTowards, [edx+TForm.handle], [ecx+TWindow.handle]
        set     ecx, TWindow:Enabled, FALSE

.parent_ok2:
        set     edx, TForm:ModalResult, 0

        get     eax, edx, TForm:OnClose
        mov     [.prev], eax

        set     edx, TForm:OnClose, __ModalOnClose

.modal_loop:
        stdcall ProcessSystemEvents
        jc      .exit_modal

        get     eax, edx, TForm:ModalResult
        test    eax, eax
        jnz     .exit_modal

        get     eax, edx, TWindow:Visible
        test    eax, eax
        jz      .exit_modal

.vis_ok:
        stdcall WaitForSystemEvent
        jmp     .modal_loop

.exit_modal:
        jecxz   @f
        set     ecx, TWindow:Enabled, TRUE
@@:

        set     edx, TForm:OnClose, [.prev]

        jecxz   @f
        stdcall _FinalizeModal, [edx+TForm.handle], [ecx+TWindow.handle]
@@:

        set     edx, TForm:Visible, FALSE
        get     eax, edx, TForm:ModalResult
        pop     edx ecx ebx
        return
endp


; prevent destroy of the form from the close button or Alt+F4
proc __ModalOnClose, .self, .reason
begin
        set     [.self], TForm:ModalResult, mrCancel
        stc
        return
endp





proc __GetRelativeXY, .window, .x, .y
begin
        push    eax esi

        mov     ecx, [.x]
        mov     edx, [.y]
        mov     esi, [.window]

.loop:
        get     eax, esi, TWindow:Parent
        test    eax, eax
        jz      .finish

        sub     ecx, [esi+TWindow._x]
        sub     edx, [esi+TWindow._y]

        mov     esi, eax
        jmp     .loop

.finish:
        pop     esi eax
        return
endp



proc __SearchWindowToFocus, .pWindow
begin
        pushad
        mov     esi, [.pWindow]

        get     eax, esi, TWindow:Enabled
        test    eax, eax
        jz      .no

        get     eax, esi, TWindow:Visible
        test    eax, eax
        jz      .no

        get     eax, esi, TWindow:WantFocus
        test    eax, eax
        jz      .children

        mov     [esp+4*regEAX], esi
        popad
        return


.children:
        get     edi, esi, TWindow:Children
        test    edi, edi
        jnz     .loopit

.no:
        xor     eax, eax
        mov     [esp+4*regEAX], eax
        popad
        return

.loopit:
        mov     ecx, [edi+TArray.count]
        jecxz   .no

.children_loop:
        dec     ecx
        js      .no

        mov     esi, [edi+TArray.array+4*ecx]

        istype  esi, TWindow
        jne     .children_loop

        stdcall __SearchWindowToFocus, esi
        test    eax, eax
        jz      .children_loop

        mov     [esp+4*regEAX], eax
        popad
        return
endp




iglobal

if used FramesPerSecond

  fpsTime dd 0

  fps     dd 0
  fpsmin  dd $ffffffff
  fpsmax  dd 0

  fpscnt  dd 0

end if

endg

fpsCount = 16

proc FramesPerSecond
begin
        pushad

        inc     [fpscnt]
        cmp     [fpscnt], fpsCount
        jne     .finish

        mov     [fpscnt], 0

        stdcall GetFineTimestamp
        mov     ecx, eax

        xchg    ecx, [fpsTime]
        test    ecx, ecx
        jz      .finish

        sub     ecx, [fpsTime]
        neg     ecx
        mov     eax, fpsCount*1000000
        xor     edx, edx

        div     ecx

        mov     [fps], eax

        cmp     eax, [fpsmin]
        jae      @f
        mov     [fpsmin], eax
@@:

        cmp     eax, [fpsmax]
        jb      @f
        mov     [fpsmax], eax
@@:


if defined options.DebugMode.ShowFPS & options.DebugMode.ShowFPS
        stdcall FileWriteString, [STDOUT], <txt '  FPS: '>
        stdcall NumToStr, [fps], ntsDec or ntsUnsigned
        push    eax
        stdcall FileWriteString, [STDOUT], eax
        stdcall StrDel ; from the stack

        stdcall FileWriteString, [STDOUT], <txt ', min: '>
        stdcall NumToStr, [fpsmin], ntsDec or ntsUnsigned
        push    eax
        stdcall FileWriteString, [STDOUT], eax
        stdcall StrDel ; from the stack

        stdcall FileWriteString, [STDOUT], <txt ', max: '>
        stdcall NumToStr, [fpsmax], ntsDec or ntsUnsigned
        push    eax
        stdcall FileWriteString, [STDOUT], eax
        stdcall StrDel ; from the stack
        stdcall FileWriteString, [STDOUT], <txt "                                           ", 13>

      if defined options.DebugMode & options.DebugMode
        stdcall FileWriteString, [STDOUT], <txt 10>
      end if
end if

.finish:
        popad
        return
endp




include '%TargetOS%/Main.asm'

endmodule