; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TAsmEdit object class
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: Advanced text editor.
;_________________________________________________________________________________________

module "TAsmEdit"

; view options: how the editor looks
aeoLeftMargin          = $01
aeoLineNumbers         = $02
aeoStripedBackground   = $04
aeoConsoleCaret        = $08

; navigation options
aeoNavXCaretWrap       = $0100            ; when the caret reaches the horizontal line boundaries (start of line and end of line) the caret wraps to the previous or next line.
aeoScrollYMore         = $0200            ; scroll Y more than needed in order to scroll less and to scroll more readable.
aeoScrollXMore         = $0400            ; scroll X more than needed

; editing options

aeoOverwriteMode       = $010000
aeoBlockSelection      = $020000
aeoAutoIndents         = $040000
aeoAutoBrackets        = $080000
aeoSmartTabs           = $100000
aeoPersistentSelection = $200000

aeoDefault = aeoLeftMargin or aeoLineNumbers or aeoConsoleCaret or aeoAutoIndents or aeoSmartTabs


; Operations for undo/redo:

chopInserted = 0
chopDeleted  = 1
chopChanged  = 2


struct TUndoInfo
  .pNext       dd ?       ; pointer to the next TUndoInfo structure in the chain.

  .caretPos    POINT
  .selStart    POINT
  .selEnd      POINT
  .operation   dd ?       ; change operation.
  .group       dd ?       ;
  .line        dd ?       ; the line number affected.
  .pText       dd ?       ; pointer to the previous TTextLine of the affected line for chopChanged and chopDeleted
ends



rstruct TTextLine
  .flags        dd  ?   ; line flags
  .ch_len       dd  ?   ; the length of the text in utf-8 characters.
                        ;
  .debugdata    dd  ?   ; pointer to debug data, associated to this line
                        ;
  .syn_array    dd  ?   ; TArray of TTextChunk items for syntax colorization.
                        ; Should be updated by calling the syntax highlighting proc.
                        ;
  .syn_context  dd  ?   ; Syntax context after the line
                        ;
  .fold_level   dd  ?   ; the level of folding.

                rd  2
                        ;
  .             TText
ends



struct TTextChunk
  .len   dw ?   ; the length of the text chunk in UTF-8 characters.
  .color db ?   ; byte, index of the color/background of the text chunk.
  .flags db ?   ; the text formatting flags. not used, reserved. For example underlining, bold, etc...
ends



object TAsmEdit, TScrollable

; private fields

  ._lines         dd    ?             ; pointer to TText structure that contains pointers.
  ._line_widths   dd    ?             ; TArray of TCounter for counting the line width. Must be kept in consistency.

  ._imgText       dd    ?
  ._img_txt_valid dd    ?

  ._imgMargin     dd    ?
  ._img_margin_valid dd ?

  ._pUndo         dd    ?
  ._pRedo         dd    ?

  ._options       dd    ?             ; behaviour bit options: aeoXXXXXX values.

  ._font_height   dd    ?
  ._font_asc      dd    ?
  ._font_desc     dd    ?

  ._font_width    dd    ?

  ._numfont_width dd    ?

  ._left_margin   dd    ?
  ._num_margin    dd    ?

  ._caretPos      POINT

  ._selStart      POINT         ; The selection start point
  ._selEnd        POINT         ;
  ._selAreaUpdate dd ?

  ._selCmd        dd    ?       ; ==0 if the shift key wasn't pressed on the last keyboard interaction.
  ._selCmdPrev    dd    ?       ; the previous value of the _selCmd field.

  ._drag_button   dd    ?

  ._width_ch      dd    ?       ; width and
  ._height_ch     dd    ?       ; height of the window in chars. Should be set on resizing and on font change.

  ._syntax_proc   dd    ?

; parameters

  param .Options, ._options, .SetOptions

  param .Selection, .GetSelection, .SetSelection
  param .Text,      .GetText,      .SetText

  param .CaretX,    ._caretPos.x, NONE
  param .CaretY,    ._caretPos.y, NONE
  param .SelStartX, ._selStart.x, NONE
  param .SelStartY, ._selStart.y, NONE
  param .SelEndX, ._selStart.x, NONE
  param .SelEndY, ._selStart.y, NONE

;  param .PosX,      ._scrX.pos, .SetPosX
;  param .PosY,      ._scrY.pos, .SetPosY

  param .SyntaxProc, ._syntax_proc, .SetSyntaxProc

; param methods

  method  .SetOptions, .value

  method  .GetSelection
  method  .SetSelection, .value

  method  .GetText
  method  .SetText, .value

  method  .ComputeLeftMarginWidth
  method  .ComputeScreenInChars

  method  .SetPos, .param, .value

  method  .SetSyntaxProc, .value

  method  ._AddSelRectangles, .pRectList
  method  ._AddSelRectanglesAbs, .pRectList


; common methods

  method .Create, .parent
  method .Destroy

  method .__UpdateImages
  method .Resize, .newWidth, .newHeight
  method .Move, .newX, .newY

  method .UpdateScrollBar

  method .UpdateCaretPos


; rendering methods

  method ._RenderTxt, .pImage, .pRect           ; the rectangle contains coordinates in character positions, relative to the begining of the window.
  method ._RenderLM,  .pImage, .from, .to       ; .from, .to in character positions, relative to the begining of the window.

; Painting methods

  method .SelfPaint, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height    ; Paints itself on the caller provided TImage.

; Text navigarion methods

  method .TextLine, .y

  method .ScrollToCaretX
  method .ScrollToCaretY

  method .CaretLeftRight, .count
  method .CaretUpDown, .count, .scroll

  method .CaretHome
  method .CaretEnd
  method .CaretScrHome
  method .CaretScrEnd
  method .CaretTxtHome
  method .CaretTxtEnd

; Selection control methods.

  method .SelStartHere
  method .SelEndHere

  method .SelChangeBegin
  method .SelChangeEnd


; undo/redo

  method .SaveForUndo, .LineNum, .operation
  method .Undo
  method .Redo


; Text editing methods

  method .AddChar, .utf8

  method .InsertNewLine, .LineNum
  method .DeleteLine, .LineNum
  method .DeleteChar, .LineNum, .CharPos

  method .JoinNextLine, .LineNum        ; if .LineNum < 0; joins the current caret line
  method .SplitLine, .LineNum, .PosX    ; if .LineNum < 0 splits on caret position.

  method .DeleteSelection
  method .ReplaceSelection, .text

  method .InsertString, .text

  method .AddTrailingSpaces, .LineNum, .PosX  ; if the caret is positioned after the line end, fills with spaces
  method .DelTrailingSpaces, .LineNum         ; deletes the trailing spaces from the line end.

; OS event handlers

  method .EventFocusIn
  method .EventFocusOut

  method .EventScroll, .direction, .command, .value
  method .EventKeyPress, .utf8, .scancode, .kbdState
  method .EventButtonPress, .button, .kbdState, .x, .y
  method .EventButtonRelease, .button, .kbdState, .x, .y
  method .EventMouseMove, .x, .y, .kbdState

endobj


uglobal
  AsmEditUndoGroup dd ?
endg




method TAsmEdit.EventMouseMove   ;, .x, .y, .kbdState
begin
        inherited [.x], [.y], [.kbdState]
        jc        .processed

        pushad
        mov     esi, [.self]

        cmp     [esi+TAsmEdit._drag_button], mbLeft
        jne     .finish

        xor     ecx, ecx

        exec    esi, TAsmEdit:SelChangeBegin

        mov     eax, [.x]
        sub     eax, [esi+TAsmEdit._left_margin]
        cdq
        idiv    [esi+TAsmEdit._font_width]
        shl     edx, 1
        sbb     eax, 0

        get     ebx, esi, TAsmEdit:PosX
        add     eax, ebx
        cmovs   eax, ecx
        mov     [esi+TAsmEdit._caretPos.x], eax
        mov     [esi+TAsmEdit._selEnd.x], eax

        mov     eax, [.y]
        cdq
        idiv    [esi+TAsmEdit._font_height]
        shl     edx, 1
        sbb     eax, 0

        get     ebx, esi, TAsmEdit:PosY
        add     eax, ebx
        cmovs   eax, ecx
        mov     [esi+TAsmEdit._caretPos.y], eax
        mov     [esi+TAsmEdit._selEnd.y], eax

        exec    esi, TAsmEdit:UpdateCaretPos
        exec    esi, TAsmEdit:SelChangeEnd

        exec    esi, TAsmEdit:ScrollToCaretX
        mov     ecx, eax
        exec    esi, TAsmEdit:ScrollToCaretY
        or      ecx, eax
        jz      .finish

; repeat the scroll if the mouse is outside the  window!
        stdcall ZeroMouseMove

.finish:
        popad

.processed:
        return
endp



method TAsmEdit.EventButtonPress        ;, .button, .kbdState, .x, .y
begin
        inherited [.button], [.kbdState], [.x], [.y]
        jc        .processed

        pushad
        mov     esi, [.self]
        mov     eax, [.button]
        mov     [esi+TAsmEdit._drag_button], eax

        stdcall SetMouseCapture, esi

        cmp     eax, mbLeft
        jne     .finish

        exec    esi, TAsmEdit:SelChangeBegin

        xor     ecx, ecx

        mov     eax, [.x]
        sub     eax, [esi+TAsmEdit._left_margin]
        cdq
        idiv    [esi+TAsmEdit._font_width]

        get     ebx, esi, TAsmEdit:PosX
        add     eax, ebx
        cmovs   eax, ecx
        mov     [esi+TAsmEdit._caretPos.x], eax
        mov     [esi+TAsmEdit._selStart.x], eax
        mov     [esi+TAsmEdit._selEnd.x], eax

        mov     eax, [.y]
        cdq
        idiv    [esi+TAsmEdit._font_height]
        get     ebx, esi, TAsmEdit:PosY
        add     eax, ebx
        cmovs   eax, ecx
        mov     [esi+TAsmEdit._caretPos.y], eax
        mov     [esi+TAsmEdit._selStart.y], eax
        mov     [esi+TAsmEdit._selEnd.y], eax

        exec    esi, TAsmEdit:UpdateCaretPos
        exec    esi, TAsmEdit:SelChangeEnd

.finish:
        popad

.processed:
        return
endp



method TAsmEdit.EventButtonRelease      ;, .button, .kbdState, .x, .y
begin
        inherited [.button], [.kbdState], [.x], [.y]

        push    eax
        mov     eax, [.self]
        and     [eax+TAsmEdit._drag_button], 0
        stdcall SetMouseCapture, 0

        pop     eax
        return
endp




method TAsmEdit.EventScroll
begin
        pushad

        mov     esi, [.self]

        cmp     [.direction], scrollX
        je      .horiz_scroll

        cmp     [.command], scWheelUp
        je      .scrollupy

        cmp     [.command], scWheelDn
        je      .scrolldny

.finish:
        exec    esi, TAsmEdit:RectChanged2, 0
        exec    esi, TAsmEdit:UpdateCaretPos
        popad
        return

.scrollupy:
        xor     edx, edx
        get     eax, esi, TAsmEdit:PosY
        sub     eax, [ASM.MouseWheel]
        cmovs   eax, edx
        set     esi, TAsmEdit:PosY, eax
        jmp     .finish


.scrolldny:
        mov     edx, [esi+TAsmEdit._lines]

        mov     eax, [edx+TText.Length]
        add     eax, [edx+TText.GapBegin]
        sub     eax, [edx+TText.GapEnd]
        shr     eax, 2
        mov     edx, eax                                ; the line count in the text.

        get     eax, esi, TAsmEdit:PosY

        add     eax, [ASM.MouseWheel]
        cmp     eax, edx
        cmova   eax, edx

        set     [.self], TAsmEdit:PosY, eax
        jmp     .finish


.horiz_scroll:
        cmp     [.command], scWheelDn
        je      .scrolldnx

        cmp     [.command], scWheelUp
        jne     .finish

.scrollupx:
        get     eax, esi, TAsmEdit:PosX
        xor     edx, edx
        sub     eax, [ASM.MouseWheel]
        cmovs   eax, edx

        set     esi, TAsmEdit:PosX, eax
        jmp     .finish


.scrolldnx:
        or      edx, -1
        shr     edx, 1

        get     eax, esi, TAsmEdit:PosX
        add     eax, [ASM.MouseWheel]
        cmovs   eax, edx
        set     esi, TAsmEdit:PosX, eax
        jmp     .finish
endp





; returns in EDX the rectangle list.

method TAsmEdit._AddSelRectanglesAbs      ;, .pRectList
begin
        pushad

        mov     esi, [.self]
        mov     edx, [.pRectList]

        test    edx, edx
        jnz     .array_ok

        stdcall CreateArray, sizeof.RECT
        mov     edx, eax

.array_ok:

        mov     eax, [esi+TAsmEdit._selStart.x]
        mov     ecx, [esi+TAsmEdit._selStart.y]
        mov     ebx, [esi+TAsmEdit._selEnd.x]
        mov     edi, [esi+TAsmEdit._selEnd.y]

        test    [esi+TAsmEdit._options], aeoBlockSelection
        jz      .add_char

; block selection

        push    eax
        stdcall AddArrayItems, edx, 1
        mov     esi, eax
        pop     eax

        mov     [esi+RECT.left], eax
        mov     [esi+RECT.right], ebx
        mov     [esi+RECT.top], ecx
        mov     [esi+RECT.bottom], edi
        stdcall RectSort, esi
        inc     [esi+RECT.bottom]
        jmp     .end_sel

.add_char:
        cmp     ecx, edi        ; caret_y and selection_y
        jl      .order_ok
        jg      .reverse

        cmp     eax, ebx        ; caret_x and caret_y
        je      .end_sel
        jl      .order_ok

.reverse:
        xchg    eax, ebx
        xchg    ecx, edi

; now eax,ecx are the begining of the selection and ebx, edi the end of the selection.
.order_ok:
        test    eax, eax
        jz      .more_than_one

        push    eax
        stdcall AddArrayItems, edx, 1
        mov     esi, eax
        pop     eax

        mov     [esi+RECT.left], eax
        mov     [esi+RECT.top], ecx
        inc     ecx
        mov     [esi+RECT.bottom], ecx
        mov     [esi+RECT.right], $7fffffff

        cmp     ecx, edi
        je      .maybe_one
        jb      .more_than_one

; one line selection.
        mov     [esi+RECT.right], ebx
        jmp     .end_sel

.maybe_one:
        test    ebx, ebx
        jz      .end_sel

.more_than_one:
        cmp     ecx, edi
        je      .final_rect

        stdcall AddArrayItems, edx, 1
        mov     esi, eax

        mov     [esi+RECT.left], 0
        mov     [esi+RECT.top], ecx
        mov     [esi+RECT.right], $7fffffff
        mov     [esi+RECT.bottom], edi

.final_rect:
        test    ebx, ebx
        jz      .end_sel

        stdcall AddArrayItems, edx, 1
        mov     esi, eax

        mov     [esi+RECT.left], 0
        mov     [esi+RECT.top], edi
        inc     edi
        mov     [esi+RECT.right], ebx
        mov     [esi+RECT.bottom], edi

.end_sel:
        mov     [esp+4*regEDX], edx
        popad
        return
endp




; returns in EDX the rectangle list.

method TAsmEdit._AddSelRectangles       ;, .pRectList
.rect1 RECT
.rect2 RECT
.rect3 RECT

.client RECT
begin
        pushad

        xor     eax, eax
        lea     edi, [.rect1]
        mov     ecx, sizeof.RECT * 4 / 4

        rep stosd

        mov     esi, [.self]
        mov     edx, [.pRectList]

        test    edx, edx
        jnz     .array_ok
        stdcall CreateArray, sizeof.RECT
        mov     edx, eax
.array_ok:

        mov     eax, [esi+TAsmEdit._width_ch]
        mov     ecx, [esi+TAsmEdit._height_ch]
        mov     [.client.right], eax
        mov     [.client.bottom], ecx

        mov     eax, [esi+TAsmEdit._selStart.x]
        mov     ecx, [esi+TAsmEdit._selStart.y]
        mov     ebx, [esi+TAsmEdit._selEnd.x]
        mov     edi, [esi+TAsmEdit._selEnd.y]

        test    [esi+TAsmEdit._options], aeoBlockSelection
        jnz     .add_block

        cmp     ecx, edi        ; caret_y and selection_y
        jl      .order_ok
        jg      .reverse

        cmp     eax, ebx        ; caret_x and caret_y
        je      .finish
        jl      .order_ok

.reverse:
        xchg    eax, ebx
        xchg    ecx, edi

; now eax,ecx is the begining of the selection and ebx, edi the end of the selection.
.order_ok:
        mov     [.rect1.left], eax
        mov     [.rect1.top], ecx
        mov     [.rect1.right], $7fffffff
        inc     ecx
        mov     [.rect1.bottom], ecx

        cmp     ecx, edi        ; begin_y and end_y
        jl      .first_ok
        je      .middle_ok

; there is only one line selection

        mov     [.rect1.right], ebx
        jmp     .rect1_only

.first_ok:
;        mov     [.rect2.left], 0       ; already 0
        mov     [.rect2.top], ecx
        mov     [.rect2.right], $7fffffff
        mov     [.rect2.bottom], edi

.middle_ok:
;        mov     [.rect3.left], 0       ; already 0
        mov     [.rect3.top], edi
        mov     [.rect3.right], ebx

        inc     edi
        mov     [.rect3.bottom], edi

; now offset rectangles

        get     eax, esi, TAsmEdit:PosX
        get     ecx, esi, TAsmEdit:PosY

        sub     [.rect1.left], eax
        sub     [.rect1.top], ecx
        sub     [.rect1.right], eax
        sub     [.rect1.bottom], ecx

        sub     [.rect2.left], eax
        sub     [.rect2.top], ecx
        sub     [.rect2.right], eax
        sub     [.rect2.bottom], ecx

        sub     [.rect3.left], eax
        sub     [.rect3.top], ecx
        sub     [.rect3.right], eax
        sub     [.rect3.bottom], ecx

        lea     ecx, [.client]
        lea     eax, [.rect1]

        stdcall RectIntersect2, eax, ecx
        jc      .rect1_ok

        stdcall RectListAdd, edx, eax

.rect1_ok:
        lea     eax, [.rect2]
        stdcall RectIntersect2, eax, ecx
        jc      .rect2_ok

        stdcall RectListAdd, edx, eax

.rect2_ok:
        lea     eax, [.rect3]
        stdcall RectIntersect2, eax, ecx
        jc      .finish

        stdcall RectListAdd, edx, eax
        jmp     .finish


.add_block:
        mov     [.rect1.left], eax
        mov     [.rect1.right], ebx
        mov     [.rect1.top], ecx
        mov     [.rect1.bottom], edi
        lea     eax, [.rect1]
        stdcall RectSort, eax
        inc     [.rect1.bottom]

.rect1_only:
        get     eax, esi, TAsmEdit:PosX
        get     ecx, esi, TAsmEdit:PosY

        sub     [.rect1.left], eax
        sub     [.rect1.top], ecx
        sub     [.rect1.right], eax
        sub     [.rect1.bottom], ecx

        lea     eax, [.rect1]
        lea     ecx, [.client]
        stdcall RectIntersect2, eax, ecx
        jc      .finish

        stdcall RectListAdd, edx, eax

.finish:
        mov     [esp+4*regEDX], edx
        popad
        return
endp




method TAsmEdit.SetPos
.render RECT
begin
        pushad
        mov     esi, [.self]

        xor     ecx, ecx
        mov     eax, [esi+TAsmEdit._width_ch]
        mov     ebx, [esi+TAsmEdit._height_ch]
        mov     [.render.left], ecx
        mov     [.render.top], ecx
        mov     [.render.right], eax
        mov     [.render.bottom], ebx

        mov     eax, [.value]

        cmp     [.param], TAsmEdit.PosY
        je      .setpos_Y

        xchg    eax, [esi+TAsmEdit._scrX.pos]
        sub     eax, [esi+TAsmEdit._scrX.pos]        ; eax < 0 = scroll left; eax > 0 = scroll right
        jz      .finish

        mov     edx, eax
        jge     @f
        neg     edx
@@:
        cmp     edx, [esi+TAsmEdit._width_ch]
        jge     .do_renderX

        test    eax, eax
        js      .scroll_left

        cmp     eax, [esi+TAsmEdit._width_ch]
        jae     .do_renderX

        mov     [.render.right], eax
        jmp     .scrollitX

.scroll_left:
        mov     ebx, [esi+TAsmEdit._width_ch]
        add     ebx, eax
        dec     ebx
        mov     [.render.left], ebx

.scrollitX:
        neg     eax
        imul    eax, [esi+TAsmEdit._font_width]
        stdcall ScrollImageH, [esi+TAsmEdit._imgText], eax

.do_renderX:
        lea     eax, [.render]
        exec    esi, TAsmEdit:_RenderTxt, [esi+TAsmEdit._imgText], eax
        exec    esi, TAsmEdit:__DrawScrollbar, TAsmEdit._scrX
        exec    esi, TAsmEdit:UpdateCaretPos
        exec    esi, TAsmEdit:RectChanged2, 0

.finish:
        popad
        return


.setpos_Y:
        xchg    eax, [esi+TAsmEdit._scrY.pos]
        sub     eax, [esi+TAsmEdit._scrY.pos]        ; eax < 0 = scroll up; eax > 0 = scroll down
        jz      .finish

        mov     edx, eax
        jge     @f
        neg     edx
@@:
        cmp     edx, [esi+TAsmEdit._height_ch]
        jge     .do_renderY

        test    eax, eax
        js      .scroll_up

        cmp     eax, [esi+TAsmEdit._height_ch]
        jae     .do_renderY

        mov     [.render.bottom], eax
        jmp     .scrollitY

.scroll_up:
        mov     ebx, [esi+TAsmEdit._height_ch]
        add     ebx, eax
        dec     ebx
        mov     [.render.top], ebx

.scrollitY:
        neg     eax
        imul    eax, [esi+TAsmEdit._font_height]
        stdcall ScrollImageV, [esi+TAsmEdit._imgText], eax
        stdcall ScrollImageV, [esi+TAsmEdit._imgMargin], eax

.do_renderY:
        lea     eax, [.render]
        exec    esi, TAsmEdit:_RenderTxt, [esi+TAsmEdit._imgText], eax
        exec    esi, TAsmEdit:_RenderLM, [esi+TAsmEdit._imgMargin], [.render.top], [.render.bottom]

        exec    esi, TAsmEdit:__DrawScrollbar, TAsmEdit._scrY
        exec    esi, TAsmEdit:UpdateCaretPos
        exec    esi, TAsmEdit:RectChanged2, 0
        jmp     .finish
endp






method TAsmEdit.SetSyntaxProc
begin
        pushad
        mov     esi, [.self]
        mov     eax, [.value]
        xchg    eax, [esi+TAsmEdit._syntax_proc]
        cmp     eax, [.value]
        je      .finish

; if the syntax procedure were changed, then
; free all syntax arrays from the lines and
; invalidate the rendered images.

        mov     edx, [esi+TAsmEdit._lines]
        xor     ecx, ecx

.loop:
        mov     eax, ecx
        shl     eax, 2
        cmp     eax, [edx+TText.GapBegin]
        jb      @f

        add     eax, [edx+TText.GapEnd]
        sub     eax, [edx+TText.GapBegin]

@@:
        cmp     eax, [edx+TText.Length]
        jae     .refresh

        mov     eax, [edx+eax]
        stdcall FreeMem, [eax+TTextLine.syn_array]
        and     [eax+TTextLine.syn_array], 0

        inc     ecx
        jmp     .loop

.refresh:
        and     [esi+TAsmEdit._img_txt_valid], 0
        and     [esi+TAsmEdit._img_margin_valid], 0
        exec    esi, TAsmEdit:RectChanged2, 0

.finish:
        popad
        return
endp




method TAsmEdit.AddChar ;, .utf8
begin
        pushad

        mov     esi, [.self]
        mov     ebx, [esi+TAsmEdit._caretPos.y]
        exec    esi, TAsmEdit:TextLine, ebx
        mov     edi, eax

        exec    esi, TAsmEdit:SaveForUndo, ebx, chopChanged
        exec    esi, TAsmEdit:AddTrailingSpaces, ebx, [esi+TAsmEdit._caretPos.x]

        stdcall TextMoveGap, [edi], [esi+TAsmEdit._caretPos.x]
        stdcall TextAddChar, [edi], [.utf8]
        mov     [edi], edx

        push    [edx+TTextLine.ch_len]
        inc     [edx+TTextLine.ch_len]
        push    [edx+TTextLine.ch_len]

        stdcall IncCount, [esi+TAsmEdit._line_widths]   ; count from the stack!
        stdcall DecCount, edx                           ; count from the stack!
        mov     [esi+TAsmEdit._line_widths], edx

        exec    esi, TAsmEdit:DelTrailingSpaces, [esi+TAsmEdit._caretPos.y]
        exec    esi, TAsmEdit:CaretLeftRight, 1

        popad
        return
endp




method TAsmEdit.EventKeyPress
.selCmd dd ?
begin
        pushad

        mov     esi, [.self]

; check the keys

        mov     eax, [.utf8]

        test    eax, eax
        jz      .no_char

        cmp     eax, $20
        jb      .ctrl_char

        cmp     eax, $7f
        je      .no_char

; here handle usual typing!

        exec    esi, TAsmEdit:AddChar, eax

        exec    esi, TAsmEdit:ComputeLeftMarginWidth
        exec    esi, TAsmEdit:__UpdateImages
        exec    esi, TAsmEdit:UpdateScrollBar
        exec    esi, TAsmEdit:UpdateCaretPos

        xor     eax, eax
        mov     [esi+TAsmEdit._img_txt_valid], eax
        mov     [esi+TAsmEdit._img_margin_valid], eax
        exec    esi, TAsmEdit:RectChanged2, eax

        jmp     .finish

.ctrl_char:

        cmp     eax, 26         ; Ctrl+Z
        je      .do_undo

        cmp     eax, 3          ; Ctrl+C
        je      .copy_to_clipboard

        cmp     eax, 22         ; Ctrl+V
        je      .paste_from_clipboard

        cmp     eax, 2          ; Ctrl+B
        je      .select_begin

        cmp     eax, 5          ; Ctrl+E
        je      .select_end

        cmp     eax, 8          ; BackSpace
        je      .backspacekey

        cmp     eax, 9          ; tab
        je      .tabkey

        cmp     eax, 13         ; Enter
        je      .enterkey

        jmp     .finish


.copy_to_clipboard:

        stdcall ClipboardWrite, <"This is clipboard test.", 13, 10>
        jmp     .finish

.paste_from_clipboard:

        stdcall ClipboardRead
        jc      .timeout

        test    eax, eax
        jz      .bad_return_string

        push    eax

        stdcall FileWriteString, [STDOUT], eax
        stdcall StrDel ; from the stack

        stdcall FileWriteString, [STDOUT], <txt 13, 10>

        jmp     .finish

.bad_return_string:
        stdcall FileWriteString, [STDOUT], <txt "Error: Clipboard returns NULL!", 13, 10>
        jmp     .finish

.timeout:
        stdcall FileWriteString, [STDOUT], <txt "Error: Clipboard timeout!", 13, 10>
        jmp     .finish


.select_begin:

        exec    esi, TAsmEdit:SelStartHere
        exec    esi, TAsmEdit:SelEndHere
        jmp     .finish

.select_end:

        exec    esi, TAsmEdit:SelEndHere
        jmp     .finish

.do_undo:
        test    [.kbdState], maskShift
        jnz     .do_redo

        exec    esi, TAsmEdit:Undo
        jmp     .finish

.do_redo:
        exec    esi, TAsmEdit:Redo
        jmp     .finish


.no_char:
        mov     eax, [.scancode]

        cmp     eax, keyF5
        je      .refresh

; moving commands

; what is the selection command?
        mov     ecx, [.kbdState]
        and     ecx, maskShift
        xchg    ecx, [esi+TAsmEdit._selCmd]
        mov     [esi+TAsmEdit._selCmdPrev], ecx

        cmp     [esi+TAsmEdit._selCmd], 0
        je      .first_sel_ok

        exec    esi, TAsmEdit:SelChangeBegin

        cmp     [esi+TAsmEdit._selCmdPrev], 0
        jne     .first_sel_ok

        exec    esi, TAsmEdit:SelStartHere
        exec    esi, TAsmEdit:SelEndHere

.first_sel_ok:
        mov     eax, [.scancode]

        cmp     eax, keyLeftNumpad
        je      .left
        cmp     eax, keyLeft
        je      .left

        cmp     eax, keyRightNumpad
        je      .right
        cmp     eax, keyRight
        je      .right

        cmp     eax, keyUpNumpad
        je      .up
        cmp     eax, keyUp
        je      .up

        cmp     eax, keyDownNumpad
        je      .down
        cmp     eax, keyDown
        je      .down

        cmp     eax, keyHomeNumpad
        je      .homekey
        cmp     eax, keyHome
        je      .homekey

        cmp     eax, keyEndNumpad
        je      .endkey
        cmp     eax, keyEnd
        je      .endkey

        cmp     eax, keyPgUpNumpad
        je      .pgupkey
        cmp     eax, keyPgUp
        je      .pgupkey

        cmp     eax, keyPgDnNumpad
        je      .pgdnkey
        cmp     eax, keyPgDown
        je      .pgdnkey

; end of moving commands

        cmp     eax, keyInsNumpad
        je      .insertkey
        cmp     eax, keyInsert
        je      .insertkey

        cmp     eax, keyDelNumpad
        je      .delkey
        cmp     eax, keyDelete
        je      .delkey

        jmp     .endselect

.tabkey:
        mov     ecx, [esi+TAsmEdit._caretPos.x]
        inc     ecx
        add     ecx, 7
        and     ecx, $fffffff8
        sub     ecx, [esi+TAsmEdit._caretPos.x]

.spadd:
        exec    esi, TAsmEdit:AddChar, ' '
        loop    .spadd

        xor     eax, eax
        mov     [esi+TAsmEdit._img_txt_valid], eax
        mov     [esi+TAsmEdit._img_margin_valid], eax
        exec    esi, TAsmEdit:RectChanged2, eax
        jmp     .finish


.enterkey:
        test    [esi+TAsmEdit._options], aeoOverwriteMode
        jnz     .split_ok

        exec    esi, TAsmEdit:SplitLine, -1, 0

.split_ok:
        exec    esi, TAsmEdit:CaretUpDown, 1, FALSE
        exec    esi, TAsmEdit:CaretHome

        and     [esi+TAsmEdit._img_txt_valid], 0
        and     [esi+TAsmEdit._img_margin_valid], 0
        exec    esi, TAsmEdit:RectChanged2, 0

        exec    esi, TAsmEdit:ComputeLeftMarginWidth
        exec    esi, TAsmEdit:__UpdateImages
        exec    esi, TAsmEdit:UpdateScrollBar
        exec    esi, TAsmEdit:UpdateCaretPos
        jmp     .finish

.refresh:
        xor     eax, eax
        mov     [esi+TAsmEdit._img_txt_valid], eax
        mov     [esi+TAsmEdit._img_margin_valid], eax
        exec    esi, TAsmEdit:RectChanged2, eax
        jmp     .finish


; caret moving commands!

.left:
        exec    esi, TAsmEdit:CaretLeftRight, -1
        jmp     .endselect


.right:
        exec    esi, TAsmEdit:CaretLeftRight, 1
        jmp     .endselect

.up:
        exec    esi, TAsmEdit:CaretUpDown, -1, FALSE
        jmp     .endselect


.down:
        exec    esi, TAsmEdit:CaretUpDown, 1, FALSE
        jmp     .endselect

.homekey:
        test    [.kbdState], maskCtrl
        jnz     .start_of_scr

        exec    esi, TAsmEdit:CaretHome
        jmp     .endselect

.start_of_scr:
        exec    esi, TAsmEdit:CaretScrHome
        jmp     .endselect


.endkey:
        test    [.kbdState], maskCtrl
        jnz     .end_of_scr

        exec    esi, TAsmEdit:CaretEnd
        jmp     .endselect

.end_of_scr:
        exec    esi, TAsmEdit:CaretScrEnd
        jmp     .endselect



.pgupkey:
        test    [.kbdState], maskCtrl
        jnz     .start_of_text

        mov     eax, [esi+TAsmEdit._height]
        xor     edx, edx
        div     [esi+TAsmEdit._font_height]
        neg     eax

        exec    esi, TAsmEdit:CaretUpDown, eax, TRUE
        jmp     .endselect

.start_of_text:
        exec    esi, TAsmEdit:CaretTxtHome
        jmp     .endselect


.pgdnkey:
        test    [.kbdState], maskCtrl
        jnz     .end_of_text

        mov     eax, [esi+TAsmEdit._height]
        xor     edx, edx
        div     [esi+TAsmEdit._font_height]

        exec    esi, TAsmEdit:CaretUpDown, eax, TRUE
        jmp     .endselect

.end_of_text:
        exec    esi, TAsmEdit:CaretTxtEnd
        jmp     .endselect


.insertkey:
        get     eax, esi, TAsmEdit:Options

        mov     edx, aeoOverwriteMode
        mov     ecx, [.kbdState]
        and     ecx, maskCtrl + maskAlt
        cmp     ecx, maskCtrl + maskAlt
        jne     .ins_mask_ok

        mov     edx, aeoBlockSelection

.ins_mask_ok:
        xor     eax, edx
        set     esi, TAsmEdit:Options, eax
        jmp     .endselect

.backspacekey:
        test    [esi+TAsmEdit._options], aeoNavXCaretWrap
        jnz     .do_back

        cmp     [esi+TAsmEdit._caretPos.x], 0
        je      .finish

.do_back:
        exec    esi, TAsmEdit:AddTrailingSpaces, -1, 0
        exec    esi, TAsmEdit:CaretLeftRight, -1

.delkey:
        exec    esi, TAsmEdit:DeleteSelection
        jnc     .del_refresh

        exec    esi, TAsmEdit:AddTrailingSpaces, -1, 0

        exec    esi, TAsmEdit:DeleteChar, -1, 0
        jnc     .del_refresh                            ; notice that the undo/redo group is not increased in order to allow
                                                        ; another change operations on this line to be grouped in one TUndoInfo
                                                        ; structure.

        test    [esi+TAsmEdit._options], aeoOverwriteMode
        jnz     .finish

        exec    esi, TAsmEdit:JoinNextLine, -1
        inc     [AsmEditUndoGroup]

.del_refresh:
        exec    esi, TAsmEdit:DelTrailingSpaces, -1

        and     [esi+TAsmEdit._img_txt_valid], 0
        and     [esi+TAsmEdit._img_margin_valid], 0

        exec    esi, TAsmEdit:RectChanged2, 0
        exec    esi, TAsmEdit:UpdateCaretPos

        exec    esi, TAsmEdit:SelChangeBegin
        exec    esi, TAsmEdit:SelStartHere
        exec    esi, TAsmEdit:SelEndHere
        exec    esi, TAsmEdit:SelChangeEnd

        jmp     .finish

.endselect:
        cmp     [esi+TAsmEdit._selCmd], 0
        je      .finish

        exec    esi, TAsmEdit:SelEndHere
        exec    esi, TAsmEdit:SelChangeEnd

.finish:
        popad
        return
endp




method TAsmEdit.EventFocusIn
begin
        stdcall CaretAttach, [.self]
        exec    [.self], TAsmEdit:UpdateCaretPos
        return
endp




method TAsmEdit.EventFocusOut
begin
        stdcall CaretAttach, 0
        return
endp



method TAsmEdit.UpdateCaretPos
.zero dd ?
begin
        pushad

        mov     esi, [.self]

        xor     edi, edi
        mov     [.zero], edi
        inc     edi

        mov     eax, [esi+TAsmEdit._caretPos.x]
        mov     ebx, [esi+TAsmEdit._caretPos.y]

        get     edx, esi, TAsmEdit:PosX
        sub     eax, edx
        cmovs   edi, [.zero]    ; hide

        imul    eax, [esi+TAsmEdit._font_width]
        add     eax, [esi+TAsmEdit._left_margin]

        cmp     eax, [esi+TAsmEdit._width]
        cmovae  edi, [.zero]

        get     edx, esi, TAsmEdit:PosY
        sub     ebx, edx
        cmovs   edi, [.zero]

        mov     edx, 2
        mov     ecx, 2
        test    [esi+TAsmEdit._options], aeoOverwriteMode
        cmovnz  ecx, [esi+TAsmEdit._font_width]
        cmovnz  edx, [esi+TAsmEdit._font_height]

        test    [esi+TAsmEdit._options], aeoConsoleCaret
        cmovnz  ecx, [esi+TAsmEdit._font_width]
        cmovz   edx, [esi+TAsmEdit._font_height]

        imul    ebx, [esi+TAsmEdit._font_height]
        add     ebx, [esi+TAsmEdit._font_height]

        cmp     ebx, [esi+TAsmEdit._height]
        cmovae  edi, [.zero]

        sub     ebx, edx

        stdcall CaretChange, eax, ebx, ecx, edx, edi

.finish:
        popad
        return

.hide_caret:
        mov     [Caret.visible], 0
        jmp     .finish
endp





method TAsmEdit._RenderTxt ; .pImage, .pRect
.rectr RECT  ; relative to the upper left corner of the window
.recta RECT  ; absolute in the document coordinates.
.rects RECT
.context dd ?
begin
        pushad

;if defined options.DebugMode & options.DebugMode
;        mov     eax, [.pRect]
;        mov     ecx, [eax+RECT.right]
;        mov     edx, [eax+RECT.bottom]
;        sub     ecx, [eax+RECT.left]
;        sub     edx, [eax+RECT.top]
;        OutputRect [eax+RECT.left], [eax+RECT.top], ecx, edx
;end if

        mov     esi, [.self]

        xor     eax, eax
        mov     ecx, [esi+TAsmEdit._width_ch]
        mov     edx, [esi+TAsmEdit._height_ch]
        mov     [.rectr.left], eax
        mov     [.rectr.top], eax
        mov     [.rectr.right], ecx
        mov     [.rectr.bottom], edx

        mov     [.recta.left], eax
        mov     [.recta.top], eax

        mov     edi, [esi+TAsmEdit._lines]
        mov     ecx, [esi+TAsmEdit._line_widths]

        mov     eax, [edi+TText.Length]
        add     eax, [edi+TText.GapBegin]
        sub     eax, [edi+TText.GapEnd]         ; the height of the document in lines x4

;        mov     edx, [ecx+TArray.count]
;        test    edx, edx
;        cmovnz  edx, [ecx + TArray.array + sizeof.TCounter * (edx-1) + TCounter.value]  ; the maximal width of the document.   ??????????????????????????
        mov     edx, $7fffffff

        shr     eax, 2
        add     eax, 1000

        mov     [.recta.right], edx
        mov     [.recta.bottom], eax

        lea     edx, [.rectr]
; Not needed in order to allow drawing of the empty areas! Need some code cleanup around.
        stdcall RectIntersect2, edx, [.pRect]
        jc      .end_draw                               ; nothing to draw

;if defined options.DebugMode & options.DebugMode
;        mov     eax, [edx+RECT.right]
;        mov     ecx, [edx+RECT.bottom]
;        sub     eax, [edx+RECT.left]
;        sub     ecx, [edx+RECT.top]
;        OutputRect [edx+RECT.left], [edx+RECT.top], eax, ecx
;end if

        get     eax, esi, TAsmEdit:PosX
        get     ecx, esi, TAsmEdit:PosY

        add     [.rectr.left], eax
        add     [.rectr.top], ecx
        add     [.rectr.right], eax
        add     [.rectr.bottom], ecx

        lea     ebx, [.recta]
        stdcall RectIntersect2, ebx, edx
        jc      .end_draw                               ; nothing to draw.

        stdcall RectCopy, edx, ebx

        sub     [.rectr.left], eax
        sub     [.rectr.top], ecx
        sub     [.rectr.right], eax
        sub     [.rectr.bottom], ecx

; Here [.recta] contains the final rectangle that need to be drawn in absolute coordinates.
;      [.rectr] contains the final rectangle that need to be drawn in coordinates, relative to the upper left corner of the window.

        mov     eax, [.rectr.left]
        mov     ecx, [.rectr.right]

        imul    eax, [esi+TAsmEdit._font_width]
        imul    ecx, [esi+TAsmEdit._font_width]

        mov     [.rectr.left], eax
        mov     [.rectr.right], ecx

        mov     eax, [.rectr.top]
        mov     ecx, [.rectr.bottom]

        imul    eax, [esi+TAsmEdit._font_height]
        imul    ecx, [esi+TAsmEdit._font_height]

        add     eax, [esi+TAsmEdit._font_asc]
        add     ecx, [esi+TAsmEdit._font_asc]

        mov     [.rectr.top], eax
        mov     [.rectr.bottom], ecx

; Here: [.rectr] contains the pixel coordinates of the rectangle that have to be drawn.
;       edi contains pointer to the TText structure of the lines.

        mov     eax, [.rectr.left]
        sub     [.rectr.right], eax     ; .rectr.right == width

        stdcall LockImg, [.pImage]

        mov     ebx, [.recta.top]

        xor     edx, edx
        lea     eax, [4*ebx]                  ; the previous line for the syntax context
        sub     eax, 4
        js      .context_ok

        cmp     eax, [edi+TText.GapBegin]
        jb      @f
        add     eax, [edi+TText.GapEnd]
        sub     eax, [edi+TText.GapBegin]
@@:
        cmp     eax, [edi+TText.Length]
        jae     .context_ok

        mov     edx, [edi+eax]
        mov     edx, [edx+TTextLine.syn_context]

.context_ok:
        mov     [.context], edx


.line_loop:
        mov     edx, [.rectr.top]
        sub     edx, [esi+TAsmEdit._font_asc]   ; needed for the background and the left margin.

        xor     ecx, ecx
        test    [esi+TAsmEdit._options], aeoStripedBackground
        cmovnz  ecx, ebx
        and     ecx, 1

        stdcall DrawSolidRect, [.pImage], [.rectr.left], edx, [.rectr.right], [esi+TAsmEdit._font_height], [ASM.clBackground + 4*ecx]

        lea     eax, [4*ebx]
        cmp     eax, [edi+TText.GapBegin]
        jb      @f
        add     eax, [edi+TText.GapEnd]
        sub     eax, [edi+TText.GapBegin]
@@:
        cmp     eax, [edi+TText.Length]
        jae     .text_ok

        mov     edx, [edi+eax]
        cmp     [edx+TTextLine.syn_array], 0
        jne     .syn_ok

        cmp     [esi+TAsmEdit._syntax_proc], 0
        je      .syn_ok

        stdcall [esi+TAsmEdit._syntax_proc], edx, [.context]

.syn_ok:

        stdcall __RenderOneLine, [.pImage], edx, [.recta.left], [.recta.right], [.rectr.left], [.rectr.top]

        mov     eax, [edx+TTextLine.syn_context]
        mov     [.context], eax

.text_ok:

        mov     eax, [esi+TAsmEdit._font_height]
        add     [.rectr.top], eax

        inc     ebx
        cmp     ebx, [.recta.bottom]
        jb      .line_loop

        stdcall UnlockImg, [.pImage]

.end_draw:
        popad
        return
endp



method TAsmEdit._RenderLM ;  .pImage, .from, .to
.last dd ?
begin
        pushad

;        OutputValue "LM from:", [.from], 10, -1
;        OutputValue "LM to:", [.to], 10, -1

        mov     esi, [.self]

        mov     eax, [esi+TAsmEdit._height_ch]
        xor     ebx, ebx
        mov     ecx, [.from]
        mov     edx, [.to]

        cmp     ecx, ebx        ; .from should be >=0
        cmovl   ecx, ebx

        cmp     edx, ebx
        cmovl   edx, ebx

        cmp     edx, eax
        cmova   edx, eax        ; edx = max(PosY+height_ch, [.to])

        cmp     ecx, edx
        jae     .end_lines

        mov     [.from], ecx
        mov     [.to], edx

        get     ebx, esi, TAsmEdit:PosY
        lea     ebx, [ebx+ecx+1]                ; line numbers are from 1

        mov     eax, [esi+TAsmEdit._font_height]
        imul    ecx, eax
        imul    edx, eax

        push    edx

        mov     edx, [esi+TAsmEdit._lines]
        mov     eax, [edx+TText.Length]
        add     eax, [edx+TText.GapBegin]
        sub     eax, [edx+TText.GapEnd]
        shr     eax, 2
        mov     [.last], eax

        pop     edx

.lines_loop:
        cmp     ecx, edx
        jae     .end_lines

; background

        mov     eax, [esi+TAsmEdit._left_margin]
        dec     eax
        stdcall DrawSolidRect, [.pImage], 0, ecx, eax, [esi+TAsmEdit._font_height], [ASM.clLeftMargin]
        stdcall DrawSolidRect, [.pImage], eax, ecx, 1, [esi+TAsmEdit._font_height], [ASM.clLeftMarginEdge]

; the number

        cmp     ebx, [.last]
        ja      .next

        pushad

        add     ecx, [esi+TAsmEdit._font_asc]

        stdcall NumToStr, ebx, ntsDec or ntsUnsigned
        push    eax

        stdcall StrPtr, eax
        push    eax

        stdcall GetTextBounds, eax, [eax+string.len], [ASM.NumFont]
        mov     edx, [esi+TAsmEdit._num_margin]
        sub     edx, eax

        pop     eax
        stdcall DrawString, [.pImage], eax, [eax+string.len], edx, ecx, [ASM.NumFont], [ASM.clLeftMarginText]

        stdcall StrDel ; from the stack

        popad

.next:
        inc     ebx
        add     ecx, [esi+TAsmEdit._font_height]
        jmp     .lines_loop


.end_lines:
        popad
        return
endp




proc __RenderOneLine, .pImage, .pLineInfo, .fromChar, .toChar, .x, .y

.chunk    dd ?  ; the index of the current chunk.
.chlen    dd ?  ; the remaining length of the current chunk.
.color    dd ?  ; the color of the current chunk.

.gap_start dd ?
.gap_end   dd ?
.txt_end   dd ?

begin
        pushad

        mov     esi, [.pLineInfo]       ; Should not be called with invalid [.pLineInfo]
        mov     edi, [esi+TTextLine.syn_array]

; search for the [.fromChar] element.

        xor     ebx, ebx                ; position inside the TTextChunk array.
        xor     ecx, ecx                ; position inside the text
        mov     edx, [.fromChar]        ; counter

        sub     [.toChar], edx          ; [.toChar] will contains the length
        jle     .finish                 ; nothing to render.

.loop_from:
        dec     edx
        js      .found

        mov     eax, [esi+TTextLine.GapBegin]
        cmp     ecx, eax
        cmovae  eax, [esi+TTextLine.GapEnd]
        sub     eax, [esi+TTextLine.GapBegin]
        add     eax, ecx

        cmp     eax, [esi+TTextLine.Length]
        jae     .finish                         ; the .fromChar is bigger than the length of the text - do nothing.

        movzx   eax, byte [esi+eax]

                                ; compute the length of the UTF-8 character by its first byte (version without branches) Is it fast?
        not     al              ; utf-8 starts with one of: 0xxxxxxx, 110xxxxx, 1110xxxx, 11110xxx, 111110xx, 1111110x
        bsr     eax, eax        ; eax =                 7,  5,  4,  3,  2,  1
        sub     eax, 7          ; eax =                 0, -2, -3, -4, -5, -6
                                ; CF  =                 0,  1,  1,  1,  1,  1
        not     eax             ; eax =                -1, +1, +2, +3, +4, +5
        sbb     eax, -2         ; eax = edx + 2 - CF = +1, +2, +3, +4, +5, +6

        add     ecx, eax
        jmp     .loop_from


.found:

; here ecx contains the position in the text of [.fromChar]

        mov     eax, [esi+TTextLine.GapBegin]
        add     eax, esi
        mov     [.gap_start], eax

        mov     eax, [esi+TTextLine.GapEnd]
        add     eax, esi
        mov     [.gap_end], eax

        mov     eax, [esi+TTextLine.Length]
        add     eax, esi
        mov     [.txt_end], eax

        mov     eax, [esi+TTextLine.GapBegin]
        cmp     ecx, eax
        cmovae  eax, [esi+TTextLine.GapEnd]
        sub     eax, [esi+TTextLine.GapBegin]
        add     eax, ecx

        add     esi, eax         ; now esi - the start of the text that have to be drawn

; now loop to the starting TTextChunk

        mov     edx, [.fromChar]
        xor     ebx, ebx
        dec     ebx

        test    edi, edi
        jz      .default_color

.chunk_loop:
        inc     ebx
        cmp     ebx, [edi+TArray.count]
        jae     .default_color

        movzx   eax, word [edi + TArray.array + sizeof.TTextChunk * ebx + TTextChunk.len]
        sub     edx, eax
        jge     .chunk_loop

        neg     edx     ; the remaining characters of this chunk

.set_color:
; here ebx - the index of the current chunk.
;      edx - the remaining characters from this chunk.

        mov     [.chunk], ebx
        mov     [.chlen], edx

        movzx   eax, byte [edi+ TArray.array + sizeof.TTextChunk * ebx + TTextChunk.color]
        and     eax, $0f
        mov     eax, [ASM.clText + 4*eax]
        mov     [.color], eax

        jmp     .draw_chunk_loop


.default_color:

        and     [.chunk], 0
        or      [.chlen], -1
        mov     eax, [ASM.clText]
        mov     [.color], eax

; and now, scan for chunks and draw them

.draw_chunk_loop:

        mov     ecx, esi              ; the start of the chunk text

.inner_chunk_loop:
        sub     [.chlen], 1
        jc      .next_chunk           ; CF=1 only once.

        dec     [.toChar]
        js      .end_chunk_final      ; draw the final chunk and exits.

        cmp     esi, [.gap_end]
        jae     .ofs_ok

        cmp     esi, [.gap_start]
        jb      .ofs_ok

        call    .draw_current_chunk

        mov     esi, [.gap_end]
        mov     ecx, esi

.ofs_ok:
        cmp     esi, [.txt_end]
        jae     .end_chunk_final

; extract one character and compute the length.

        movzx   eax, byte [esi]

; compute the length of the UTF-8 character by its first byte. See above for explanation.

        not     al
        bsr     eax, eax
        sub     eax, 7
        not     eax
        sbb     eax, -2

        add     esi, eax
        jmp     .inner_chunk_loop


.next_chunk:
        call    .draw_current_chunk

        mov     ebx, [.chunk]
        inc     ebx
        cmp     ebx, [edi+TArray.count]
        ja      .default_color

        movzx   edx, word [edi+ TArray.array + sizeof.TTextChunk * ebx + TTextChunk.len]
        jmp     .set_color

.end_chunk_final:
        call    .draw_current_chunk


.finish:
        popad
        return



.draw_current_chunk:

        mov     eax, esi
        sub     eax, ecx
        jz      .draw_ok        ; zero length string

        stdcall DrawString, [.pImage], ecx, eax, [.x], [.y], [ASM.Font], [.color]
        add     [.x], eax

.draw_ok:
        retn

endp


method TAsmEdit.SelfPaint       ;.pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height    ; Paints itself on the caller provided TImage.

.rect_src RECT
.rect_lm  RECT
.rect_txt RECT

.rect_ch  RECT
.rect_dst RECT

.offs_x dd ?
.offs_y dd ?

begin
        pushad

        mov     esi, [.self]

        mov     ecx, [.xDst]
        mov     edx, [.yDst]
        sub     ecx, [.xSrc]
        sub     edx, [.ySrc]

        mov     [.offs_x], ecx
        mov     [.offs_y], edx

        mov     ecx, [.xSrc]
        mov     edx, [.ySrc]
        mov     [.rect_src.left], ecx
        mov     [.rect_src.top], edx
        add     ecx, [.width]
        add     edx, [.height]
        mov     [.rect_src.right], ecx
        mov     [.rect_src.bottom], edx

        xor     eax, eax
        mov     ebx, [esi+TAsmEdit._left_margin]
        mov     ecx, [esi+TAsmEdit._width]
        mov     edx, [esi+TAsmEdit._height]

        mov     [.rect_lm.left], eax
        mov     [.rect_lm.top], eax
        mov     [.rect_lm.right], ebx
        mov     [.rect_lm.bottom], edx

        mov     [.rect_txt.left], ebx
        mov     [.rect_txt.top], eax
        mov     [.rect_txt.right], ecx
        mov     [.rect_txt.bottom], edx

        lea     eax, [.rect_lm]
        lea     edx, [.rect_src]
        stdcall RectIntersect2, eax, edx
        jc      .left_margin_ok

        cmp     [esi+TAsmEdit._img_margin_valid], 0
        jne     .img_lm_ok

        mov     eax, [.rect_lm.top]
        xor     edx, edx
        div     [esi+TAsmEdit._font_height]
        mov     ecx, eax

        mov     eax, [.rect_lm.bottom]
        xor     edx, edx
        div     [esi+TAsmEdit._font_height]
        test    edx, edx
        jz      @f
        inc     eax
@@:
        exec    esi, TAsmEdit:_RenderLM, [esi+TAsmEdit._imgMargin], ecx, eax
        inc     [esi+TAsmEdit._img_margin_valid]

.img_lm_ok:
        mov     ecx, [.rect_lm.left]
        mov     edx, [.rect_lm.top]
        sub     [.rect_lm.right], ecx
        sub     [.rect_lm.bottom], edx

        add     ecx, [.offs_x]
        add     edx, [.offs_y]

        stdcall BlendImage, [.pDstImage], ecx, edx, [esi+TAsmEdit._imgMargin], [.rect_lm.left], [.rect_lm.top], [.rect_lm.right], [.rect_txt.bottom]

.left_margin_ok:

        lea     eax, [.rect_txt]
        lea     edx, [.rect_src]
        stdcall RectIntersect2, eax, edx
        jc      .text_ok

        cmp     [esi+TAsmEdit._img_txt_valid], 0
        jne     .img_txt_ok


        mov     eax, [.rect_txt.left]
        sub     eax, [esi+TAsmEdit._left_margin]
        jl      .img_txt_ok

        xor     edx, edx
        mov     ecx, [esi+TAsmEdit._font_width]

        div     ecx
        mov     [.rect_ch.left], eax

        mov     eax, [.rect_txt.right]
        sub     eax, [esi+TAsmEdit._left_margin]
        jl      .img_txt_ok

        xor     edx, edx
        div     ecx
        test    edx, edx
        jz      @f
        inc     eax
@@:
        mov     [.rect_ch.right], eax

        mov     eax, [.rect_txt.top]
        mov     ecx, [esi+TAsmEdit._font_height]
        xor     edx, edx

        div     ecx

        mov     [.rect_ch.top], eax

        mov     eax, [.rect_txt.bottom]
        xor     edx, edx
        div     ecx
        test    edx, edx
        jz      @f
        inc     eax
@@:
        mov     [.rect_ch.bottom], eax

        lea     eax, [.rect_ch]
        exec    esi, TAsmEdit:_RenderTxt, [esi+TAsmEdit._imgText], eax
        inc     [esi+TAsmEdit._img_txt_valid]

.img_txt_ok:

        mov     ecx, [.rect_txt.left]
        mov     edx, [.rect_txt.top]
        mov     ebx, [.rect_txt.right]
        mov     edi, [.rect_txt.bottom]

        sub     ebx, ecx        ; width
        sub     edi, edx        ; height

        mov     eax, ecx
        add     ecx, [.offs_x]
        add     edx, [.offs_y]
        sub     eax, [esi+TAsmEdit._left_margin]

        stdcall BlendImage, [.pDstImage], ecx, edx, [esi+TAsmEdit._imgText], eax, [.rect_txt.top], ebx, edi

.text_ok:

; Here, draw the selection if any.

        exec    esi, TAsmEdit:_AddSelRectangles, 0
        push    edx

        mov     ecx, [edx+TArray.count]
        test    ecx, ecx
        jz      .selection_ok

        mov     eax, [.xSrc]
        mov     ebx, [.ySrc]
        mov     [.rect_dst.left], eax
        mov     [.rect_dst.top], ebx

        add     eax, [.width]
        add     ebx, [.height]
        mov     [.rect_dst.right], eax
        mov     [.rect_dst.bottom], ebx

        lea     edi, [edx+TArray.array]
        lea     ebx, [.rect_dst]

.loop:
        mov     eax, [edi+RECT.left]
        mov     edx, [edi+RECT.top]

        imul    eax, [esi+TAsmEdit._font_width]
        imul    edx, [esi+TAsmEdit._font_height]

        add     eax, [esi+TAsmEdit._left_margin]

        mov     [edi+RECT.left], eax
        mov     [edi+RECT.top], edx

        mov     eax, [edi+RECT.right]
        mov     edx, [edi+RECT.bottom]

        imul    eax, [esi+TAsmEdit._font_width]
        imul    edx, [esi+TAsmEdit._font_height]

        add     eax, [esi+TAsmEdit._left_margin]

        mov     [edi+RECT.right], eax
        mov     [edi+RECT.bottom], edx

        stdcall RectIntersect2, edi, ebx
        jc      .next

        push    [ASM.clSelection]

        mov     edx, [edi+RECT.bottom]
        mov     eax, [edi+RECT.right]
        sub     edx, [edi+RECT.top]
        sub     eax, [edi+RECT.left]

        push    edx     ; height
        push    eax     ; width

        mov     eax, [edi+RECT.left]
        mov     edx, [edi+RECT.top]
        add     eax, [.offs_x]
        add     edx, [.offs_y]

        stdcall BlendSolidRect, [.pDstImage], eax, edx

.next:
        add     edi, sizeof.RECT
        dec     ecx
        jnz     .loop

.selection_ok:
        stdcall FreeMem ; from the stack

        popad
        inherited       [.pDstImage], [.xDst], [.yDst], [.xSrc], [.ySrc], [.width], [.height]
        return
endp






method TAsmEdit.UpdateScrollBar
begin
        pushad
        mov     esi, [.self]

        xor     eax, eax
        cmp     [esi+TAsmEdit._font_width], eax
        lea     eax, [eax+1]
        je      .xpage_ok

        mov     eax, [esi+TAsmEdit._width]

        sub     eax, [esi+TAsmEdit._left_margin]
        sub     eax, [GUI.scrollWidth]
        cdq
        idiv    [esi+TAsmEdit._font_width]
        cmp     eax, 0
        jg      .xpage_ok

        xor     eax, eax
        inc     eax

.xpage_ok:

        mov     [esi+TAsmEdit._scrX.page], eax

        mov     ecx, [esi+TAsmEdit._line_widths]
        mov     edx, [ecx+TArray.count]
        test    edx, edx
        jz      .xsize_ok
        mov     edx, [ecx+TArray.array+sizeof.TCounter*edx+TCounter.value - sizeof.TCounter]
        sub     edx, eax
        jge     .xsize_ok
        xor     edx, edx
.xsize_ok:
        mov     [esi+TAsmEdit._scrX.max], edx

; set Y scrollers page and

        xor     eax, eax
        cmp     [esi+TAsmEdit._font_height], eax
        lea     eax, [eax+1]
        je      .ypage_ok

        mov     eax, [esi+TAsmEdit._height]
        sub     eax, [GUI.scrollWidth]
        cdq
        idiv    [esi+TAsmEdit._font_height]
        test    eax, eax
        jnz     .ypage_ok

        inc     eax

.ypage_ok:

        mov     [esi+TAsmEdit._scrY.page], eax

        mov     edx, [esi+TAsmEdit._lines]
        mov     ecx, [edx+TText.Length]
        add     ecx, [edx+TText.GapBegin]
        sub     ecx, [edx+TText.GapEnd]
        shr     ecx, 2
        sub     ecx, eax
        jg      .ymax_ok
        xor     ecx, ecx
.ymax_ok:
        mov     [esi+TAsmEdit._scrY.max], ecx

        popad
        return
endp



; general methods


method TAsmEdit.Create
begin
        inherited [.parent]

        pushad

        mov     edi, [.self]

        mov     [edi+TAsmEdit._cursor], mcArrow;mcText
        mov     [edi+TAsmEdit.__want_focus], TRUE

        mov     [edi+TAsmEdit._options], aeoDefault

        stdcall GetFontMetrics, [ASM.Font]

        mov     [edi+TAsmEdit._font_height], eax
        mov     [edi+TAsmEdit._font_asc], ebx
        mov     [edi+TAsmEdit._font_desc], edx

        OutputValue "   Font height:", eax, 10, -1
        OutputValue " Font ascender:", ebx, 10, -1
        OutputValue "Font descender:", edx, 10, -1

        stdcall GetTextBounds, "ABCDEFGH", 8, [ASM.Font]
        sar     eax, 3
        mov     [edi+TAsmEdit._font_width], eax

        OutputValue "    Font width:", eax, 10, -1

        stdcall GetTextBounds, "99999999", 8, [ASM.NumFont]
        sar     eax, 3
        mov     [edi+TAsmEdit._numfont_width], eax

;        mov     [edi+TAsmEdit._syntax_proc], SyntaxFASM
; test only!

        set     edi, TAsmEdit:Text, 0
;        set     edi, TAsmEdit:Text, utf8_long    ; 0   ; set the needed text structures, but empty.
;        mov     [edi+TAsmEdit._caret_x], 4
;        mov     [edi+TAsmEdit._caret_y], 4

        popad
        return
endp




method TAsmEdit.Destroy
begin
        push    ecx
        mov     ecx, [.self]

        stdcall TextFree, [ecx+TAsmEdit._lines]

        pop     ecx
        inherited
        return
endp



method TAsmEdit.__UpdateImages
begin
        inherited
        pushad
        mov     esi, [.self]

; Handle ._imgText

        mov     edi, [esi+TAsmEdit._imgText]
        mov     ecx, [esi+TAsmEdit._width]
        mov     edx, [esi+TAsmEdit._height]

        sub     ecx, [esi+TAsmEdit._left_margin]
        mov     eax, [esi+TAsmEdit._img_txt_valid]

        call    .process_one_img

        mov     [esi+TAsmEdit._imgText], edi
        mov     [esi+TAsmEdit._img_txt_valid], eax

; Process ._imgLeftMargin

        mov     edi, [esi+TAsmEdit._imgMargin]
        mov     ecx, [esi+TAsmEdit._left_margin]
        mov     eax, [esi+TAsmEdit._img_margin_valid]

        call    .process_one_img

        mov     [esi+TAsmEdit._imgMargin], edi
        mov     [esi+TAsmEdit._img_margin_valid], eax

        exec    esi, TAsmEdit:RectChanged2, 0

        popad
        return

; Changes the size of one image:
; Arguments:
;   edi - pointer to the image
;   ecx - desired width
;   edx - desired height
;   eax - image_valid flag.
; Returns:
;   edi - pointer to the image. Can be changed.
;   eax - valid image flag. If the image has been reallocated, eax = FALSE; If the image was only resized, but the previous pixels
;         are preserved, the flag is not changed.
;
.process_one_img:

        test    edi, edi
        jz      .create_new

        cmp     ecx, [edi+TImage.width]
        ja      .destroy

        cmp     edx, [edi+TImage.height]
        jbe     .set_wrap

.destroy:

        stdcall DestroyImage, edi

.create_new:

        lea     eax, [ecx+SCREEN_MORE_X]
        lea     ebx, [edx+SCREEN_MORE_Y]

        stdcall CreateImage, eax, ebx
        mov     edi, eax
        xor     eax, eax

.set_wrap:
        stdcall SetImageWrap, edi, ecx, edx
        retn

endp


method TAsmEdit.Resize
.rect     RECT
.rect_ch  RECT
begin
        pushad

        mov     esi, [.self]
        mov     ebx, [esi+TAsmEdit._width]      ; the old values
        mov     ecx, [esi+TAsmEdit._height]     ; of the height and width
        mov     edi, [esi+TAsmEdit._width_ch]   ;
        mov     edx, [esi+TAsmEdit._height_ch]

        inherited [.newWidth], [.newHeight]
        exec    [.self], TAsmEdit:ComputeScreenInChars

        cmp     [esi+TAsmEdit._width], ebx
        jbe     .render_right_ok

        dec     edi
        jns     @f
        xor     edi, edi
@@:

        mov     [.rect.left], ebx
        mov     [.rect_ch.left], edi

        mov     ebx, [esi+TAsmEdit._width]
        mov     edi, [esi+TAsmEdit._width_ch]
        mov     [.rect.right], ebx
        mov     [.rect_ch.right], edi

        xor     eax, eax
        mov     [.rect.top], eax
        mov     [.rect_ch.top], eax

        mov     eax, [esi+TAsmEdit._height]
        mov     ebx, [esi+TAsmEdit._height_ch]
        mov     [.rect.bottom], eax
        mov     [.rect_ch.bottom], edx

        lea     eax, [.rect_ch]
        exec    esi, TAsmEdit:_RenderTxt, [esi+TAsmEdit._imgText], eax

        lea     eax, [.rect]
        exec    esi, TAsmEdit:RectChanged2, eax

.render_right_ok:

        cmp     [esi+TAsmEdit._height], ecx
        jbe     .render_bottom_ok

        dec     edx
        jns     @f
        xor     edx, edx
@@:

        mov     [.rect.top], ecx
        mov     [.rect_ch.top], edx

        mov     ecx, [esi+TAsmEdit._height]
        mov     edx, [esi+TAsmEdit._height_ch]
        mov     [.rect.bottom], ecx
        mov     [.rect_ch.bottom], edx

        xor     eax, eax
        mov     [.rect.left], eax
        mov     [.rect_ch.left], eax

        mov     eax, [esi+TAsmEdit._width]
        mov     edx, [esi+TAsmEdit._width_ch]
        mov     [.rect.right], eax
        mov     [.rect_ch.right], edx

        lea     eax, [.rect_ch]
        exec    esi, TAsmEdit:_RenderTxt, [esi+TAsmEdit._imgText], eax
        exec    esi, TAsmEdit:_RenderLM, [esi+TAsmEdit._imgMargin], [.rect_ch.top], [.rect_ch.bottom]

        lea     eax, [.rect]
        exec    esi, TAsmEdit:RectChanged2, eax

.render_bottom_ok:

        exec    [.self], TAsmEdit:UpdateScrollBar
;        exec    [.self], TAsmEdit:RectChanged2, 0

        popad
        return
endp



method TAsmEdit.Move
begin
        mov     eax, [Caret.pWindow]
        push    eax
        cmp     eax, [.self]
        jne     @f
        stdcall CaretAttach, 0
@@:
        inherited [.newX], [.newY]

        pop     eax
        cmp     eax, [.self]
        jne     @f
        stdcall CaretAttach, eax
        exec    [.self], TAsmEdit:UpdateCaretPos
@@:
        return
endp


; parameters get/set methods

method TAsmEdit.SetOptions
begin
        pushad

        mov     esi, [.self]
        mov     eax, [.value]

        cmp     eax, [esi+TAsmEdit._options]
        je      .finish

        mov     [esi+TAsmEdit._options], eax
        xor     eax, eax
        mov     [esi+TAsmEdit._img_txt_valid], eax
        mov     [esi+TAsmEdit._img_margin_valid], eax

        exec    esi, TAsmEdit:RectChanged2, eax
        exec    esi, TAsmEdit:ComputeLeftMarginWidth
        exec    esi, TAsmEdit:UpdateCaretPos

.finish:
        popad
        return
endp




method TAsmEdit.GetSelection
.lines dd ?
begin
        pushad

        exec    [.self], TAsmEdit:_AddSelRectanglesAbs, 0
        mov     [.lines], edx

        lea     edi, [edx+TArray.array]
        mov     ecx, [edx+TArray.count]

        stdcall StrNew
        mov     edx, eax

.rect_loop:
        dec     ecx
        js      .finish

        mov     ebx, [edi+RECT.bottom]

.y_loop:
        dec     ebx
        cmp     ebx, [edi+RECT.top]
        jl      .next_rect

        exec    [.self], TAsmEdit:TextLine, ebx
        mov     esi, [eax]

        mov     eax, [esi+TTextLine.ch_len]
        cmp     eax, [edi+RECT.right]
        cmova   eax, [edi+RECT.right]

        stdcall TextIndexToPos, esi, eax   ; move the gap after the rectangle right bound.
        push    eax

        stdcall TextMoveGap, esi, eax
        stdcall TextIndexToPos, esi, [edi+RECT.left]

        add     esi, eax
        sub     [esp], eax
        stdcall StrCatMem, edx, esi ; length from the stack

        stdcall StrCharCat, edx, $0d0a
        jmp     .y_loop

.next_rect:
        add     edi, sizeof.RECT
        jmp     .rect_loop

.finish:
        stdcall FreeMem, [.lines]

        mov     [esp+4*regEAX], edx     ; the result string. StrDel when not needed.
        popad
        return
endp




method TAsmEdit.SetSelection
begin

        return
endp




; returns two numbers:
;   eax - the whole left margin width
;   edx - the number part of the left margin.


method  TAsmEdit.ComputeLeftMarginWidth
begin
        pushad

        mov     esi, [.self]
        xor     ebx, ebx

        mov     [esi+TAsmEdit._num_margin], ebx

        test    [esi+TAsmEdit._options], aeoLeftMargin
        jz      .marginok

        test    [esi+TAsmEdit._options], aeoLineNumbers
        jz      .marginok1

        mov     ecx, [esi+TAsmEdit._lines]
        mov     eax, [ecx+TText.Length]
        add     eax, [ecx+TText.GapBegin]
        sub     eax, [ecx+TText.GapEnd]
        shr     eax, 2                          ; the number of the lines.
        test    eax, eax
        jz      .marginok1

        call    .digits
        imul    ecx, [esi+TAsmEdit._numfont_width]
        add     ebx, ecx
        add     ebx, [ASM.xspcLMPadding]

        mov     [esi+TAsmEdit._num_margin], ebx

.marginok1:

        mov     eax, [ASM.iconBreakpoints]
        add     ebx, [eax+TImage.width]

        mov     eax, [ASM.xspcIconBrkPnts]
        lea     ebx, [ebx+2*eax]           ; left/right margins to the image.

        mov     eax, [ASM.iconFoldings]
        add     ebx, [eax+TImage.width]

        add     ebx, 3+1                   ; 3px for word wrap cut + 1px whitespace at the line beginning.

.marginok:
        xchg    [esi+TAsmEdit._left_margin], ebx
        cmp     ebx, [esi+TAsmEdit._left_margin]
        je      .finish

;        exec    [.self], TAsmEdit:UpdateCaretPosition
        exec    [.self], TAsmEdit:ComputeScreenInChars

.finish:
        popad
        return


.digits:
        xor     ecx, ecx
        cmp     eax, 1          ; 0
        jb      .edig

        inc     ecx
        cmp     eax, 10         ; 1
        jb      .edig

        inc     ecx
        cmp     eax, 100        ; 2
        jb      .edig

        inc     ecx
        cmp     eax, 1000       ; 3
        jb      .edig

        inc     ecx
        cmp     eax, 10000      ; 4
        jb      .edig

        inc     ecx
        cmp     eax, 100000     ; 5
        jb      .edig

        inc     ecx
        cmp     eax, 1000000    ; 6
        jb      .edig

        inc     ecx
        cmp     eax, 10000000   ; 7
        jb      .edig

        inc     ecx
        cmp     eax, 100000000  ; 8
        jb      .edig

        inc     ecx
        cmp     eax, 1000000000 ; 9
        jb      .edig

        inc     ecx             ; 10

.edig:
        retn

endp





method TAsmEdit.ComputeScreenInChars
begin
        pushad
        mov     esi, [.self]

; width
        mov     eax, [esi+TAsmEdit._width]
        sub     eax, [esi+TAsmEdit._left_margin]
        cdq

        idiv    [esi+TAsmEdit._font_width]
        test    edx, edx
        jz      @f
        inc     eax
@@:
        mov     [esi+TAsmEdit._width_ch], eax

; height
        mov     eax, [esi+TAsmEdit._height]
        cdq

        idiv    [esi+TAsmEdit._font_height]
        test    edx, edx
        jz      @f
        inc     eax
@@:
        mov     [esi+TAsmEdit._height_ch], eax

        popad



        return
endp





method TAsmEdit.GetText
begin

        return
endp




method TAsmEdit.SetText
.context dd ?
begin
        pushad

        mov     ebx, [.self]

        xor     eax, eax
        mov     [.context], eax

; free the previous TText structure for the TAsmEdit._lines

        mov     esi, [ebx+TAsmEdit._lines]
        test    esi, esi
        jz      .create_new_lines

        xor     ecx, ecx

.free_loop:
        mov     eax, [esi+TText.GapBegin]
        cmp     ecx, eax
        cmovae  eax, [esi+TText.GapEnd]
        sub     eax, [esi+TText.GapBegin]
        add     eax, ecx

        cmp     eax, [esi+TText.Length]
        jae     .end_free_loop

        stdcall TextFree, [esi+eax]
        add     ecx, 4
        jmp     .free_loop

.end_free_loop:

        stdcall TextFree, esi

.create_new_lines:

        stdcall TextCreate, sizeof.TText
        mov     [ebx+TAsmEdit._lines], eax

; new line widths array.

        stdcall FreeMem, [ebx+TAsmEdit._line_widths]
        stdcall CreateArray, sizeof.TCounter
        mov     [ebx+TAsmEdit._line_widths], eax

        mov     edi, [.value]
        test    edi, edi
        jz      .end_text

.text_loop:

        cmp     byte [edi], 0
        je      .end_text

        exec    ebx, TAsmEdit:InsertNewLine, -1
        mov     esi, eax

        xor     ecx, ecx

.line_loop:
        stdcall DecodeUtf8, [edi]
        add     edi, edx

        test    eax, eax
        jz      .cr_ok

        cmp     eax, $0d
        je      .new_line
        cmp     eax, $0a
        je      .new_line

        stdcall EncodeUtf8, eax
        jc      .end_text

        stdcall TextAddChar, [esi], eax
        mov     [esi], edx
        jc      .end_text

        inc     ecx                             ; ecx counts the characters!
        jmp     .line_loop

.new_line:
        xor     eax, $0d xor $0a
        cmp     [edi], al
        jne     .cr_ok
        inc     edi

.cr_ok:
        push    eax

        mov     esi, [esi]
        mov     [esi+TTextLine.ch_len], ecx
        stdcall IncCount, [ebx+TAsmEdit._line_widths], ecx
        mov     [ebx+TAsmEdit._line_widths], edx

        mov     eax, [ebx+TAsmEdit._syntax_proc]
        test    eax, eax
        jz      .syntax_ok

        stdcall eax, esi, [.context]
        mov     [.context], eax

.syntax_ok:
        pop     eax
        test    eax, eax
        jnz     .text_loop

.end_text:
        exec    ebx, TAsmEdit:ComputeLeftMarginWidth
        exec    ebx, TAsmEdit:UpdateScrollBar


.finish:

        popad
        return
endp





; Navigation manipulation methods.

; Returns a pointer to variable where TTextLine with absolute coordinate [.y] is stored.
; If [.y] is outside the text, appends new lines up to [.y] and returns the result.
method TAsmEdit.TextLine  ; .y
begin
        push    esi ebx

        mov     eax, [.self]
        mov     esi, [eax+TAsmEdit._lines]

        mov     eax, [.y]
        shl     eax, 2

        cmp     eax, [esi+TText.GapBegin]
        jb      @f
        add     eax, [esi+TText.GapEnd]
        sub     eax, [esi+TText.GapBegin]
@@:
        cmp     eax, [esi+TText.Length]
        jae     .create_new

        lea     eax, [esi+eax]

.finish:
        pop     ebx esi
        return

.create_new:
        mov     ebx, [esi+TText.Length]
        sub     ebx, [esi+TText.GapEnd]
        add     ebx, [esi+TText.GapBegin]
        shr     ebx, 2

.loop:
        exec    [.self], TAsmEdit:InsertNewLine, ebx
        push    eax
        exec    [.self], TAsmEdit:SaveForUndo, ebx, chopInserted
        pop     eax

        cmp     ebx, [.y]
        lea     ebx, [ebx+1]
        jne     .loop

        inc     [AsmEditUndoGroup]

        jmp     .finish
endp



method TAsmEdit.ScrollToCaretX
begin
        pushad

        mov     esi, [.self]
        xor     ecx, ecx
        mov     [esp+4*regEAX], ecx

        mov     eax, [esi+TAsmEdit._width]
        sub     eax, [esi+TAsmEdit._left_margin]
        xor     edx, edx
        div     [esi+TAsmEdit._font_width]

        dec     eax
        mov     ebx, eax    ; max column

        shr     eax, 2
        test    [esi+TAsmEdit._options], aeoScrollXMore
        cmovz   edi, ecx
        cmovnz  edi, eax        ; scroll more than needed.

        get     edx, esi, TAsmEdit:PosX

        mov     eax, [esi+TAsmEdit._caretPos.x]
        sub     eax, edx                       ; offset from the screen home.
        jz      .finish                        ; no need to scroll
        js      .do_scroll

        sub     eax, ebx
        jle     .finish

        neg     edi

.do_scroll:
        sub     eax, edi
        add     edx, eax
        cmovs   edx, ecx

        set     esi, TAsmEdit:PosX, edx
        exec    esi, TAsmEdit:RectChanged2, ecx

        inc     dword [esp+4*regEAX]

.finish:
        exec    esi, TAsmEdit:UpdateCaretPos
        popad
        return
endp




method TAsmEdit.ScrollToCaretY
begin
        pushad

        mov     esi, [.self]
        xor     ecx, ecx
        mov     [esp+4*regEAX], ecx

        mov     eax, [esi+TAsmEdit._height]
        xor     edx, edx
        div     [esi+TAsmEdit._font_height]

        dec     eax
        mov     ebx, eax  ; max line

        shr     eax, 2
        test    [esi+TAsmEdit._options], aeoScrollYMore
        cmovz   edi, ecx
        cmovnz  edi, eax        ; scroll more than needed.

        get     edx, esi, TAsmEdit:PosY

        mov     eax, [esi+TAsmEdit._caretPos.y]
        sub     eax, edx                       ; offset from the screen home.
        jz      .finish                        ; no need to scroll
        js      .do_scroll

        sub     eax, ebx
        jle     .finish

        neg     edi

.do_scroll:

        sub     eax, edi
        add     edx, eax
        cmovs   edx, ecx

        set     esi, TAsmEdit:PosY, edx
        exec    esi, TAsmEdit:RectChanged2, ecx

        inc     dword [esp+4*regEAX]

.finish:
        exec    esi, TAsmEdit:UpdateCaretPos
        popad
        return
endp






method TAsmEdit.CaretLeftRight
begin
        pushad

        mov     esi, [.self]
        xor     edx, edx                    ; minimal caret pos 0

        mov     ebx, [esi+TAsmEdit._caretPos.x]
        add     ebx, [.count]
        jns     .leftok

        xor     ebx, ebx
        test    [esi+TAsmEdit._options], aeoNavXCaretWrap
        jz      .set_pos

        cmp     [esi+TAsmEdit._caretPos.y], ebx
        je      .finish

        exec    esi, TAsmEdit:CaretUpDown, -1, FALSE
        exec    esi, TAsmEdit:CaretEnd
        jmp     .finish

.leftok:
        test    [esi+TAsmEdit._options], aeoNavXCaretWrap
        jz      .set_pos

        exec    esi, TAsmEdit:TextLine, [esi+TAsmEdit._caretPos.y]

        mov     eax, [eax]
        cmp     ebx, [eax+TTextLine.ch_len]
        jbe     .set_pos

        exec    esi, TAsmEdit:CaretUpDown, 1, FALSE
        exec    esi, TAsmEdit:CaretHome
        jmp     .finish

.set_pos:
        mov     [esi+TAsmEdit._caretPos.x], ebx

        exec    esi, TAsmEdit:ScrollToCaretX

.finish:
        popad
        return
endp





method TAsmEdit.CaretUpDown
.rect RECT
begin
        pushad

        mov     esi, [.self]
        xor     edx, edx                    ; minimal caret pos 0

        mov     ebx, [esi+TAsmEdit._caretPos.y]
        get     ecx, esi, TAsmEdit:PosY

        add     ebx, [.count]
        cmovs   ebx, edx

        add     ecx, [.count]
        cmovs   ecx, edx

        mov     edi, [esi+TAsmEdit._lines]

        mov     eax, [edi+TText.Length]
        add     eax, [edi+TText.GapBegin]
        sub     eax, [edi+TText.GapEnd]
        shr     eax, 2

        cmp     ebx, eax
        cmova   ebx, eax

        add     eax, 2
        sub     eax, [esi+TAsmEdit._height_ch]
        cmovs   eax, edx

        cmp     ecx, eax
        cmova   ecx, eax

        mov     [esi+TAsmEdit._caretPos.y], ebx

        cmp     [.scroll], edx
        je      .scroll_ok

        set     esi, TAsmEdit:PosY, ecx
        exec    esi, TAsmEdit:RectChanged2, edx

.scroll_ok:

        exec    esi, TAsmEdit:ScrollToCaretY

.finish:
        popad
        return
endp




method TAsmEdit.CaretHome
begin
        push    eax
        mov     eax, [.self]
        and     [eax+TAsmEdit._caretPos.x], 0
        exec    eax, TAsmEdit:ScrollToCaretX
        pop     eax
        return
endp


method TAsmEdit.CaretEnd
begin
        push    esi
        mov     esi, [.self]

        exec    esi, TAsmEdit:TextLine, [esi+TAsmEdit._caretPos.y]      ; returns CF=1 and eax=0 on error.
        mov     eax, [eax]
        mov     eax, [eax+TTextLine.ch_len]

        mov     [esi+TAsmEdit._caretPos.x], eax
        exec    esi, TAsmEdit:ScrollToCaretX

        pop     esi
        return
endp


method TAsmEdit.CaretScrHome
begin
        push    esi

        mov     esi, [.self]

        get     eax, esi, TAsmEdit:PosY

        mov     [esi+TAsmEdit._caretPos.y], eax
        exec    esi, TAsmEdit:UpdateCaretPos

        pop     esi
        return
endp


method TAsmEdit.CaretScrEnd
begin
        push    esi edx

        mov     esi, [.self]

        mov     eax, [esi+TAsmEdit._height]
        xor     edx, edx
        div     [esi+TAsmEdit._font_height]
        add     eax, [esi+TAsmEdit._scrY.pos]
        dec     eax
        js      .finish

        mov     [esi+TAsmEdit._caretPos.y], eax
        exec    esi, TAsmEdit:UpdateCaretPos

.finish:
        pop     edx esi
        return
endp



method TAsmEdit.CaretTxtHome
begin
        push    eax
        mov     eax, [.self]
        and     [eax+TAsmEdit._caretPos.y], 0
        exec    eax, TAsmEdit:ScrollToCaretY
        pop     eax
        return
endp


method TAsmEdit.CaretTxtEnd
begin
        push    esi edx

        mov     esi, [.self]
        mov     edx, [esi+TAsmEdit._lines]

        mov     eax, [edx+TText.Length]
        add     eax, [edx+TText.GapBegin]
        sub     eax, [edx+TText.GapEnd]
        shr     eax, 2

        mov     [esi+TAsmEdit._caretPos.y], eax
        exec    esi, TAsmEdit:ScrollToCaretY

        pop     edx esi
        return
endp


; Selection control methods


method TAsmEdit.SelStartHere
begin
        pushad

        mov     esi, [.self]
        exec    esi, TAsmEdit:SelChangeBegin
        mov     eax, [esi+TAsmEdit._caretPos.x]
        mov     edx, [esi+TAsmEdit._caretPos.y]
        mov     [esi+TAsmEdit._selStart.x], eax
        mov     [esi+TAsmEdit._selStart.y], edx
        exec    esi, TAsmEdit:SelChangeEnd

        popad
        return
endp



method TAsmEdit.SelEndHere
begin
        pushad

        mov     esi, [.self]
        exec    esi, TAsmEdit:SelChangeBegin
        mov     eax, [esi+TAsmEdit._caretPos.x]
        mov     edx, [esi+TAsmEdit._caretPos.y]
        mov     [esi+TAsmEdit._selEnd.x], eax
        mov     [esi+TAsmEdit._selEnd.y], edx
        exec    esi, TAsmEdit:SelChangeEnd

        popad
        return
endp


method TAsmEdit.SelChangeBegin
begin
        pushad
        mov     esi, [.self]
        xor     edx, edx
        xchg    edx, [esi+TAsmEdit._selAreaUpdate]
        test    edx, edx
        jz      .old_ok

        stdcall FreeMem, edx
        xor     edx, edx

.old_ok:
        exec    esi, TAsmEdit:_AddSelRectangles, edx
        mov     [esi+TAsmEdit._selAreaUpdate], edx

        popad
        return
endp


method TAsmEdit.SelChangeEnd
begin
        pushad
        mov     esi, [.self]
        mov     edx, [esi+TAsmEdit._selAreaUpdate]

        exec    esi, TAsmEdit:_AddSelRectangles, edx
        mov     ecx, [edx+TArray.count]
        lea     edi, [edx+TArray.array]

.loop:
        dec     ecx
        js      .freelist

        mov     eax, [edi+RECT.left]
        mov     ebx, [edi+RECT.top]

        imul    eax, [esi+TAsmEdit._font_width]
        imul    ebx, [esi+TAsmEdit._font_height]

        add     eax, [esi+TAsmEdit._left_margin]

        mov     [edi+RECT.left], eax
        mov     [edi+RECT.top], ebx

        mov     eax, [edi+RECT.right]
        mov     ebx, [edi+RECT.bottom]

        imul    eax, [esi+TAsmEdit._font_width]
        imul    ebx, [esi+TAsmEdit._font_height]

        add     eax, [esi+TAsmEdit._left_margin]

        mov     [edi+RECT.right], eax
        mov     [edi+RECT.bottom], ebx

        exec    esi, TAsmEdit:RectChanged2, edi

        add     edi, sizeof.RECT
        jmp     .loop

.freelist:
        stdcall FreeMem, edx
        xor     edx, edx
        mov     [esi+TAsmEdit._selAreaUpdate], edx

        popad
        return
endp


; Line handling and formatting methods.



; Undo/Redo methods


method TAsmEdit.SaveForUndo
begin
        pushad

        mov     esi, [.self]
        mov     edx, [esi+TAsmEdit._pUndo]
        mov     ebx, [.LineNum]
        mov     ecx, [.operation]

        test    edx, edx
        jz      .add_new

        cmp     [edx+TUndoInfo.line], ebx
        jne     .add_new

        cmp     ecx, chopChanged
        jne     .add_new

        mov     eax, [AsmEditUndoGroup]
        cmp     eax, [edx+TUndoInfo.group]
        jne     .add_new

        cmp     [edx+TUndoInfo.operation], ecx
        je      .finish                         ; group the chopChanged for the same line.

; add new record in the chain.

.add_new:
        OutputValue "Add new TUndoInfo for operation: ", ecx, 10, -1

        stdcall GetMem, sizeof.TUndoInfo
        jc      .finish

        mov     edi, eax

        mov     [edi+TUndoInfo.pNext], edx
        mov     [edi+TUndoInfo.operation], ecx

        mov     eax, [AsmEditUndoGroup]
        mov     [edi+TUndoInfo.group], eax

        mov     [esi+TAsmEdit._pUndo], edi

        mov     ecx, [esi+TAsmEdit._caretPos.x]
        mov     edx, [esi+TAsmEdit._caretPos.y]

        mov     [edi+TUndoInfo.line], ebx
        mov     [edi+TUndoInfo.caretPos.x], ecx
        mov     [edi+TUndoInfo.caretPos.y], edx

        mov     ecx, [esi+TAsmEdit._selStart.x]
        mov     edx, [esi+TAsmEdit._selStart.y]
        mov     [edi+TUndoInfo.selStart.x], ecx
        mov     [edi+TUndoInfo.selStart.y], edx

        mov     ecx, [esi+TAsmEdit._selEnd.x]
        mov     edx, [esi+TAsmEdit._selEnd.y]
        mov     [edi+TUndoInfo.selEnd.x], ecx
        mov     [edi+TUndoInfo.selEnd.y], edx

        mov     edx, [esi+TAsmEdit._lines]
        shl     ebx, 2

        pos2ofs ebx, edx
        stdcall TextDup, [edx+ebx]

        mov     [edi+TUndoInfo.pText], eax
        mov     [eax+TTextLine.syn_array], 0    ; don't copy syntax array pointer!

; now free all redo records, because every new stored undo point invalidates the whole redo list.

.clean_redo:

        mov     ecx, [esi+TAsmEdit._pRedo]
        and     [esi+TAsmEdit._pRedo], 0

.clean_loop:
        jecxz      .finish

        stdcall TextFree, [ecx+TUndoInfo.pText]

        push    ecx
        mov     ecx, [ecx+TUndoInfo.pNext]

        stdcall FreeMem ; from the stack
        jmp     .clean_loop

.finish:
        popad
        return
endp



method TAsmEdit.Undo
.group dd ?
begin
        pushad

        mov     esi, [.self]
        mov     edi, [esi+TAsmEdit._pUndo]

        test    edi, edi
        jz      .finish         ; nothing to undo.

        mov     eax, [edi+TUndoInfo.group]
        mov     [.group], eax

.loop:
        mov     edx, [edi+TUndoInfo.pNext]
        mov     [esi+TAsmEdit._pUndo], edx           ; remove from the UNDO list.

        mov     edx, [esi+TAsmEdit._pRedo]
        mov     [edi+TUndoInfo.pNext], edx
        mov     [esi+TAsmEdit._pRedo], edi           ; insert in the REDO list.

        call    __ApplyUndo

        mov     edi, [esi+TAsmEdit._pUndo]
        test    edi, edi
        jz      .refresh

        mov     eax, [edi+TUndoInfo.group]
        cmp     eax, [.group]
        je      .loop

.refresh:
        exec    esi, TAsmEdit:ScrollToCaretY
        exec    esi, TAsmEdit:ScrollToCaretX

        and     [esi+TAsmEdit._img_txt_valid], 0
        and     [esi+TAsmEdit._img_margin_valid], 0
        exec    esi, TAsmEdit:RectChanged2, 0

        exec    esi, TAsmEdit:UpdateCaretPos
        exec    esi, TAsmEdit:UpdateScrollBar

.finish:
        popad
        return
endp



method TAsmEdit.Redo
.group dd ?
begin
        pushad

        mov     esi, [.self]
        mov     ebx, [esi+TAsmEdit._lines]

        mov     edi, [esi+TAsmEdit._pRedo]

        test    edi, edi
        jz      .finish

        mov     eax, [edi+TUndoInfo.group]
        mov     [.group], eax

.loop:
        mov     edx, [edi+TUndoInfo.pNext]
        mov     [esi+TAsmEdit._pRedo], edx           ; remove from the REDO list.

        mov     edx, [esi+TAsmEdit._pUndo]
        mov     [edi+TUndoInfo.pNext], edx
        mov     [esi+TAsmEdit._pUndo], edi           ; insert in the UNDO list.

        call    __ApplyUndo

        mov     edi, [esi+TAsmEdit._pRedo]
        test    edi, edi
        jz      .refresh

        mov     eax, [edi+TUndoInfo.group]
        cmp     eax, [.group]
        je      .loop

.refresh:
        exec    esi, TAsmEdit:ScrollToCaretY
        exec    esi, TAsmEdit:ScrollToCaretX

        and     [esi+TAsmEdit._img_txt_valid], 0
        and     [esi+TAsmEdit._img_margin_valid], 0
        exec    esi, TAsmEdit:RectChanged2, 0

        exec    esi, TAsmEdit:UpdateCaretPos
        exec    esi, TAsmEdit:UpdateScrollBar

.finish:
        popad
        return
endp


; esi - pointer to TAsmEdit
; edi - pointer to TUndoInfo

proc __ApplyUndo
begin
        pushad

        push    [edi+TUndoInfo.caretPos.x] [esi+TAsmEdit._caretPos.x]
        pop     [edi+TUndoInfo.caretPos.x] [esi+TAsmEdit._caretPos.x] ; xchg caret_x

        push    [edi+TUndoInfo.caretPos.y] [esi+TAsmEdit._caretPos.y]
        pop     [edi+TUndoInfo.caretPos.y] [esi+TAsmEdit._caretPos.y] ; xchg caret_y

        push    [edi+TUndoInfo.selStart.x] [esi+TAsmEdit._selStart.x]
        pop     [edi+TUndoInfo.selStart.x] [esi+TAsmEdit._selStart.x] ; xchg selStart_x

        push    [edi+TUndoInfo.selStart.y] [esi+TAsmEdit._selStart.y]
        pop     [edi+TUndoInfo.selStart.y] [esi+TAsmEdit._selStart.y] ; xchg selStart_y

        push    [edi+TUndoInfo.selEnd.x] [esi+TAsmEdit._selEnd.x]
        pop     [edi+TUndoInfo.selEnd.x] [esi+TAsmEdit._selEnd.x] ; xchg selEnd_x

        push    [edi+TUndoInfo.selEnd.y] [esi+TAsmEdit._selEnd.y]
        pop     [edi+TUndoInfo.selEnd.y] [esi+TAsmEdit._selEnd.y] ; xchg selEnd_y

        mov     ebx, [esi+TAsmEdit._lines]
        mov     eax, [edi+TUndoInfo.operation]

        cmp     eax, chopChanged
        je      .op_changed

        cmp     eax, chopDeleted
        je      .op_deleted

        cmp     eax, chopInserted
        je      .op_inserted

        int3           ; INVALID OPERATION
        jmp     .finish

.op_inserted:
        mov     [edi+TUndoInfo.operation], chopDeleted

        mov     eax, [edi+TUndoInfo.line]
        shl     eax, 2
        stdcall TextMoveGap, ebx, eax

        mov     eax, [ebx+TText.GapEnd]
        mov     ecx, [ebx+eax]

        mov     [edi+TUndoInfo.pText], ecx

        stdcall FreeMem, [ecx+TTextLine.syn_array]
        and     [ecx+TTextLine.syn_array], 0
        add     [ebx+TText.GapEnd], 4

        stdcall DecCount, [esi+TAsmEdit._line_widths], [ecx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx

        jmp     .finish

.op_deleted:

        mov     [edi+TUndoInfo.operation], chopInserted

        mov     eax, [edi+TUndoInfo.line]
        shl     eax, 2
        stdcall TextMoveGap, ebx, eax
        stdcall TextSetGapSize, ebx, 4
        mov     [esi+TAsmEdit._lines], edx
        mov     ebx, edx
        mov     eax, [ebx+TText.GapBegin]
        and     dword [ebx+eax], 0
        add     [ebx+TText.GapBegin], 4
        jmp     .restore_txt


.op_changed:
        mov     eax, [edi+TUndoInfo.line]
        shl     eax, 2          ; in bytes
        pos2ofs eax, ebx

.restore_txt:
        pushd   [edi+TUndoInfo.pText] [ebx+eax]
        popd    [edi+TUndoInfo.pText] [ebx+eax]      ; xchg TTextLine structures of the line.

        mov     ecx, [edi+TUndoInfo.pText]
        jecxz   @f
        stdcall FreeMem, [ecx+TTextLine.syn_array]
        and     [ecx+TTextLine.syn_array], 0
@@:

        mov     ecx, [ebx+eax]
        and     [ecx+TTextLine.syn_array], 0

; fix the horizontal scrollbar size.

        stdcall IncCount, [esi+TAsmEdit._line_widths], [ecx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx

        mov     ecx, [edi+TUndoInfo.pText]
        jecxz   @f
        stdcall DecCount, [esi+TAsmEdit._line_widths], [ecx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx
@@:

.finish:
;        OutputValue "Operation after: ", [edi+TUndoInfo.operation], 10, -1

        popad
        return
endp





; Text editing methods




method TAsmEdit.InsertNewLine   ;, .LineNum
begin
        pushad

        mov     esi, [.self]
        mov     edx, [esi+TAsmEdit._lines]

        stdcall TextSetGapSize, edx, 4
        mov     [esi+TAsmEdit._lines], edx
        jc      .finish

        mov     eax, [.LineNum]
        shl     eax, 2

        stdcall TextMoveGap, edx, eax

        mov     ebx, [edx+TText.GapBegin]
        add     ebx, edx

        stdcall TextCreate, sizeof.TTextLine
        jc      .finish

        mov     [ebx], eax
        mov     [esp+4*regEAX], ebx                     ; return value in EAX is the pointer to the new inserted TText.

        add     [edx+TText.GapBegin], 4

        stdcall IncCount, [esi+TAsmEdit._line_widths], 0
        mov     [esi+TAsmEdit._line_widths], edx

        clc

.finish:
        popad
        return
endp



method TAsmEdit.DeleteLine      ;, .LineNum
begin
        pushad

        mov     esi, [.self]
        mov     edi, [esi+TAsmEdit._lines]
        mov     ebx, [.LineNum]
        test    ebx, ebx
        cmovs   ebx, [esi+TAsmEdit._caretPos.y]
        mov     [.LineNum], ebx

        shl     ebx, 2
        stdcall TextMoveGap, edi, ebx

        mov     ebx, [edi+TText.GapEnd]
        cmp     ebx, [edi+TText.Length]
        jae     .not_deleted

        exec    esi, TAsmEdit:SaveForUndo, [.LineNum], chopDeleted

        mov     edx, [edi+ebx]
        stdcall FreeMem, [edx+TTextLine.syn_array]
        stdcall TextFree, edx

        add     [edi+TText.GapEnd], 4

        clc
        popad
        return

.not_deleted:
        stc
        popad
        return
endp


method TAsmEdit.DeleteChar  ; .LineNum, .CharPos
begin
        pushad

        mov     esi, [.self]
        mov     edi, [esi+TAsmEdit._lines]
        mov     ebx, [.LineNum]
        mov     eax, [.CharPos]
        test    ebx, ebx
        cmovs   ebx, [esi+TAsmEdit._caretPos.y]
        cmovs   eax, [esi+TAsmEdit._caretPos.x]
        mov     [.CharPos], eax
        mov     [.LineNum], ebx

        shl     ebx, 2
        cmp     ebx, [edi+TText.GapBegin]
        jb      .skip
        add     ebx, [edi+TText.GapEnd]
        sub     ebx, [edi+TText.GapBegin]
.skip:
        cmp     ebx, [edi+TText.Length]
        jae     .not_deleted

        mov     ebx, [edi+ebx]
        stdcall TextIndexToPos, ebx, [.CharPos]
        jc      .not_deleted

        stdcall TextMoveGap, ebx, eax
        mov     edx, [ebx+TTextLine.GapEnd]
        cmp     edx, [ebx+TTextLine.Length]
        jae     .not_deleted

        exec    esi, TAsmEdit:SaveForUndo, [.LineNum], chopChanged

        movzx   eax, byte [ebx+edx]

                                ; compute the length of the UTF-8 character by its first byte (version without branches) Is it fast?
        not     al              ; utf-8 starts with one of: 0xxxxxxx, 110xxxxx, 1110xxxx, 11110xxx, 111110xx, 1111110x
        bsr     eax, eax        ; eax =                 7,  5,  4,  3,  2,  1
        sub     eax, 7          ; eax =                 0, -2, -3, -4, -5, -6
                                ; CF  =                 0,  1,  1,  1,  1,  1
        not     eax             ; eax =                -1, +1, +2, +3, +4, +5
        sbb     eax, -2         ; eax = edx + 2 - CF = +1, +2, +3, +4, +5, +6

        add     edx, eax
        cmp     edx, [ebx+TTextLine.Length]
        cmova   edx, [ebx+TTextLine.Length]
        mov     [ebx+TTextLine.GapEnd], edx

        stdcall FreeMem, [ebx+TTextLine.syn_array]
        and     [ebx+TTextLine.syn_array], 0            ; reset the syntax array of the line...

        stdcall DecCount, [esi+TAsmEdit._line_widths], [ebx+TTextLine.ch_len]
        dec     [ebx+TTextLine.ch_len]
        stdcall IncCount, edx, [ebx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx

        and     [esi+TAsmEdit._img_txt_valid], 0
        and     [esi+TAsmEdit._img_margin_valid], 0
        exec    esi, TAsmEdit:RectChanged2, 0
        exec    esi, TAsmEdit:UpdateCaretPos

        clc
        popad
        return

.not_deleted:
        stc
        popad
        return
endp


method TAsmEdit.JoinNextLine    ; .LineNum
begin
        pushad

        mov     esi, [.self]
        mov     edi, [esi+TAsmEdit._lines]
        mov     ebx, [.LineNum]
        test    ebx, ebx
        cmovs   ebx, [esi+TAsmEdit._caretPos.y]
        mov     [.LineNum], ebx

        shl     ebx, 2
        lea     ecx, [ebx+4]

        stdcall TextMoveGap, edi, ecx   ; move the gap between the two lines.

; here ebx is before the gap and ecx is after!

        add     ecx, [edi+TText.GapEnd]
        sub     ecx, [edi+TText.GapBegin]
        cmp     ecx, [edi+TText.Length]
        jae     .finish ; nothing to join

        cmp     ebx, [edi+TText.Length]
        jae     .finish

        exec    esi, TAsmEdit:SaveForUndo, [.LineNum], chopChanged
        inc     [.LineNum]
        exec    esi, TAsmEdit:SaveForUndo, [.LineNum], chopDeleted

        add     [edi+TText.GapEnd], 4   ; free the place of the next line...

; first fix the new line widths...

        mov     edx, [esi+TAsmEdit._line_widths]

        mov     eax, [edi+ebx]
        push    [eax+TTextLine.ch_len]
        stdcall DecCount, edx, [eax+TTextLine.ch_len]

        mov     eax, [edi+ecx]
        push    [eax+TTextLine.ch_len]
        stdcall DecCount, edx, [eax+TTextLine.ch_len]

        mov     [esi+TAsmEdit._line_widths], edx

        pop     eax edx
        add     eax, edx
        mov     edx, [edi+ebx]
        mov     [edx+TTextLine.ch_len], eax

        stdcall IncCount, [esi+TAsmEdit._line_widths], eax
        mov     [esi+TAsmEdit._line_widths], edx

; then join the texts

        stdcall TextAddText, [edi+ebx], -1, [edi+ecx]
        mov     [edi+ebx], edx

; then free the next line TTextLine structure...

        mov     eax, [edi+ecx]
        stdcall FreeMem, [eax+TTextLine.syn_array]
        stdcall TextFree, eax     ; delete the next line... but what about the debug info and other attributes?

; reset the syntax coloring of the result line
        mov     eax, [edi+ebx]
        stdcall FreeMem, [eax+TTextLine.syn_array]
        and     [eax+TTextLine.syn_array], 0

.finish:
        popad
        return
endp



method TAsmEdit.SplitLine
begin
        pushad

        mov     esi, [.self]
        mov     ebx, [.LineNum]
        mov     ecx, [.PosX]
        test    ebx, ebx
        cmovs   ebx, [esi+TAsmEdit._caretPos.y]
        cmovs   ecx, [esi+TAsmEdit._caretPos.x]
        mov     [.LineNum], ebx
        mov     [.PosX], ecx

        exec    esi, TAsmEdit:TextLine, ebx     ; ensure the line ebx exists.

        inc     ebx
        exec    esi, TAsmEdit:InsertNewLine, ebx
        mov     edi, eax
        exec    esi, TAsmEdit:SaveForUndo, ebx, chopInserted

        mov     edx, [esi+TAsmEdit._lines]
        mov     ebx, [edx+4*ebx-4]                ; the gap, after the InsertNewLine is located just after the ebx line.

        stdcall TextIndexToPos, ebx, [.PosX]
        jc      .finish                         ; the posX is outside the text, so there is nothing to split!

        push    eax
        exec    esi, TAsmEdit:SaveForUndo, [.LineNum], chopChanged
        pop     eax

        stdcall TextMoveGap, ebx, eax
        stdcall DecCount, [esi+TAsmEdit._line_widths], [ebx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx

        mov     eax, [ebx+TTextLine.GapEnd]
        mov     ecx, [ebx+TTextLine.Length]
        mov     [ebx+TTextLine.GapEnd], ecx
        sub     ecx, eax
        add     eax, ebx

        push    eax ecx

        stdcall StrLenUtf8, eax, ecx
        push    eax eax

        mov     ecx, [edi]

        stdcall FreeMem, [ecx+TTextLine.syn_array]
        and     [ecx+TTextLine.syn_array], 0

        stdcall FreeMem, [ebx+TTextLine.syn_array]
        and     [ebx+TTextLine.syn_array], 0

        stdcall DecCount, [esi+TAsmEdit._line_widths], [ecx+TTextLine.ch_len]
        pop     eax
        mov     [ecx+TTextLine.ch_len], eax
        stdcall IncCount, edx, eax

        pop     eax
        sub     [ebx+TTextLine.ch_len], eax
        stdcall IncCount, edx, [ebx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx

        pop     ecx eax
        stdcall TextAddBytes, [edi], -1, eax, ecx
        mov     [edi], edx

        exec    esi, TAsmEdit:DelTrailingSpaces, [.LineNum]

.finish:
        inc     [AsmEditUndoGroup]
        popad
        return
endp



method TAsmEdit.DeleteSelection
.rect RECT
.list dd ?
begin
        pushad

        exec    [.self], TAsmEdit:_AddSelRectanglesAbs, 0
        mov     [.list], edx

        lea     edi, [edx+TArray.array]
        mov     ecx, [edx+TArray.count]
        test    ecx, ecx
        jz      .exit_empty

        mov     eax, ecx
        shl     eax, 4  ; sizeof.RECT
        add     edi, eax

.rect_loop:

        sub     edi, sizeof.RECT
        dec     ecx
        js      .finish

        mov     ebx, [edi+RECT.bottom]

.y_loop:
        dec     ebx
        cmp     ebx, [edi+RECT.top]
        jl      .rect_loop

        exec    [.self], TAsmEdit:SaveForUndo, ebx, chopChanged

        exec    [.self], TAsmEdit:AddTrailingSpaces, ebx, [edi+RECT.left]
        exec    [.self], TAsmEdit:TextLine, ebx
        mov     esi, [eax]

        stdcall TextIndexToPos, esi, [edi+RECT.left]
        stdcall TextMoveGap, esi, eax

        mov     eax, [esi+TTextLine.ch_len]
        cmp     eax, [edi+RECT.right]
        cmova   eax, [edi+RECT.right]                   ; delete up to the min(length, RECT.right) utf-8 character position!

        push    ecx
        push    eax             ; char count
        mov     ecx, [.self]

        stdcall TextIndexToPos, esi, eax
        sub     eax, [esi+TTextLine.GapBegin]
        add     [esi+TTextLine.GapEnd], eax             ; delete eax bytes!

        stdcall DecCount, [ecx+TAsmEdit._line_widths], [esi+TTextLine.ch_len]

        pop     eax
        sub     eax, [edi+RECT.left]    ; deleted characters!
        sub     [esi+TTextLine.ch_len], eax

        stdcall IncCount, edx, [esi+TTextLine.ch_len]
        mov     [ecx+TAsmEdit._line_widths], edx

        pop     ecx

; join the lines, if the selection is right to infinity:

        cmp     [edi+RECT.right], $7fffffff
        je      .joinnext

        exec    [.self], TAsmEdit:DelTrailingSpaces, ebx
        jmp     .y_loop

.joinnext:
        exec    [.self], TAsmEdit:JoinNextLine, ebx
        jmp     .y_loop

.finish:
        mov     esi, [.self]

        get     eax, esi, TAsmEdit:SelStartX
        get     ecx, esi, TAsmEdit:SelEndX
        cmp     eax, ecx
        cmovg   eax, ecx

        get     edx, esi, TAsmEdit:SelStartY
        get     ecx, esi, TAsmEdit:SelEndY
        cmp     edx, ecx
        cmovg   edx, ecx

        mov     [esi+TAsmEdit._caretPos.x], eax
        mov     [esi+TAsmEdit._caretPos.y], edx

        xor     eax, eax
        mov     ecx, [esi+TAsmEdit._width_ch]
        mov     edx, [esi+TAsmEdit._height_ch]

        mov     [.rect.left], eax
        mov     [.rect.top], eax
        mov     [.rect.right], ecx
        mov     [.rect.bottom], edx
        lea     eax, [.rect]
        exec    esi, TAsmEdit:_RenderTxt, [esi+TAsmEdit._imgText], eax
        exec    esi, TAsmEdit:_RenderLM, [esi+TAsmEdit._imgMargin], [.rect.top], [.rect.bottom]
        exec    esi, TAsmEdit:RectChanged2, 0
        exec    esi, TAsmEdit:ScrollToCaretX
        exec    esi, TAsmEdit:ScrollToCaretY

.exit:
        inc     [AsmEditUndoGroup]
        stdcall FreeMem, [.list]
        clc
        popad
        return

.exit_empty:

        stdcall FreeMem, [.list]
        stc
        popad
        return
endp


method TAsmEdit.InsertString    ;, .text
begin


        return
endp



method TAsmEdit.ReplaceSelection        ;, .text
begin






        return
endp





method TAsmEdit.AddTrailingSpaces  ; .LineNum, .CharX
begin
        pushad

        mov     esi, [.self]
        mov     ebx, [.LineNum]
        mov     eax, [.PosX]
        test    ebx, ebx
        cmovs   ebx, [esi+TAsmEdit._caretPos.y]
        cmovs   eax, [esi+TAsmEdit._caretPos.x]
        mov     [.PosX], eax

        mov     edi, [esi+TAsmEdit._lines]

        shl     ebx, 2
        cmp     ebx, [edi+TText.GapBegin]
        jb      .skip
        add     ebx, [edi+TText.GapEnd]
        sub     ebx, [edi+TText.GapBegin]
.skip:
        cmp     ebx, [edi+TText.Length]
        jae     .finish         ; the line is outside the text. Add empty lines???

        mov     edx, [edi+ebx]

        mov     ecx, [.PosX]
        sub     ecx, [edx+TTextLine.ch_len]
        jle     .finish                         ; the caret is inside the line.

        push    edx
        stdcall DecCount, [esi+TAsmEdit._line_widths], [edx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx
        pop     edx

        add     [edx+TTextLine.ch_len], ecx

        push    edx
        stdcall IncCount, [esi+TAsmEdit._line_widths], [edx+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx
        pop     edx

        stdcall TextMoveGap, edx, -1
        stdcall TextSetGapSize, edx, ecx
        mov     [edi+ebx], edx

        mov     edi, edx
        add     edi, [edx+TTextLine.GapBegin]
        add     [edx+TTextLine.GapBegin], ecx
        mov     al, " "
        rep stosb

.finish:
        popad
        return
endp




method TAsmEdit.DelTrailingSpaces       ;, ;.LineNum
begin
        pushad

        mov     esi, [.self]
        mov     ebx, [.LineNum]
        test    ebx, ebx
        cmovs   ebx, [esi+TAsmEdit._caretPos.y]

        mov     edi, [esi+TAsmEdit._lines]

        shl     ebx, 2
        cmp     ebx, [edi+TText.GapBegin]
        jb      .skip
        add     ebx, [edi+TText.GapEnd]
        sub     ebx, [edi+TText.GapBegin]
.skip:
        cmp     ebx, [edi+TText.Length]
        jae     .finish         ; the line is outside the text. Add empty lines???

        mov     edi, [edi+ebx]

        stdcall TextMoveGap, edi, -1

        mov     edx, edi
        add     edx, [edi+TTextLine.GapBegin]

; search back the first non-whitespace...

.loop:
        dec     edx
        cmp     edx, edi
        jl      .found

        cmp     byte [edx], ' '
        jbe     .loop

.found:
        inc     edx
        mov     ecx, edx
        sub     ecx, edi

        xchg    ecx, [edi+TTextLine.GapBegin]
        sub     ecx, [edi+TTextLine.GapBegin]

        stdcall DecCount, [esi+TAsmEdit._line_widths], [edi+TTextLine.ch_len]
        sub     [edi+TTextLine.ch_len], ecx
        stdcall IncCount, edx, [edi+TTextLine.ch_len]
        mov     [esi+TAsmEdit._line_widths], edx

        stdcall FreeMem, [edi+TTextLine.syn_array]
        and     [edi+TTextLine.syn_array], 0

.finish:
        popad
        return
endp



endmodule


include "themes/asmedit.asm"
include "syntax/SyntaxFASM.asm"