; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TButton object class
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: Represents GUI button.
;_________________________________________________________________________________________
module "TButton library"


; Button states for the field TButton._state; The height of TButton._canvas = btnStateCount * TButton._height

bsNeutral    = 0
bsHovered    = 1
bsPressed    = 2
bsDisabled   = 3
btnStateCount = 4

; icon positions

iposLeft   = 0
iposRight  = 1
iposTop    = 2
iposBottom = 3
iposCenter = 4


object TButton, TWindow

; private fields

  ._state        dd ?      ; internal state variable.

  ._textalign    dd ?
  ._icon         dd ?
  ._iconpos      dd ?

  ._menu         dd ?

  ._OnClick      dd ?

  ._ModalResult  dd ?

  ._canvas       dd ?
  ._canvas_valid dd ?

; parameters

  param .OnClick, ._OnClick, ._OnClick

  param .TextAlign,    ._textalign,  .SimpleSetParam:3
  param .Icon,         ._icon,       .SetIcon
  param .IconPosition, ._iconpos,    .SimpleSetParam:3

  param .State,        .GetState,    ._state

  param .ModalResult, ._ModalResult, ._ModalResult
  param .Menu, ._menu, ._menu

; methods

  ; parameters handling methods

  method .GetState

  method .SetIcon, .value
  method .SimpleSetParam, .param, .get, .value

  ; system events methods

  method .Create, .Parent
  method .ForceRefresh

  method ._RenderView
  method .UpdateAction, .action

  method .SelfPaint, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height
  method .Resize, .newWidth, .newHeight
  method .SetCaption, .value

  method .EventFocusIn
  method .EventFocusOut

  method .EventMouseEnter
  method .EventMouseLeave
  method .EventMouseMove, .x, .y, .kbdState

  method .EventButtonPress, .button, .kbdState, .x, .y
  method .EventButtonRelease, .button, .kbdState, .x, .y

  method .EventKeyPress, .utf8, .scancode, .kbdState
  method .EventKeyRelease, .scancode, .kbdState

endobj


interface TButton.OnClick, .self, .button


; new

method TButton.Create
begin
        inherited [.Parent]

        mov     eax, [.self]
        mov     [eax+TButton._cursor], mcHand
        mov     [eax+TButton.__want_focus], TRUE

        return
endp



method TButton.ForceRefresh
begin
        mov     eax, [.self]
        and     [eax+TButton._canvas_valid], 0
        inherited
        return
endp


method TButton.SelfPaint
begin
        pushad
        mov     esi, [.self]

        cmp     [esi+TButton._canvas_valid], 0
        jne     @f
        exec    esi, TButton:_RenderView
@@:
        mov     ebx, [esi+TButton._canvas]
        test    ebx, ebx
        jz      .finish

        get     edx, esi, TButton:State
        imul    edx, [esi+TButton._height]
        add     edx, [.ySrc]

        stdcall BlendImage, [.pDstImage], [.xDst], [.yDst], ebx, [.xSrc], edx, [.width], [.height]

.finish:
        popad
        inherited [.pDstImage], [.xDst], [.yDst], [.xSrc], [.ySrc], [.width], [.height]
        return
endp



method TButton.Resize ; .newWidth, .newHeight
begin
        inherited [.newWidth], [.newHeight]
        pushad

        mov     esi, [.self]
        mov     ecx, [.newWidth]
        mov     edx, [.newHeight]
        shl     edx, 2

        mov     ebx, [esi+TButton._canvas]
        test    ebx, ebx
        jz      .create_new

        cmp     ecx, [ebx+TImage.width]
        jg      .destroy

        cmp     edx, [ebx+TImage.height]
        jle     .size_ok

.destroy:
        stdcall DestroyImage, ebx
        and     [esi+TButton._canvas], 0

.create_new:
        test    ecx, ecx
        jz      .size_ok

        test    edx, edx
        jz      .size_ok

        stdcall CreateImage, ecx, edx
        mov     [esi+TButton._canvas], eax

.size_ok:
        and     [esi+TButton._canvas_valid], 0
        exec    esi, TButton:RectChanged2, 0
        popad
        return
endp




method TButton.GetState
begin
        mov     eax, [.self]
        cmp     [eax+TButton._enabled], 0
        je      .disabled
        mov     eax, [eax+TButton._state]
        return

.disabled:
        mov     eax, bsDisabled
        return
endp



method TButton.SetCaption       ;, .value
begin
        inherited [.value]

        push    eax
        mov     eax, [.self]
        and     [eax+TButton._canvas_valid], 0
        pop     eax
        exec    [.self], TButton:RectChanged2, 0
        return
endp



method TButton.SetIcon;, .value
begin
        pushad

        mov     esi, [.self]
        mov     eax, [esi+TButton._icon]
        test    eax, eax
        jz      .load_new

        stdcall DestroyImage, eax
        xor     eax, eax

.load_new:

        cmp     [.value], 0
        je      @f

        stdcall CreateImagePNG, [.value], -1

@@:
        mov     [esi+TButton._icon], eax

        and     [esi+TButton._canvas_valid], 0
        exec    esi, TButton:RectChanged2, 0

        popad
        return
endp



method TButton.UpdateAction
begin
        exec    [.self], TButton:_RenderView
        exec    [.self], TWindow:RectChanged2, 0
        return
endp




method TButton._RenderView

 .bounds TBounds
 .client TBounds

 .clTxt dd ?
 .canvas dd ?

 .caption dd ?
 .enabled dd ?
 .icon    dd ?
 .icon_x  dd ?
 .icon_y  dd ?

begin
        pushad

        mov     esi, [.self]
        mov     eax, [esi+TButton._canvas]
        test    eax, eax
        jz      .finish

        mov     [.canvas], eax

        get     edx, esi, TButton:Action
        test    edx, edx
        jz      .get_normal

        get     [.caption], edx, TAction:Caption
        get     [.icon],    edx, TAction:Icon
        jmp     .render_states

.get_normal:

        get     [.caption], esi, TButton:Caption
        get     [.icon], esi, TButton:Icon

.render_states:

        xor     eax, eax        ; Y coordinate
        mov     ecx, [esi+TButton._width]
        mov     edx, [esi+TButton._height]

        mov     [.icon_x], eax
        mov     [.icon_y], eax

        mov     [.bounds.x], eax
        mov     [.client.x], eax
        mov     [.bounds.y], eax
        mov     [.client.y], eax
        mov     [.bounds.width], ecx
        mov     [.client.width], ecx
        mov     [.bounds.height], edx
        mov     [.client.height], edx

        mov     ebx, [GUI.btnMarginX]
        mov     edx, [GUI.btnMarginY]

        add     [.bounds.x], ebx
        add     [.bounds.y], edx
        sub     [.bounds.width], ebx
        sub     [.bounds.height], edx
        sub     [.bounds.width], ebx
        sub     [.bounds.height], edx

        mov     edi, [.icon]
        test    edi, edi
        jz      .icon_pos_ok

        mov     eax, [.bounds.width]
        sub     eax, [edi+TImage.width]
        sar     eax, 1
        add     eax, [.bounds.x]                ; x coordinate of the icon.

        mov     ecx, [.bounds.height]
        sub     ecx, [edi+TImage.height]
        sar     ecx, 1
        add     ecx, [.bounds.y]                ; y coordinate of the icon.

        cmp     [esi+TButton._iconpos], iposCenter
        je      .icon_pos_store

        cmp     [esi+TButton._iconpos], iposLeft
        jne     @f

; align left
        mov     eax, [.bounds.x]

        add     ebx, [edi+TImage.width]
        add     [.bounds.x], ebx
        sub     [.bounds.width], ebx
        jmp     .icon_pos_store

@@:
        cmp     [esi+TButton._iconpos], iposRight
        jne     @f

; align right
        mov     eax, [.bounds.x]
        add     eax, [.bounds.width]
        sub     eax, [edi+TImage.width]

        add     ebx, [edi+TImage.width]
        sub     [.bounds.width], ebx
        jmp     .icon_pos_store

@@:
        cmp     [esi+TButton._iconpos], iposTop
        jne     @f

; align top
        mov     ecx, [.bounds.y]
        add     edx, [edi+TImage.height]

        add     [.bounds.y], edx
        sub     [.bounds.height], edx
        jmp     .icon_pos_store

@@:
        cmp     [esi+TButton._iconpos], iposBottom
        jne     .icon_pos_ok

; align bottom
        mov     ecx, [.bounds.y]
        add     ecx, [.bounds.height]
        sub     ecx, [edi+TImage.height]

        add     edx, [edi+TImage.height]
        sub     [.bounds.height], edx

.icon_pos_store:

        mov     [.icon_x], eax
        mov     [.icon_y], ecx

.icon_pos_ok:

        xor     ebx, ebx        ; 0 == bsNeutral

.loop_y:

; draw the border and the background

        lea     eax, [.client]
        stdcall [DrawBox], [.canvas], eax, [GUI.clBtnBk+4*ebx], [GUI.btnBorder+4*ebx], [GUI.boxBorderWidth]

; draw the icon

        mov     edi, [.icon]
        test    edi, edi
        jz      .draw_text

        mov     ecx, [.icon_y]
        mov     eax, [.icon_x]
        add     ecx, [.client.y]

        cmp     ebx, bsPressed
        jne     .blend_icon

        add     eax, [GUI.btnPressedOfsX]
        add     ecx, [GUI.btnPressedOfsY]

.blend_icon:
        xor     edx, edx
        stdcall BlendImage, [.canvas], eax, ecx, edi, edx, edx, [edi+TImage.width], [edi+TImage.height]

.draw_text:

        get     edi, esi, TButton:TextAlign
        cmp     [.caption], 0
        je      .textok

        mov     eax, [.bounds.x]
        mov     ecx, [.bounds.y]

        cmp     ebx, bsPressed
        jne     @f
        add     eax, [GUI.btnPressedOfsX]
        add     ecx, [GUI.btnPressedOfsY]
@@:

        stdcall DrawTextBox, [.canvas], [.caption], eax, ecx, [.bounds.width], [.bounds.height], 2, edi, [GUI.DefaultFont], [GUI.clBtnTxt+4*ebx]

.textok:
        cmp     esi, [__FocusedWindow]
        jne     .focus_ok

        mov     eax, [.client.y]
        add     eax, [.client.height]
        sub     eax, 2
        stdcall DrawSolidRect, [.canvas], 0, eax, [esi+TButton._width], 2, [GUI.clBorderFocused]

.focus_ok:

        get     eax, esi, TWindow:height
        add     [.bounds.y], eax
        add     [.client.y], eax

        inc     ebx
        cmp     ebx, bsDisabled
        jbe     .loop_y

.finish:
        inc     [esi+TButton._canvas_valid]
        popad
        return
endp





method TButton.SimpleSetParam
begin
        push eax esi edi

        mov     esi, [.self]
        mov     edi, esi
        add     esi, [.get]

        mov     eax, [.value]
        cmp     eax, [esi]
        je      .finish

        mov     [esi], eax
        cmp     [.param], 0
        je      @f
        and     [edi+TButton._canvas_valid], 0
@@:
        exec    edi, TButton:RectChanged2, 0

.finish:
        pop     edi esi eax
        return
endp



method TButton.EventFocusIn
begin
        push    eax
        mov     eax, [.self]
        and     [eax+TButton._canvas_valid], 0
        exec    eax, TWindow:RectChanged2, 0
        pop     eax
        return
endp


method TButton.EventFocusOut
begin
        push    eax
        mov     eax, [.self]
        and     [eax+TButton._canvas_valid], 0
        exec    eax, TWindow:RectChanged2, 0
        pop     eax
        return
endp



method TButton.EventMouseEnter
begin
        exec    [.self], TButton:SimpleSetParam, 0, TButton._state, bsHovered
        inherited
        return
endp


method TButton.EventMouseLeave
begin
        exec    [.self], TButton:SimpleSetParam, 0, TButton._state, bsNeutral
        inherited
        return
endp


method TButton.EventMouseMove
begin
        exec    [.self], TButton:SimpleSetParam, 0, TButton._state, bsHovered
        inherited [.x], [.y], [.kbdState]
        return
endp


method TButton.EventButtonPress
begin
        exec    [.self], TButton:SimpleSetParam, 0, TButton._state, bsPressed
        return
endp



method TButton.EventButtonRelease
begin
        pushad

        mov     esi, [.self]

        cmp     [esi+TButton._state], bsPressed
        jne     .finish

        cmp     [.x], -1         ; when called from the keyboard event as a mouse emulation the .x and .y parameters are both -1.
        jne     .hover

        cmp     [.y], -1
        je      .state_ok

.hover:
        exec    [.self], TButton:SimpleSetParam, 0, TButton._state, bsHovered

.state_ok:
        cmp     [esi+TButton._ModalResult], mrNone
        je      @f

        get     eax, esi, TButton:Parent

        istype  eax, TForm
        jne     @f

        set     eax, TForm:ModalResult, [esi+TButton._ModalResult]

@@:
        get     eax, esi, TButton:Menu
        test    eax, eax
        jz      .menu_ok

        exec    eax, TMenu:Show, esi, 0, 0

.menu_ok:
        get     eax, esi, TButton:OnClick
        test    eax, eax
        jz      .click_ok

        stdcall eax, esi, [.button]       ; OnClick event

.click_ok:
        get     eax, esi, TButton:Action
        test    eax, eax
        jz      .finish

        exec    eax, TAction:Execute, [.self]

.finish:
        popad
        return
endp




method TButton.EventKeyPress
begin
        push    ecx

        cmp     [.utf8], ' '
        jne     .finish

        mov     ecx, [.self]
        exec    ecx, TButton:SimpleSetParam, 0, TButton._state, bsPressed
        exec    ecx, TWindow:EventButtonRelease, mbLeft, [.kbdState], -1, -1

.finish:
        inherited [.utf8], [.scancode], [.kbdState]
        pop     ecx
        return
endp



method TButton.EventKeyRelease
begin
        exec    [.self], TButton:SimpleSetParam, 0, TButton._state, bsNeutral
        return
endp



endmodule