; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TGrid object class
;
;  Target OS: Any
;
;  Dependencies:
;  Notes:
;_________________________________________________________________________________________
module "TGrid library"



object TGrid, TScrollable

; private fields

  ._canvas        dd ?    ; To be split later on 4 different canvases. One for fixed[0,0], one for fixed columns, one for the fixed rows and one for the regular cells.
                          ; this way the amount of drawing on refresh will be minimized.
  ._canvas_valid  dd ?


  ._captured dd ?

; properties
  .ColCount  dd ?
  .RowCount  dd ?
  .ColWidth  dd ?               ; default column width - every column can have different width though.
  .RowHeight dd ?               ; default row height - every row can have diferent height.

  .FixedCols dd ?               ; count of the fixed columns [0..FixedCols-1]
  .FixedRows dd ?               ; count of the fixed rows [0..FixedRows-1]
  ._flags    dd ?               ; grid flags ( dgf* flags)

  .ptrGridData dd ?             ; one pointer for the user.

  ._focused   POINT             ; coordinates of the focused cell.
  ._on_focus_change dd ?        ; event handler for focus changed.

; These 2 arrays contains respectively ColCount+1 and RowCount+1 elements.
  .ptrColX dd ?                 ; pointer to array of dwords with X coordinates of the columns.
  .ptrRowY dd ?                 ; pointer to array of dwords with Y coordinates of the rows.

  .selection RECT               ; rectangle with current selection.
  .sel_flags  dd ?              ; see gsfXXXXX constants below
  .resize_rc  dd ?              ; row or column to resize.
  .resize_ofs dd ?

; Parameters

  param    .Flags, ._flags, ._flags

  param    .OnFocusChange, ._on_focus_change, ._on_focus_change

; Abstract methods

  abstract .DrawCell, .canvas, .ptrRect, .Col, .Row

; Methods

  method .ForceRefresh
  method .__UpdateImages
  method .SelfPaint, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height    ; Paints itself on the caller provided TImage.

  method .__RenderGrid
  method .__AdjustScrollbars
  method .__AdjustOneScroller, .kind, .ptrArray, .Count, .Fixed, .client, .meansize

  method .Create, .parent
  method .Destroy

  method .Resize, .newW, .newH


  method .EnsureVisible, .Col, .Row
  method .HitTest, .x, .y
  method .GetCellXY, .Col, .Row
  method .SetCounts, .ColCount, .RowCount
  method .SetColWidth, .Col, .Width
  method .SetColumnsWidth, .BegCol, .EndCol, .Width
  method .SetRowHeight, .Row, .Height
  method .SetRowsHeight, .BegRow, .EndRow, .Height

  method .SetFocused, .x, .y

; system event handlers

  method .EventMouseMove, .x, .y, .kbdState
;  method .EventMouseLeave

  method .EventButtonPress, .button, .kbdState, .x, .y
  method .EventButtonRelease, .button, .kbdState, .x, .y
  method .EventScroll, .direction, .command, .value

  method .EventKeyPress, .utf8, .scancode, .kbdState

endobj


fcfUnfocused = 0
fcfFocused = 1

interface TGrid.OnFocusChange, .pGrid, .x, .y, .fFocused


; TGrid features flags

; implemented
gfGridLines     = 1
gfGridLinesFixed = 2
gfResizeColumns = 4
gfResizeRows    = 8
gfRowFocus      = $10
gfSingleCellSelect = $20

; not implemented
;gfSelectRows    = $10   ; clicks on the fixed column selects entire row. Else, click column event.
;gfSelectColumns = $20   ; clicks on the fixed row selects entire column. Else, click row event.
;gfSelectCells   = $40   ; rectangle of cells can be selected. Else, only focused cell is changed or
                        ; if gfSelectRows or gfSelectColumns is set, select the respective row or column.



; grid selections flags for the field TGrid.sel_flags

gsfSelectEntireRow = 1
gsfSelectEntireCol = 2
gsfResizeColWidth = 4
gsfResizeRowHeight = 8


; cell possible states:

cellRegular  = 0
cellFixed    = 1
cellSelected = 2
cellFocused  = 3


method TGrid.Create
begin
        inherited [.parent]
        push    eax esi

        mov     esi, [.self]

        stdcall GetMem, 4
        mov     [esi+TGrid.ptrColX], eax
        stdcall GetMem, 4
        mov     [esi+TGrid.ptrRowY], eax

; Default values for sizes.

        mov     [esi+TGrid.__want_focus], TRUE

        mov     [esi+TGrid.ColWidth], 64
        mov     [esi+TGrid.RowHeight], 20
        mov     [esi+TGrid._flags], gfResizeColumns or gfResizeRows or gfGridLinesFixed or gfGridLines   ;gfGridLinesFixed or gfGridLines or gfResizeColumns or gfResizeRows

        or      eax, -1
        mov     [esi+TGrid._focused.x], eax
        mov     [esi+TGrid._focused.y], eax
        mov     [esi+TGrid.selection.left], eax
        mov     [esi+TGrid.selection.top], eax
        mov     [esi+TGrid.selection.right], eax
        mov     [esi+TGrid.selection.bottom], eax

        exec    esi, TGrid:SetCounts, 1, 1
        pop     esi eax
        return
endp


method TGrid.Destroy
begin
        pushad
        mov     esi, [.self]
        stdcall FreeMem, [esi+TGrid.ptrColX]
        stdcall FreeMem, [esi+TGrid.ptrRowY]

        popad
        inherited
        return
endp





method TGrid.ForceRefresh
begin
        mov     eax, [.self]
        and     [eax+TGrid._canvas_valid], 0
        inherited
        return
endp



method TGrid.SelfPaint       ; .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height
begin
        pushad
        mov     esi, [.self]

        cmp     [esi+TGrid._canvas_valid], 0
        jne     @f
        exec    esi, TGrid:__RenderGrid
@@:
        mov     ebx, [esi+TGrid._canvas]
        test    ebx, ebx
        jz      .finish

        stdcall BlendImage, [.pDstImage], [.xDst], [.yDst], ebx, [.xSrc], [.ySrc], [.width], [.height]

.finish:
        popad
        inherited [.pDstImage], [.xDst], [.yDst], [.xSrc], [.ySrc], [.width], [.height]
        return
endp




; returns CF = 1 if the focus has been changed.

method TGrid.SetFocused ; .x, .y
begin
        pushad

        mov     esi, [.self]
        mov     ecx, [.x]
        mov     edx, [.y]
        mov     edi, [esi+TGrid._on_focus_change]

        mov     eax, [esi+TGrid.ColCount]
        mov     ebx, [esi+TGrid.RowCount]
        dec     eax
        dec     ebx

        cmp     ecx, eax
        cmova   ecx, eax

        cmp     edx, ebx
        cmova   edx, ebx

        cmp     ecx, [esi+TGrid._focused.x]
        jne     .changeit

        cmp     edx, [esi+TGrid._focused.y]
        je      .finish

.changeit:
        test    edi, edi
        jz      .old_ok

        pushad
        stdcall edi, esi, [esi+TGrid._focused.x], [esi+TGrid._focused.y], fcfUnfocused
        popad

.old_ok:
        mov     [esi+TGrid._focused.x], ecx
        mov     [esi+TGrid._focused.y], edx
        test    edi, edi
        jz      .end_ok

        stdcall edi, esi, ecx, edx, fcfFocused

.end_ok:
        stc

.finish:
        popad
        return
endp



method TGrid.__UpdateImages
begin
        pushad
        mov     esi, [.self]

        mov     ebx, [esi+TGrid._visible]
        mov     edi, [esi+TGrid._canvas]

        mov     ecx, [esi+TGrid._width]
        mov     edx, [esi+TGrid._height]

        test    ebx, ebx
        jz      .destroy

        test    edi, edi
        jz      .create_new

; check the size

        cmp     ecx, [edi+TImage.width]
        ja      .destroy

        cmp     edx, [edi+TImage.height]
        jbe     .setwrap

.destroy:
        stdcall DestroyImage, edi                         ; canvas

        xor     eax, eax
        mov     [esi+TGrid._canvas], eax

        test    ebx, ebx
        jz      .exit


.create_new:

; new size - make it a little bit bigger.

        lea     eax, [ecx+256]
        lea     ebx, [edx+256]

        stdcall CreateImage, eax, ebx
        mov     [esi+TGrid._canvas], eax
        mov     edi, eax

.setwrap:
        mov     [edi+TImage.wrapW], ecx
        mov     [edi+TImage.wrapH], edx

.exit:
        and     [esi+TGrid._canvas_valid], 0
        popad
        inherited
        return
endp




method TGrid.Resize
begin
        inherited [.newW], [.newH]
        exec    [.self], TGrid:__AdjustScrollbars
        return
endp



method TGrid.EventScroll        ; , .direction, .command, .value
begin
        pushad
        mov     esi, [.self]

        cmp     [.command], scTrack
        je      .refresh

        mov     ebx, [.direction]
        shl     ebx, 2

        lea     edi, [esi+TGrid._scrX + 4*ebx]

        mov     ecx, [.value]
        imul    ecx, [esi+TGrid.ColWidth + ebx]

        cmp     [.command], scWheelUp
        jne     @f
        neg     ecx
@@:
        mov     eax, [edi+TScroller.pos]
        add     eax, ecx
        jns     @f
        xor     eax, eax
@@:
        cmp     eax, [edi+TScroller.max]
        cmova   eax, [edi+TScroller.max]

        mov     [edi+TScroller.pos], eax
        sub     edi, esi
        exec    esi, TGrid:__DrawScrollbar, edi

.refresh:
        and     [esi+TGrid._canvas_valid], 0
        exec    esi, TGrid:RectChanged2, 0

        popad
        return
endp




method TGrid.EventMouseMove     ;, .x, .y, .kbdState
begin
        inherited [.x], [.y], [.kbdState]
        jc        .processed

        pushad
        mov     esi, [.self]

; first the mouse cursor:

        mov     ebx, mcArrow

        exec    esi, TGrid:HitTest, [.x], [.y]
        cmovc   eax, [esi+TGrid.selection.right]
        cmovc   edx, [esi+TGrid.selection.bottom]

        test    eax, eax
        jns     @f
        mov     ebx, mcSizeV
@@:
        test    edx, edx
        jns     @f
        mov     ebx, mcSizeH
@@:

        cmp     [esi+TGrid.sel_flags], 0
        jne     .mouse_cursor_ok

        push    eax edx

        mov     [esi+TWindow._cursor], ebx
        stdcall GetStockCursor, ebx

        get     ecx, esi, TGrid:OSWindow
        stdcall SetMouseCursor, [ecx+TWindow.handle], eax

        pop     edx eax

        test    eax, eax
        js      .end_mouse_move
        test    edx, edx
        js      .end_mouse_move

.mouse_cursor_ok:
        cmp     [esi+TGrid._captured], 0
        je      .end_mouse_move

; here process the drag operations:

        mov     edi, [esi+TGrid.sel_flags]

        test    edi, gsfResizeColWidth or gsfResizeRowHeight
        jnz     .resize_row_col

        cmp     edi, gsfSelectEntireRow
        cmove   eax, [esi+TGrid.ColCount]

        cmp     edi, gsfSelectEntireCol
        cmove   edx, [esi+TGrid.RowCount]

        mov     [esi + TGrid.selection.right], eax
        mov     [esi + TGrid.selection.bottom], edx

        cmp     edi, gsfSelectEntireRow
        cmove   eax, [esi+TGrid.FixedCols]

        cmp     edi, gsfSelectEntireCol
        cmove   edx, [esi+TGrid.FixedRows]

        exec    esi, TGrid:SetFocused, eax, edx
        jnc     .end_mouse_move

.ensure:
        exec    esi, TGrid:EnsureVisible, [esi+TGrid.selection.right], [esi+TGrid.selection.bottom]
        and     [esi+TGrid._canvas_valid], 0
        exec    esi, TGrid:RectChanged2, 0

.end_mouse_move:
        popad

.processed:
        return


.resize_row_col:

        test    edi, gsfResizeColWidth
        jz      .col_width_ok

        mov     ebx, [esi+TGrid.ptrColX]
        mov     eax, [esi+TGrid.resize_rc]
        mov     ecx, [.x]
        sub     ecx, [ebx+4*eax]        ; the start coordinate.
        cmp     eax, [esi+TGrid.FixedCols]
        jb      .xpos_ok
        add     ecx, [esi+TGrid._scrX.pos]
.xpos_ok:
        sub     ecx, [esi+TGrid.resize_ofs]
        cmp     ecx, 4
        jge     @f
        mov     ecx, 4
@@:
        exec    esi, TGrid:SetColWidth, eax, ecx

        mov     eax, [esi+TGrid.resize_rc]
        xor     edx, edx
        jmp     .ensure

.col_width_ok:
        test    edi, gsfResizeRowHeight
        jz      .end_mouse_move

        mov     ebx, [esi+TGrid.ptrRowY]
        mov     eax, [esi+TGrid.resize_rc]
        mov     ecx, [.y]
        sub     ecx, [ebx+4*eax]        ; the start coordinate.
        cmp     eax, [esi+TGrid.FixedCols]
        jb      .ypos_ok
        add     ecx, [esi+TGrid._scrY.pos]
.ypos_ok:
        sub     ecx, [esi+TGrid.resize_ofs]
        cmp     ecx, 4
        jge     @f
        mov     ecx, 4
@@:
        exec    esi, TGrid:SetRowHeight, eax, ecx

        mov     edx, [esi+TGrid.resize_rc]
        xor     eax, eax
        jmp     .ensure
endp




method TGrid.EventButtonPress        ;, .button, .kbdState, .x, .y
begin
        inherited [.button], [.kbdState], [.x], [.y]
        jc        .processed

        pushad
        cmp     [.button], mbLeft
        jne     .finish

        mov     esi, [.self]

        exec    esi, TGrid:HitTest, [.x], [.y]
        jc      .finish

        stdcall SetMouseCapture, [.self]
        mov     [esi+TGrid._captured], 1

        test    eax, eax
        js      .resize_row
        test    edx, edx
        js      .resize_col

        and     [esi+TGrid.sel_flags], 0
        mov     [esi+TGrid.selection.left], eax
        mov     [esi+TGrid.selection.top], edx
        mov     [esi+TGrid.selection.right], eax
        mov     [esi+TGrid.selection.bottom], edx

        cmp     eax, [esi+TGrid.FixedCols]
        jge     .notleft

; if the cursor is on the left fixed cells
        or      [esi+TGrid.sel_flags], gsfSelectEntireRow
        mov     ebx, [esi+TGrid.FixedCols]  ; xbeg
        mov     ecx, [esi+TGrid.ColCount]   ; xend
        mov     [esi+TGrid.selection.left], ebx
        mov     [esi+TGrid.selection.right], ecx
        mov     eax, [esi+TGrid.FixedCols]

.notleft:
        cmp     edx, [esi+TGrid.FixedRows]
        jge     .nottop

; if the cursor is on the top fixed cells
        or      [esi+TGrid.sel_flags], gsfSelectEntireCol
        mov     ebx, [esi+TGrid.FixedRows]  ; ybeg
        mov     ecx, [esi+TGrid.RowCount]   ; yend
        mov     [esi+TGrid.selection.top], ebx
        mov     [esi+TGrid.selection.bottom], ecx
        mov     edx, [esi+TGrid.FixedCols]

.nottop:

        exec    esi, TGrid:SetFocused, eax, edx

                and     [esi+TGrid._canvas_valid], 0
        exec    esi, TGrid:RectChanged2, 0

.finish:
        popad

.processed:
        return


.resize_row:
        mov     [esi+TGrid.sel_flags], gsfResizeRowHeight
        mov     eax, edx
        mov     ecx, [esi+TGrid.ptrRowY]
        mov     edx, [.y]
        cmp     eax, [esi+TGrid.FixedRows]
        jb      .do_resize
        add     edx, [esi+TGrid._scrY.pos]
        jmp     .do_resize

.resize_col:
        mov     [esi+TGrid.sel_flags], gsfResizeColWidth
        mov     ecx, [esi+TGrid.ptrColX]
        mov     edx, [.x]
        cmp     eax, [esi+TGrid.FixedCols]
        jb      .do_resize
        add     edx, [esi+TGrid._scrX.pos]

.do_resize:
        mov     [esi+TGrid.resize_rc], eax
        sub     edx, [ecx+4*eax+4]      ; Y coordinate on the next cell.
        mov     [esi+TGrid.resize_ofs], edx
        jmp     .finish
endp





method TGrid.EventButtonRelease      ;, .button, .kbdState, .x, .y
begin
        inherited [.button], [.kbdState], [.x], [.y]
        pushad

        mov     esi, [.self]
        cmp     [esi+TGrid._captured], 0
        je      .finish

        stdcall SetMouseCapture, 0
        mov     [esi+TGrid._captured], 0

        and     [esi+TGrid.sel_flags], 0
        lea     edi, [esi+TGrid.selection]
        stdcall RectSort, edi

        and     [esi+TGrid._canvas_valid], 0
        exec    esi, TGrid:RectChanged2, 0

.finish:
        popad
        return
endp





method TGrid.EventKeyPress    ;, .utf8, .scancode, .kbdState
begin
        cmp     [.utf8], 0
        jne     .finish

        pushad

        mov     esi, [.self]
        mov     eax, [.scancode]

        mov     ecx, [esi+TGrid._focused.x]
        mov     edx, [esi+TGrid._focused.y]

        cmp     eax, keyUpNumpad
        je      .up
        cmp     eax, keyUp
        je      .up

        cmp     eax, keyDownNumpad
        je      .down
        cmp     eax, keyDown
        je      .down

        cmp     eax, keyPgUpNumpad
        je      .pageup
        cmp     eax, keyPgUp
        je      .pageup

        cmp     eax, keyPgDnNumpad
        je      .pagedown
        cmp     eax, keyPgDown
        je      .pagedown

        cmp     eax, keyLeftNumpad
        je      .left
        cmp     eax, keyLeft
        je      .left

        cmp     eax, keyRightNumpad
        je      .right
        cmp     eax, keyRight
        je      .right

        cmp     eax, keyHomeNumpad
        je      .home
        cmp     eax, keyHome
        je      .home

        cmp     eax, keyEndNumpad
        je      .end
        cmp     eax, keyEnd
        je      .end


.key_end:
        popad

.finish:
        inherited [.utf8], [.scancode], [.kbdState]
        return


.xy_ok:
        mov     [esi+TGrid.selection.right], ecx
        mov     [esi+TGrid.selection.bottom], edx

        test    [.kbdState], maskShift
        jnz     .selection_ok

        mov     [esi+TGrid.selection.left], ecx
        mov     [esi+TGrid.selection.top], edx

.selection_ok:

        exec    esi, TGrid:SetFocused, ecx, edx
        jnc     .key_end

        exec    esi, TGrid:EnsureVisible, [esi+TGrid._focused.x], [esi+TGrid._focused.y]
        exec    esi, TGrid:__AdjustScrollbars

        xor     eax, eax
        mov     [esi+TGrid._canvas_valid], eax
        exec    esi, TGrid:RectChanged2, eax
        jmp     .key_end

;------------------------------------------------------------------

.up:
        dec     edx
        cmp     edx, [esi+TGrid.FixedRows]
        cmovl   edx, [esi+TGrid.FixedRows]
        jmp     .xy_ok

.down:
        mov     eax, [esi+TGrid.RowCount]
        dec     eax

        inc     edx
        cmp     edx, eax
        cmovg   edx, eax
        jmp     .xy_ok


.left:
        dec     ecx
        cmp     ecx, [esi+TGrid.FixedCols]
        cmovl   ecx, [esi+TGrid.FixedCols]
        jmp     .xy_ok

.right:
        mov     eax, [esi+TGrid.ColCount]
        dec     eax

        inc     ecx
        cmp     ecx, eax
        cmovg   ecx, eax
        jmp     .xy_ok


.home:
        mov     edx, [esi+TGrid.FixedRows]
        jmp     .xy_ok


.end:
        mov     edx, [esi+TGrid.RowCount]
        dec     edx
        jmp     .xy_ok


; It is a little bit more complex.

.pageup:
.pagedown:
        jmp     .xy_ok



endp





method TGrid.__RenderGrid
.rcFixedTL RECT
.rcFixedT  RECT
.rcFixedL  RECT
.rcFloat   RECT
.rcClient  RECT
.ptrRect   dd ?
begin
        pushad

        mov     esi, [.self]

        lea     eax, [.rcClient]
        mov     [.ptrRect], eax

        mov     eax, [esi+TGrid._width]
        mov     ecx, [esi+TGrid._height]

        xor     edx, edx

        mov     [.rcClient.left], edx
        mov     [.rcClient.top], edx
        mov     [.rcClient.right], eax
        mov     [.rcClient.bottom], ecx

        mov     [.rcFixedT.right], eax
        mov     [.rcFixedL.bottom], ecx
        mov     [.rcFloat.right], eax
        mov     [.rcFloat.bottom], ecx

        mov     [.rcFixedTL.left], edx
        mov     [.rcFixedTL.top], edx
        mov     [.rcFixedT.top], edx
        mov     [.rcFixedL.left], edx

        mov     ecx, [esi+TGrid.FixedCols]
        mov     eax, [esi+TGrid.ptrColX]
        test    eax, eax
        jnz     .colsizeok

        imul    ecx, [esi+TGrid.ColWidth]   ; default width if there is no size array
        jmp     .x1ok

.colsizeok:
        mov     ecx, [eax+4*ecx]
.x1ok:
        dec     ecx
        mov     edx, [esi+TGrid.FixedRows]
        mov     eax, [esi+TGrid.ptrRowY]
        test    eax, eax
        jnz     .rowsizeok

        imul    edx, [esi+TGrid.ColWidth]   ; default width if there is no size array
        jmp     .y1ok

.rowsizeok:
        mov     edx, [eax+4*edx]
.y1ok:
        dec     edx
        mov     [.rcFixedTL.right], ecx
        mov     [.rcFixedTL.bottom], edx
        mov     [.rcFixedT.left], ecx
        mov     [.rcFixedT.bottom], edx
        mov     [.rcFixedL.right], ecx
        mov     [.rcFixedL.top], edx
        mov     [.rcFloat.left], ecx
        mov     [.rcFloat.top], edx

; Here all 4 rectangles are ready. Some of them can be empty of course.
; Lets make intersections with the invalidated rectangle.

.drawFloat:
        lea     eax, [.rcFloat]
        stdcall RectIntersect, eax, eax, [.ptrRect]
        jc      .drawFixedT

        lea     eax, [.rcFloat]
        stdcall TGrid.__DrawCellArea, [esi+TGrid._canvas], eax, 0

.drawFixedT:
        lea     eax, [.rcFixedT]
        stdcall RectIntersect, eax, eax, [.ptrRect]
        jc      .drawFixedL

        stdcall TGrid.__DrawCellArea, [esi+TGrid._canvas], eax, ffVertical

.drawFixedL:
        lea     eax, [.rcFixedL]
        stdcall RectIntersect, eax, eax, [.ptrRect]
        jc      .drawFixedTL

        stdcall TGrid.__DrawCellArea, [esi+TGrid._canvas], eax, ffHorizontal

.drawFixedTL:
        lea     eax, [.rcFixedTL]
        stdcall RectIntersect, eax, eax, [.ptrRect]
        jc      .finish

        stdcall TGrid.__DrawCellArea, [esi+TGrid._canvas], eax, ffHorizontal or ffVertical

.finish:
        popad
        return
endp




; .fFixed contains flags:
ffHorizontal = 1
ffVertical   = 2

proc TGrid.__DrawCellArea, .canvas, .ptrClipRect, .fFixed
  .Cell  RECT
  .xofs  dd ?
  .yofs  dd ?
  .clip  RECT
  .sel   RECT
  .bnd   TBounds
begin
        pushad

; selection rectangle in cells coordinates.

        lea     edi, [.sel]
        lea     eax, [esi+TGrid.selection]
        stdcall RectCopy, edi, eax
        stdcall RectSort, edi
        inc     [.sel.right]
        inc     [.sel.bottom]

        mov     eax, [esi+TGrid.FixedRows]
        mov     ecx, [esi+TGrid.FixedCols]

; The fixed cells can not be selected!!!
; [edi+RECT.top] := max([edi+RECT.top], [esi+TGrid.FixedRows])
; [edi+RECT.left] := max([edi+RECT.left], [esi+TGrid.FixedCols])

        cmp     eax, [edi+RECT.top]
        cmovl   eax, [edi+RECT.top]

        cmp     ecx, [edi+RECT.left]
        cmovl   ecx, [edi+RECT.left]

        mov     [edi+RECT.top], eax
        mov     [edi+RECT.left], ecx

        mov     eax, [esi+TGrid.FixedCols]
        mov     ecx, [esi+TGrid.ColCount]

        test    [esi+TGrid._flags], gfRowFocus
        jz      @f

        mov     [edi+RECT.left], eax
        mov     [edi+RECT.right], ecx

@@:

; drawing clip rectange.
        lea     eax, [.bnd]
        stdcall RectToBounds, eax, [.ptrClipRect]

; erase the background
        cmp     [.fFixed], 0
        cmove   eax, [GUI.clGridCellBk + 4*cellRegular]
        cmovne  eax, [GUI.clGridCellBk + 4*cellFixed]

        mov     edi, [.ptrClipRect]

        mov     ecx, [edi+RECT.right]
        mov     edx, [edi+RECT.bottom]
        sub     ecx, [edi+RECT.left]
        sub     edx, [edi+RECT.top]
        stdcall DrawSolidRect, [.canvas], [edi+RECT.left], [edi+RECT.top], ecx, edx, eax

; If the cells are not fixed, compute the real coordinates of the rectangle.
        xor     eax, eax
        mov     [.xofs], eax
        mov     [.yofs], eax

        test    [.fFixed], ffHorizontal
        jnz     .horizOK

        mov     eax, [esi+TGrid._scrX.pos]

        mov     [.xofs], eax
        add     [edi+RECT.left], eax
        add     [edi+RECT.right], eax

.horizOK:
        test    [.fFixed], ffVertical
        jnz     .vertOK

        mov     eax, [esi+TGrid._scrY.pos]

        mov     [.yofs], eax
        add     [edi+RECT.top], eax
        add     [edi+RECT.bottom], eax

; Determine the cell range for the given clip rectangle.
.vertOK:
        xor     eax, eax
        mov     ebx, [esi+TGrid.ColCount]
        dec     ebx
        stdcall __SearchCoordinate, [edi+RECT.left], [esi+TGrid.ptrColX], 0, ebx
        mov     [.Cell.left], eax
;        OutputValue "Cell.left: ", eax, 10, -1

        mov     eax, [esi+TGrid.ColCount]
        dec     eax
        stdcall __SearchCoordinate, [edi+RECT.right], [esi+TGrid.ptrColX], 0, eax
        mov     [.Cell.right], eax
;        OutputValue "Cell.right: ", eax, 10, -1

        xor     eax, eax
        mov     ebx, [esi+TGrid.RowCount]
        dec     ebx
        stdcall  __SearchCoordinate, [edi+RECT.top], [esi+TGrid.ptrRowY], 0, ebx
        mov     [.Cell.top], eax
;        OutputValue "Cell.top: ", eax, 10, -1

        mov     eax, [esi+TGrid.RowCount]
        dec     eax
        stdcall __SearchCoordinate, [edi+RECT.bottom], [esi+TGrid.ptrRowY], 0, eax
        mov     [.Cell.bottom], eax
;        OutputValue "Cell.bottom: ", eax, 10, -1

        mov     ebx, [.Cell.top]

.yloop:
        cmp     ebx, [.Cell.bottom]
        jg      .endyloop

        mov     eax, [esi+TGrid.ptrRowY]
        mov     ecx, [4*ebx+eax+4]              ; bottom
        mov     eax, [4*ebx+eax]                ; top
        sub     ecx, [.yofs]
        sub     eax, [.yofs]

        mov     [.clip.bottom], ecx
        mov     [.clip.top], eax

        mov     edi, [.Cell.left]

.xloop:
        cmp     edi, [.Cell.right]
        jg      .endxloop

        mov     eax, [esi+TGrid.ptrColX]
        mov     ecx, [4*edi+eax+4]
        mov     eax, [4*edi+eax]
        sub     ecx, [.xofs]
        sub     eax, [.xofs]

        mov     [.clip.right], ecx
        mov     [.clip.left], eax

; draw the background if the cell is selected or focused:

        cmp     ebx, [esi+TGrid._focused.y]
        jne     .is_selected

        test    [esi+TGrid._flags], gfRowFocus
        jnz     .cell_focused

        cmp     edi, [esi+TGrid._focused.x]
        jne     .is_selected

.cell_focused:
        mov     edx, [GUI.clGridCellBk+4*cellFocused]
        jmp     .draw_bk

.is_selected:

        lea     eax, [.sel]
        stdcall PointInRect, eax, edi, ebx
        jnc     .bkgroundok

        mov     edx, [GUI.clGridCellBk + 4*cellSelected]

.draw_bk:
        mov     eax, [.clip.right]
        mov     ecx, [.clip.bottom]
        sub     eax, [.clip.left]
        sub     ecx, [.clip.top]
        stdcall DrawSolidRect, [.canvas], [.clip.left], [.clip.top], eax, ecx, edx

.bkgroundok:
        lea     eax, [.clip]
        exec    esi, TGrid:DrawCell, [.canvas], eax, edi, ebx

; draw the gridlines:

        push    ebx

        xor     ebx, ebx
        mov     eax, gfGridLines

        cmp     [.fFixed], FALSE
        je      .datagridline

        mov     ebx, 8
        mov     eax, gfGridLinesFixed

.datagridline:
;        OutputValue "Flags: ", [esi+TGrid._flags], 16, 8

        test    [esi+TGrid._flags], eax
        jz      .gridok

        mov     edx, [.clip.left]
        mov     ecx, [.clip.right]
        mov     eax, [.clip.bottom]
        dec     ecx
        sub     eax, [.clip.top]
        stdcall BlendSolidRect, [.canvas], edx, [.clip.top], 1, eax, [GUI.clGridLines+ebx]   ; left
        stdcall BlendSolidRect, [.canvas], ecx, [.clip.top], 1, eax, [GUI.clGridLines+4+ebx] ; right

        mov     edx, [.clip.top]
        mov     ecx, [.clip.right]
        mov     eax, [.clip.bottom]
        sub     ecx, [.clip.left]
        dec     eax
        stdcall BlendSolidRect, [.canvas], [.clip.left], edx, ecx, 1, [GUI.clGridLines+ebx]    ; top
        stdcall BlendSolidRect, [.canvas], [.clip.left], eax, ecx, 1, [GUI.clGridLines+4+ebx]  ; bottom

.gridok:
        pop     ebx

        inc     edi
        jmp     .xloop

.endxloop:
        inc     ebx
        jmp     .yloop

.endyloop:

        popad
        return
endp






;------------------------------------------------------------------
; If .Col, .Row is inside the visible client area, does nothing.
; , if .Col, .Row is outside this area, provides scrolling to
; make it visible to the center of the client area.
;------------------------------------------------------------------
method TGrid.EnsureVisible      ;, .Col, .Row
.client RECT
begin
        pushad
        mov     edi, [.self]

        mov     ebx, [edi+TGrid.FixedCols]
        mov     ecx, [edi+TGrid.FixedRows]
        shl     ebx, 2
        shl     ecx, 2
        add     ebx, [edi+TGrid.ptrColX]
        add     ecx, [edi+TGrid.ptrRowY]

        mov     ebx, [ebx]
        mov     ecx, [ecx]

        mov     [.client.left], ebx
        mov     [.client.top], ecx
        mov     ebx, [edi+TGrid._width]
        mov     ecx, [edi+TGrid._height]
        mov     [.client.right], ebx
        mov     [.client.bottom], ecx

        mov     eax, [.Col]
        cmp     eax, [edi+TGrid.FixedCols]
        jb      .posvertical                    ; fixed columns are always inside the screen

        get     ecx, edi, TGrid:PosX
        add     [.client.left], ecx
        add     [.client.right], ecx

        mov     edx, [edi+TGrid.ptrColX]
        mov     ecx, [.Col]
        mov     ebx, [edx+4*ecx]
        mov     ecx, [edx+4*ecx+4]

        cmp     ecx, [.client.right]
        jle     .rightok

        sub     ecx, [.client.right]
        add     ecx, [.client.left]

        get     eax, edi, TGrid:PosX
        sub     eax, [.client.left]
        add     ecx, eax

        set     edi, TGrid:PosX, ecx
        jmp     .posvertical

.rightok:
        cmp     ebx, [.client.left]
        jge     .posvertical

        get     eax, edi, TGrid:PosX
        sub     eax, [.client.left]
        add     ebx, eax

        set     edi, TGrid:PosX, ebx

.posvertical:
        mov     eax, [.Row]
        cmp     eax, [edi+TGrid.FixedRows]
        jb      .endpos

        get     ecx, edi, TGrid:PosY
        add     [.client.top], ecx
        add     [.client.bottom], ecx

        mov     edx, [edi+TGrid.ptrRowY]
        mov     ecx, [.Row]
        mov     ebx, [edx+4*ecx]
        mov     ecx, [edx+4*ecx+4]

        cmp     ecx, [.client.bottom]
        jle     .bottomok

        sub     ecx, [.client.bottom]
        add     ecx, [.client.top]

        get     eax, edi, TGrid:PosY
        sub     eax, [.client.top]
        add     ecx, eax

        set     edi, TGrid:PosY, ecx
        jmp     .endpos

.bottomok:
        cmp     ebx, [.client.top]
        jge     .endpos

        get     eax, edi, TGrid:PosY
        sub     eax, [.client.top]
        add     ebx, eax

        set     edi, TGrid:PosY, ebx

.endpos:
        popad
        return
endp



; Return:
;   CF = 1: no cell on the given coordinates.
;   CF = 0: a cell has been found on these coordinates.
;   eax: column pointed or -1.
;   edx: row pointed or -1
;   If eax = -1 then the cursor is in the row resize area for the row in edx
;   if edx = -1 then the cursor is in the column resize area for the column in eax

method TGrid.HitTest ; .x, .y
begin
        pushad

        mov     esi, [.self]

        stdcall __SearchCoordinate, [.y], [esi+TGrid.ptrRowY], 0, [esi+TGrid.FixedRows]
        jc      .col_not_resize

        mov     ebx, eax

        mov     edi, [.x]
        stdcall __SearchCoordinate, edi, [esi+TGrid.ptrColX], 0, [esi+TGrid.FixedCols]
        jnc     .maybe_col_resize

        add     edi, [esi+TGrid._scrX.pos]
        stdcall __SearchCoordinate, edi, [esi+TGrid.ptrColX], [esi+TGrid.FixedCols], [esi+TGrid.ColCount]
        jc      .cell_not_found

.maybe_col_resize:

        test    [esi+TGrid._flags], gfResizeColumns
        jz      .col_select

        sub     edx, edi
        sub     edi, ecx

        cmp     edx, 3
        jle     .col_resize

        cmp     edi, 2
        jg      .col_select

        dec     eax

.col_resize:
        mov     [esp+4*regEAX], eax
        mov     dword [esp+4*regEDX], -1

        clc
        popad
        return

.col_select:
        cmp     eax, [esi+TGrid.FixedCols]
        jb      .col_not_resize

        mov     [esp+4*regEAX], eax
        mov     [esp+4*regEDX], ebx

        clc
        popad
        return

.cell_not_found:
        stc
        popad
        return

; the Y coordinate is not on the fixed rows, so it is either row resize or simple data cell selection.

.col_not_resize:

        stdcall __SearchCoordinate, [.x], [esi+TGrid.ptrColX], 0, [esi+TGrid.FixedCols]
        jc      .select_data_cell

        mov     ebx, eax

        mov     edi, [.y]
        stdcall __SearchCoordinate, edi, [esi+TGrid.ptrRowY], 0, [esi+TGrid.FixedRows]
        jnc     .maybe_row_resize

        add     edi, [esi+TGrid._scrY.pos]
        stdcall __SearchCoordinate, edi, [esi+TGrid.ptrRowY], [esi+TGrid.FixedRows], [esi+TGrid.RowCount]
        jc      .cell_not_found

.maybe_row_resize:
        test    [esi+TGrid._flags], gfResizeRows
        jz      .row_select

        sub     edx, edi
        sub     edi, ecx

        cmp     edx, 3
        jle     .row_resize

        cmp     edi, 2
        jg      .row_select

        dec     eax

.row_resize:
        mov     [esp+4*regEDX], eax
        mov     dword [esp+4*regEAX], -1
        clc
        popad
        return


.select_data_cell:
        mov     edi, [.x]
        add     edi, [esi+TGrid._scrX.pos]
        stdcall __SearchCoordinate, edi, [esi+TGrid.ptrColX], [esi+TGrid.FixedCols], [esi+TGrid.ColCount]
        jc      .cell_not_found

        mov     ebx, eax

        mov     edi, [.y]
        add     edi, [esi+TGrid._scrY.pos]
        stdcall __SearchCoordinate, edi, [esi+TGrid.ptrRowY], [esi+TGrid.FixedRows], [esi+TGrid.RowCount]
        jc      .cell_not_found

.row_select:
        mov     [esp+4*regEAX], ebx
        mov     [esp+4*regEDX], eax
        clc
        popad
        return


endp






method TGrid.GetCellXY  ;, .Col, .Row
begin
        push    esi

        mov     esi, [.self]
        mov     eax, [.Col]
        mov     edx, [.Row]
        shl     eax, 2
        shl     edx, 2
        add     eax, [esi+TGrid.ptrColX]
        add     edx, [esi+TGrid.ptrRowY]
        mov     eax, [eax]
        mov     edx, [edx]

        pop     esi
        return
endp



method TGrid.SetCounts  ;, .ColCount, .RowCount
.ycount dd ?
.ptrNewColX  dd ?
.ptrNewRowY  dd ?
begin
        pushad

        mov     [.ptrNewColX], 0
        mov     [.ptrNewRowY], 0
        mov     ebx, [.self]

; resize column coordinates array...
        mov     eax, [ebx+TGrid.ptrColX]
        mov     [.ptrNewColX], eax

        mov     eax, [.ColCount]
        cmp     eax, [ebx+TGrid.ColCount]
        je      .resizey

        mov     eax, [.ColCount]
        lea     eax, [4*eax+4]          ; byte size of the array.
        stdcall ResizeMem, [ebx+TGrid.ptrColX], eax
        mov     [.ptrNewColX], eax
        test    eax, eax
        jz      .finish

        mov     edi, eax

        mov     ecx, [ebx+TGrid.ColCount]
        cmp     ecx, [.ColCount]
        jge     .resizey

        mov     eax, [edi+4*ecx]

.newsizesx:
        inc     ecx
        cmp     ecx, [.ColCount]
        jg      .resizey

        add     eax, [ebx+TGrid.ColWidth]
        mov     [edi+4*ecx], eax
        jmp     .newsizesx

.resizey:
; resize row coordinates array...
        mov     eax, [ebx+TGrid.ptrRowY]
        mov     [.ptrNewRowY], eax

        mov     eax, [.RowCount]
        cmp     eax, [ebx+TGrid.RowCount]
        je      .finish

        mov     eax, [.RowCount]
        lea     eax, [4*eax+4]          ; byte size of the array.
        stdcall ResizeMem,[ebx+TGrid.ptrRowY], eax
        mov     [.ptrNewRowY], eax
        test    eax, eax
        jz      .finish

        mov     edi, eax

        mov     ecx, [ebx+TGrid.RowCount]
        cmp     ecx, [.RowCount]
        jg      .finish

        mov     eax, [edi+4*ecx]

.newsizesy:
        inc     ecx
        cmp     ecx, [.RowCount]
        jg      .finish

        add     eax, [ebx+TGrid.RowHeight]
        mov     [edi+4*ecx], eax
        jmp     .newsizesy

.finish:
        xor     eax, eax
        cmp     eax, [.ptrNewColX]
        je      .error1
        cmp     eax, [.ptrNewRowY]
        je      .error2

        mov     ecx, [.RowCount]
        mov     edx, [.ColCount]
        mov     [ebx+TGrid.RowCount], ecx
        mov     [ebx+TGrid.ColCount], edx

        mov     ecx, [.ptrNewColX]
        mov     edx, [.ptrNewRowY]
        mov     [ebx+TGrid.ptrColX], ecx
        mov     [ebx+TGrid.ptrRowY], edx

        exec    ebx, TGrid:__AdjustScrollbars

        clc
        popad
        return

.error2:
        stdcall FreeMem, [.ptrNewColX]

.error1:
        stc
        popad
        return
endp


method TGrid.SetColWidth;, .Col, .Width
begin
        pushad
        mov     edi, [.self]
        mov     eax, [edi+TGrid.ptrColX]
        mov     ecx, [.Col]

        lea     eax, [eax+4*ecx+4]

        mov     edx, [.Width]
        sub     ecx, [edi+TGrid.ColCount]
        jge     .finish         ; error, the index is greater than count.

        sub     edx, [eax]
        neg     ecx
        add     edx, [eax-4]

.loop:
        add     [eax], edx
        lea     eax, [eax+4]
        loop    .loop

        exec    edi, TGrid:__AdjustOneScroller, sbHorz, [edi+TGrid.ptrColX], [edi+TGrid.ColCount], [edi+TGrid.FixedCols], [edi+TGrid._width], [edi+TGrid.ColWidth]

.finish:
        popad
        return
endp



method TGrid.SetColumnsWidth    ;, .BegCol, .EndCol, .Width
begin
        pushad

        mov     edi, [.self]
        mov     ebx, [.BegCol]
        cmp     ebx, [edi+TGrid.ColCount]
        jae     .endloop

        mov     eax, [edi+TGrid.ColCount]
        cmp     [.EndCol], eax
        jl      .do_resize

        dec     eax
        mov     [.EndCol], eax

.do_resize:
        mov     esi, [edi+TGrid.ptrColX]
        mov     ecx, [esi+4*ebx]        ; the start X coordinate.

.loop:
        add     ecx, [.Width]
        inc     ebx
        cmp     ebx, [.EndCol]
        ja      .endloop

        mov     [esi+4*ebx], ecx
        jmp     .loop

.endloop:
        sub     ecx, [esi+4*ebx]   ; the delta.

.loop2:
        cmp     ebx, [edi+TGrid.ColCount]
        ja      .endloop2

        add     [esi+4*ebx], ecx
        inc     ebx
        jmp     .loop2

.endloop2:

        exec    edi, TGrid:__AdjustOneScroller, sbHorz, [edi+TGrid.ptrColX], [edi+TGrid.ColCount], [edi+TGrid.FixedCols], [edi+TGrid._width], [edi+TGrid.ColWidth]

        popad
        return
endp



method TGrid.SetRowHeight       ;, .Row, .Height
begin
        pushad
        mov     edi, [.self]
        mov     eax, [edi+TGrid.ptrRowY]
        mov     ecx, [.Row]

        lea     eax, [eax+4*ecx+4]

        mov     edx, [.Height]
        sub     ecx, [edi+TGrid.RowCount]
        jge     .finish         ; error, the index is greater than count.

        sub     edx, [eax]
        neg     ecx
        add     edx, [eax-4]

.loop:
        add     [eax], edx
        lea     eax, [eax+4]
        loop    .loop

        exec    edi, TGrid:__AdjustOneScroller, sbVert, [edi+TGrid.ptrRowY], [edi+TGrid.RowCount], [edi+TGrid.FixedRows], [edi+TGrid._height], [edi+TGrid.RowHeight]

.finish:
        popad
        return
endp



method TGrid.SetRowsHeight      ;.BegRow, .EndRow, .Height
begin
        pushad

        mov     edi, [.self]
        mov     ebx, [.BegRow]
        cmp     ebx, [edi+TGrid.RowCount]
        jae     .endloop

        mov     eax, [edi+TGrid.RowCount]
        cmp     [.EndRow], eax
        jl      .do_resize

        dec     eax
        mov     [.EndRow], eax

.do_resize:
        mov     esi, [edi+TGrid.ptrRowY]
        mov     ecx, [esi+4*ebx]        ; the start Y coordinate.

.loop:
        add     ecx, [.Height]
        inc     ebx
        cmp     ebx, [.EndRow]
        ja      .endloop

        mov     [esi+4*ebx], ecx
        jmp     .loop

.endloop:
        sub     ecx, [esi+4*ebx]   ; the delta.

.loop2:
        cmp     ebx, [edi+TGrid.RowCount]
        ja      .endloop2

        add     [esi+4*ebx], ecx
        inc     ebx
        jmp     .loop2

.endloop2:

        exec    edi, TGrid:__AdjustOneScroller, sbVert, [edi+TGrid.ptrRowY], [edi+TGrid.RowCount], [edi+TGrid.FixedRows], [edi+TGrid._height], [edi+TGrid.RowHeight]

        popad
        return
endp



; Arguments:
;  .x : some coordinate.
;  .xarray: array or dword with coordinates of the cells.
;  .from: the count of the array. It contains N+1 elements where N is the number of cells in the table.
;  .to: from which cell to search.
; returns:
;   eax - the index of the cell where the coordinate is located.
;   ecx - the left coordinate of the cell.
;   edx - the right coordinate of the cell+1 (the left coordinate of the next cell).

proc __SearchCoordinate, .x, .xarray, .from, .to
begin
        pushad

        mov     esi, [.xarray]
        mov     eax, [.from]   ; left
        mov     edx, [.to]     ; right

        cmp     eax, edx
        jge     .not_found

        mov     ebx, [.x]
        cmp     ebx, [esi+4*eax]
        jl      .not_found      ; it is less than the first element.

        cmp     ebx, [esi+4*edx]
        jge     .not_found      ; it is greater than the last column.

; binary search between

.loop:
        cmp     eax, edx
        je      .found

        lea     ecx, [eax+edx]
        shr     ecx, 1

        cmp     ebx, [esi+4*ecx]
        cmovl   edx, ecx
        jl      .loop

        cmp     ebx, [esi+4*ecx+4]
        cmovge  eax, ecx
        jge     .loop

.found:
        mov     [esp+4*regEAX], ecx     ; the index of the
        mov     eax, [esi+4*ecx]
        mov     edx, [esi+4*ecx+4]
        mov     [esp+4*regECX], eax
        mov     [esp+4*regEDX], edx
        clc
        popad
        return

.not_found:
        stc
        popad
        return
endp



;------------------------------------------------------------------------
; TGrid private methods.
; These functions are for internal use. Don't call them outside TGrid
; control.
;------------------------------------------------------------------------

;---------------------------------------------------------
; Auto adjusts the scrollbars
;---------------------------------------------------------

sbHorz = 0
sbVert = 1

method TGrid.__AdjustScrollbars
begin
        push    ebx

        mov     ebx, [.self]
        exec    ebx, TGrid:__AdjustOneScroller, sbVert, [ebx+TGrid.ptrRowY], [ebx+TGrid.RowCount], [ebx+TGrid.FixedRows], [ebx+TGrid._height], [ebx+TGrid.RowHeight]
        exec    ebx, TGrid:__AdjustOneScroller, sbHorz, [ebx+TGrid.ptrColX], [ebx+TGrid.ColCount], [ebx+TGrid.FixedCols], [ebx+TGrid._width], [ebx+TGrid.ColWidth]

        pop     ebx
        return
endp


method TGrid.__AdjustOneScroller        ;, .kind, .ptrArray, .Count, .Fixed, .client, .meansize
begin
        pushad

        mov     ebx, [.self]

        mov     esi, [.ptrArray]
        test    esi, esi
        jz      .finish

        mov     ecx, [.Count]
        mov     edx, [.Fixed]

        mov     ecx, [esi+4*ecx]        ; total height of all cells
        mov     edx, [esi+4*edx]        ; height of the fixed cells

        sub     ecx, [.client]
        jns     @f
        xor     ecx, ecx
@@:

        mov     eax, [.client]          ; scroller page size
        sub     edx, eax
        neg     edx
        jns     @f
        xor     edx, edx
@@:

        cmp     [.kind], sbHorz
        jne     .vert

; horizonal
        get     eax, ebx, TGrid:PosX
        cmp     eax, ecx
        cmova   eax, ecx

        set     ebx, TGrid:PosX, eax
        set     ebx, TGrid:MaxX, ecx
        set     ebx, TGrid:PageX, edx

.finish:
        popad
        return

.vert:
        get     eax, ebx, TGrid:PosY
        cmp     eax, ecx
        cmova   eax, ecx

        set     ebx, TGrid:PosY, eax
        set     ebx, TGrid:MaxY, ecx
        set     ebx, TGrid:PageY, edx
        jmp     .finish

endp



endmodule



