; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TWindow object class
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: TWindow is the common ancestor of all visible on the screen controls.
;         This file contains only OS independent part of the library and includes
;         the OS dependent files.
;_________________________________________________________________________________________
module "TWindow library"

; Interfaces for the OS dependent procedures

; NOTICE: Some of them are not needed, so can be removed.


;-----------------------------------------------------------------------------------------
; Creates OS window, according to the content of TWindow structure [.pWindow]
;
; Notice, that this OS windows are created ONLY for the visible top level windows.
;
;-----------------------------------------------------------------------------------------
interface _CreateWindow, .pWindow

;-----------------------------------------------------------------------------------------
; Destroy the OS window. The OS windows are actually destroyed when invisible and then
; recreated when needed.
;-----------------------------------------------------------------------------------------
interface _DestroyWindow, .hwnd

;-----------------------------------------------------------------------------------------
; Converts the client coordinates to screen coordinates.
;-----------------------------------------------------------------------------------------
interface _ClientToScreen, .hwnd, .x, .y

;-----------------------------------------------------------------------------------------
; Changes the visibility of the window.
;-----------------------------------------------------------------------------------------
interface _ShowWindow, .hwnd, .flag


interface _RefreshWindowRect, .hwnd, .x, .y, .width, .height
interface _SetWindowTextUtf8, .hwnd, .ptrUtf8
interface _SetModalTowards, .hwnd, .hwndParent
interface _FinalizeModal, .hwnd, .hwndParent
interface _GetWindowStruct, .hwin
interface _SetWindowStruct, .hwin, .value


; probably not needed, but will keep them for some time.

;interface _SetWindowBorder, .hwnd, .brdType
;interface _GetWindowBounds, .hwnd, .pBounds
;interface _SetWindowBounds, .hwnd, .pBounds
;interface _CreateNullWindow
;interface _GetParent, .hwnd
;interface _GetChildren, .hwnd
;interface _GetVisible, .hwnd
;interface _SetFocus, .hwnd
;interface _AddChild, .hwnd, .child
;interface _GetWindowTextUtf8, .hwnd, .ptrUtf8
;interface _EnableWindow, .hwnd, .flag
;interface _SetTransientTowards, .hwnd, .hwndParent


include '%TargetOS%/windows.asm'



; Border kind of window.
borderNone    = 0
borderFull    = 1
borderModal   = 2
borderToolbox = 3

; SelfPaint operations

object TWindow, TObject

  .handle         dd ?   ; it is handle to the system provided window.

  ._x             dd ?
  ._y             dd ?
  ._width         dd ?
  ._height        dd ?

  ._border        dd ?    ; has meaning only for the top level windows!

  ._cursor        dd ?    ; mouse cursor.

  ._caption       dd ?    ; string handle with window caption.

  ._enabled       dd ?

  ._visible       dd ?

  .__want_focus   dd ?

  .__action       dd ?

  .__main         dd ?    ; this is the window that is main for this window. has meaning only for top level windows.

; window screen image

  ._screen        dd ?    ; TImage containing the image of the window with all children composited.
  ._rect_list     dd ?    ; array of rectangles containing all the changed rectangles of the window.

  ._user_data     dd ?

  if defined options.Threads & options.Threads
    ._rect_mx       TMutex
  end if

; window user event handlers

  ._OnSelfPaint    dd ?
  ._OnScroll       dd ?
  ._OnKeyPressed   dd ?

; parameters

  param .OSWindow, .GetOSWindow, NONE           ; the first window down the tree, that has window handle.

  param .Action, .__action, .SetAction

  param .x,      ._x,      .SetX
  param .y,      ._y,      .SetY
  param .width,  ._width,  .SetWidth
  param .height, ._height, .SetHeight
  param .SplitCell, NONE, .SetSplitCell

  param .border, ._border, ._border

  param ._MainWindow, .__main, .__main

  param .Cursor, ._cursor, ._cursor

  param .Visible, ._visible, .SetVisible
  param .Caption, .GetCaption, .SetCaption

  param .Enabled, .GetEnabled, .SetEnabled
  param .GlobalEnabled, .GetGlobalEnabled, NONE

  param .WantFocus, .__want_focus, .__want_focus

  param .Screen, .GetImgScreen, NONE

; Event handlers

  param .OnSelfPaint, ._OnSelfPaint, .SetSelfPaint

  param .OnScroll, ._OnScroll, ._OnScroll
  param .OnKeyPressed, ._OnKeyPressed, ._OnKeyPressed

  param .UserData, ._user_data, ._user_data

; private methods

  method .__UpdateHandle
  method .__UpdateImages
  method .__UpdateFocus


; parameter methods

  method .SetParent, .value

  method .SetAction, .value

  abstract .UpdateAction, .action    ; to be defined only for controls that can be attached to TAction

  method .SetX, .value
  method .SetY, .value
  method .SetWidth, .value
  method .SetHeight, .value

  method .Move, .newX, .newY
  method .Resize, .newWidth, .newHeight

  method .GetEnabled
  method .SetEnabled, .value

  method .GetGlobalEnabled

  method .GetCaption
  method .SetCaption, .value

  method .SetVisible, .value
  method .SetSelfPaint, .value

  method .SetSplitCell, .value

  method .GetImgScreen

  method .GetOSWindow


; other methods

  method .Create, .parent
  method .Destroy


; Parent/child methods definitions of abstract methods, inherited from TObject

  method .ChildByXY, .x, .y, .onlyenabled

  method .RectChanged2, .pRect     ; invalidates the rectangle, so all children windows in this rectangle are
                                   ; to be redrawn. Called from the children window on any visual changes.
                                   ; notice, that if the child changed the size ot position, it needs special
                                   ; processing of the rectangle. Actually the changed rectangle is the
                                   ; bounding rectangle of the new and old rectangles.

  method .RectChangedXY2, .x, .y, .width, .height
;  method .RectChangedList, .pRectList

  method .ForceRefresh  ; Use this method only for serious refresh tasks, for example on global theme change.

  method  .ClientToScreenXY, .x, .y

; Painting methods

  method .SelfPaint, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height    ; Paints itself on the caller provided TImage.

; Layout methods

  method .Autosize      ; Returns in EAX the prefered width and in EDX the prefered height of the Window
                        ; By default (in TWindow) returns the width and height of the window
                        ; Called by the splitgrid layout library.

; System event methods

  method .EventFocusIn
  method .EventFocusOut

  method .EventMouseEnter
  method .EventMouseLeave

  method .EventMouseMove, .x, .y, .kbdState

  method .EventButtonPress, .button, .kbdState, .x, .y
  method .EventButtonRelease, .button, .kbdState, .x, .y

  method .EventScroll, .direction, .command, .value

  method .EventKeyPress, .utf8, .scancode, .kbdState
  method .EventKeyRelease, .scancode, .kbdState

  method .CloseRequest, .reason

endobj


interface TWindow.OnSelfPaint, .self, .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height
interface TWindow.OnScroll, .self, .direction, .command, .value
interface TWindow.OnKeyPressed, .self, .utf8, .scancode, .kbdState



; OS independent code.
;_________________________________________________________________________________________



method TWindow.Create
begin
        inherited [.parent]

        set     [.self], TWindow:Enabled, TRUE
        set     [.self], TWindow:Cursor, mcArrow
        set     [.self], TWindow:border, borderFull

if defined options.Threads & options.Threads
        push    esi
        mov     esi, [.self]
        lea     esi, [esi+TWindow._rect_mx]
        stdcall MutexCreate, 0, esi
        stdcall MutexRelease, esi
        pop     esi
end if
        return
endp


;_________________________________________________________________________________________



method TWindow.Destroy
begin
        pushad

        mov     esi, [.self]

; hide the window, destroy the handle and remove the focus.

        set     esi, TWindow:Visible, FALSE

        cmp     esi, [__LastPointedWindow]
        jne     @f
        mov     [__LastPointedWindow], 0
@@:

; detach from action if any

        mov     ecx, [esi+TWindow.__action]
        jecxz   @f

        exec    ecx, TAction:Detach, esi

@@:

; destroy the screen images if any. And here they must be 0 anyway.

        xor     ecx, ecx
        xchg    ecx, [esi+TWindow._screen]
        jecxz   .screen_ok

        stdcall DestroyImage, ecx

.screen_ok:

; free the string handles

        stdcall StrDel, [esi+TWindow._caption]

if defined options.Threads & options.Threads
        lea     eax, [esi+TWindow._rect_mx]
        stdcall MutexDestroy, eax
end if

        popad

        inherited
        return
endp


;_________________________________________________________________________________________

; Paints itself on the caller provided TImage.
; The base implementation in TWindow simply calls the user event handler TWindow:OnPaint and
; renders the children windows on the provided TImage/rectangle.
; The implementations for the controls can use any method of paint, but must to call
; "inherited" at the end;

method TWindow.SelfPaint  ; .pDstImage, .xDst, .yDst, .xSrc, .ySrc, .width, .height
.src  RECT
.dest RECT
.ofsx dd ?
.ofsy dd ?
begin
        pushad

        mov     esi, [.self]
        stdcall LockImg, [.pDstImage]

        mov     eax, [esi+TWindow._OnSelfPaint]
        test    eax, eax
        jz      .user_event_processed

        push    esi
        stdcall eax, esi, [.pDstImage], [.xDst], [.yDst], [.xSrc], [.ySrc], [.width], [.height]
        pop     esi

.user_event_processed:

; The source rectangle:

        mov     ecx, [.xDst]
        mov     edx, [.yDst]
        sub     ecx, [.xSrc]
        sub     edx, [.ySrc]
        mov     [.ofsx], ecx    ; .ofsx and .ofsy actually are the X and Y position of the top-left corner
        mov     [.ofsy], edx    ; of the window on the destination image.

        mov     ecx, [.xDst]
        mov     edx, [.yDst]
        mov     [.src.left], ecx
        mov     [.src.top], edx
        add     ecx, [.width]
        add     edx, [.height]
        mov     [.src.right], ecx
        mov     [.src.bottom], edx
        lea     edi, [.src]

; then blend the children windows

        get     ebx, esi, TObject:Children
        test    ebx, ebx
        jz      .end_loop

        or      ecx, -1

.children_loop:

        inc     ecx
        cmp     ecx, [ebx+TArray.count]
        je      .end_loop

        mov     esi, [ebx+4*ecx+TArray.array]

        istype  esi, TWindow
        jne     .children_loop

        get     eax, esi, TWindow:Visible
        test    eax, eax
        jz      .children_loop

        mov     eax, [esi+TWindow._x]
        mov     edx, [esi+TWindow._y]
        add     eax, [.ofsx]
        add     edx, [.ofsy]
        mov     [.dest.left], eax
        mov     [.dest.top], edx
        add     eax, [esi+TWindow._width]
        add     edx, [esi+TWindow._height]
        mov     [.dest.right], eax
        mov     [.dest.bottom], edx

        lea     eax, [.dest]
        stdcall RectIntersect, eax, eax, edi
        jc      .children_loop                  ; the rectangles does not intersect, so the child should not to be painted.

        mov     eax, [.dest.left]
        mov     edx, [.dest.top]
        sub     [.dest.right], eax      ; source rectangle width
        sub     [.dest.bottom], edx     ; source rectangle height

        sub     eax, [esi+TWindow._x]
        sub     edx, [esi+TWindow._y]
        sub     eax, [.ofsx]
        sub     edx, [.ofsy]

        exec    esi, TWindow:SelfPaint, [.pDstImage], [.dest.left], [.dest.top], eax, edx, [.dest.right], [.dest.bottom]
        jmp     .children_loop

.end_loop:
        stdcall UnlockImg, [.pDstImage]
        popad
        return
endp



;_________________________________________________________________________________________

method TWindow.GetImgScreen
.client RECT
.rect   RECT
begin
        pushad

        xor     eax, eax
        mov     esi, [.self]

        cmp     [esi+TWindow._screen], eax
        je      .finish_screen

        stdcall LockImg, [esi+TWindow._screen]
        jc      .finish_screen

if defined options.Threads & options.Threads
        lea     eax, [esi+TWindow._rect_mx]
        stdcall WaitForMutex, eax, 1000
        jc      .finish_unlock
end if

        mov     edi, [esi+TWindow._rect_list]   ; list of the rectangles that need rendering.
        test    edi, edi
        jz      .finish_release                 ; no changes, the screen is up to date.

        mov     ecx, [edi+TArray.count]
        test    ecx, ecx
        jz      .finish_free

        lea     edi, [edi+TArray.array]

        xor     eax, eax
        mov     ebx, [esi+TWindow._width]
        mov     edx, [esi+TWindow._height]

        mov     [.client.left], eax
        mov     [.client.top], eax
        mov     [.client.right], ebx
        mov     [.client.bottom], edx

        lea     ebx, [.client]
        lea     edx, [.rect]

.rect_loop:
        stdcall RectIntersect, edx, ebx, edi
        jc      .next_rect

        mov     eax, [edx+RECT.left]
        sub     [edx+RECT.right], eax

        mov     eax, [edx+RECT.top]
        sub     [edx+RECT.bottom], eax

        exec    esi, TWindow:SelfPaint, [esi+TWindow._screen], [edx+RECT.left], [edx+RECT.top], [edx+RECT.left], [edx+RECT.top], [edx+RECT.right], [edx+RECT.bottom]

.next_rect:

        add     edi, sizeof.RECT
        loop    .rect_loop

.finish_free:

        xor     eax, eax
        xchg    [esi+TWindow._rect_list], eax

        stdcall FreeMem, eax

.finish_release:

if defined options.Threads & options.Threads
        lea     eax, [esi+TWindow._rect_mx]
        stdcall MutexRelease, eax
end if

.finish_unlock:

        stdcall UnlockImg, [esi+TWindow._screen]

.finish_screen:
        mov     eax, [esi+TWindow._screen]
        mov     [esp+4*regEAX], eax
        popad
        return
endp

;_________________________________________________________________________________________



method TWindow.SetAction
begin
        pushad

        mov     esi, [.self]
        mov     ecx, [.value]
        mov     edx, [esi+TWindow.__action]

        cmp     ecx, edx
        je      .finish

        mov     [esi+TWindow.__action], ecx

        test    edx, edx
        jz      .old_ok

        exec    edx, TAction:Detach, esi

.old_ok:
        test    ecx, ecx
        jz      .new_ok

        exec    ecx, TAction:Attach, esi

.new_ok:
        getm    eax, esi, TWindow:UpdateAction
        test    eax, eax
        jz      .finish

        stdcall eax, esi, ecx

.finish:
        popad
        return
endp


;_________________________________________________________________________________________


method TWindow.GetCaption
begin
        mov     eax, [.self]
        mov     eax, [eax+TWindow._caption]
        test    eax, eax
        jz      .finish
        stdcall StrDup, eax
.finish:
        return
endp


;_________________________________________________________________________________________


method TWindow.SetCaption
begin
        push    eax esi

        mov     esi, [.self]

        lea     eax, [esi+TWindow._caption]
        stdcall SetString, eax, [.value]
        stdcall StrPtr, [eax]

        cmp     [esi+TWindow.handle], 0
        je      @f

        stdcall _SetWindowTextUtf8, [esi+TWindow.handle], eax

@@:
        pop     esi eax
        return
endp


;_________________________________________________________________________________________



method TWindow.SetVisible
begin
        push    eax esi

        mov     esi, [.self]

        mov     eax, [.value]
        cmp     eax, [esi+TWindow._visible]
        je      .finish

        mov     [esi+TWindow._visible], eax

        exec    esi, TWindow:__UpdateHandle
        exec    esi, TWindow:__UpdateImages
        exec    esi, TWindow:__UpdateFocus

        exec    esi, TWindow:RectChanged2, 0

.finish:
        pop     esi eax
        return
endp


;_________________________________________________________________________________________

method TWindow.SetSelfPaint
begin
        push    esi eax

        mov     esi, [.self]
        mov     eax, [.value]
        cmp     eax, [esi+TWindow._OnSelfPaint]
        je      .finish

        mov     [esi+TWindow._OnSelfPaint], eax
        exec    esi, TWindow:RectChanged2, 0            ; repaint on change of the paint handler.

.finish:
        pop     eax esi
        return
endp

;_________________________________________________________________________________________


method TWindow.SetSplitCell
begin
        push    eax

        get     eax, [.self], TWindow:Parent
        istype  eax, TForm
        jne     .finish

        exec    eax, TForm:AttachControlToCell, [.self], [.value]

.finish:
        pop     eax
        return
endp


;_________________________________________________________________________________________


method TWindow.GetGlobalEnabled
begin
        push    ecx

        mov     ecx, [.self]

        get     eax, ecx, TWindow:Action
        test    eax, eax
        jz      .action_ok

        get     eax, eax, TAction:Enabled
        jmp     .check_parent

.action_ok:
        mov     eax, [ecx+TWindow._enabled]

.check_parent:
        test    eax, eax
        jz      .finish

        get     ecx, [.self], TWindow:Parent
        jecxz   .finish

        get     eax, ecx, TWindow:GlobalEnabled

.finish:
        pop     ecx
        return
endp



;_________________________________________________________________________________________



method TWindow.GetEnabled
begin
        push    ecx

        mov     ecx, [.self]

        get     eax, ecx, TWindow:Action
        test    eax, eax
        jz      .action_ok

        get     eax, eax, TAction:Enabled
        jmp     .finish

.action_ok:
        mov     eax, [ecx+TWindow._enabled]

.finish:
        pop     ecx
        return
endp




method TWindow.SetEnabled
begin
        push    ecx eax
        mov     ecx, [.self]

        mov     eax, [.value]
        xchg    eax, [ecx+TWindow._enabled]
        cmp     eax, [ecx+TWindow._enabled]
        je      .finish

        exec    ecx, TWindow:__UpdateFocus

.finish:
        pop     eax ecx
        return
endp





method TWindow.GetOSWindow
begin
        push    esi

        mov     esi, [.self]

.loop:
        cmp     [esi+TWindow.handle], 0
        jne     .finish

        get     esi, esi, TWindow:Parent
        test    esi, esi
        jnz     .loop

.finish:
        mov     eax, esi
        pop     esi
        return
endp


;_________________________________________________________________________________________

;---------------------------------------------------------------------------------
; Arguments:
;   .self
;   .x
;   .y
;   .onlyenabled
;
; Returns:
;   eax - pointer to TWindow
;   ecx - X coordinates, translated to the client coordinates of the child.
;   edx - Y coordinates, translated to the client coordinates of the child.
;
; If children window was not found, the method returns:
; eax = [.self]
; ecx = [.x]
; edx = [.y]
;---------------------------------------------------------------------------------

method TWindow.ChildByXY
begin
        push    ebx esi edi

        mov     eax, [.self]
        mov     ecx, [.x]
        mov     edx, [.y]

        mov     esi, [eax+TObject.__pChildren]
        test    esi, esi
        jz      .finish

        mov     edi, [esi+TArray.count]
        lea     esi, [esi+TArray.array]


.loop:
        dec     edi
        js      .finish

        mov     ebx, [esi+4*edi]

        istype  ebx, TWindow
        jne     .loop

        cmp     [ebx+TWindow._visible], 0
        je      .loop

        sub     ecx, [ebx+TWindow._x]
        jl      .no2

        sub     edx, [ebx+TWindow._y]
        jl      .no

        cmp     ecx, [ebx+TWindow._width]
        jge     .no

        cmp     edx, [ebx+TWindow._height]
        jge      .no

        cmp     [.onlyenabled], FALSE
        je      .enabled_ok

        push    eax
        get     eax, ebx, TWindow:GlobalEnabled
        test    eax, eax
        pop     eax
        jz      .no


.enabled_ok:
        pushad
        stdcall CreateImage, 1, 1
        mov     esi, eax

        exec    ebx, TWindow:SelfPaint, esi, 0, 0, ecx, edx, 1, 1
        mov     eax, [esi+TImage.pPixels]
        movzx   eax, byte [eax+3]

        stdcall DestroyImage, esi

        test    eax, eax
        popad
        jnz     .yes

.no:
        add     edx, [ebx+TWindow._y]
.no2:
        add     ecx, [ebx+TWindow._x]
        jmp     .loop

.yes:
        mov     eax, ebx
        exec    eax, TWindow:ChildByXY, ecx, edx, [.onlyenabled]

.finish:
        pop     edi esi ebx
        return

endp




;______________________________________________________________________________________________
;
; Includes the passed rectangle in the invalid rectangles list. Later this rectangle will need
; to be repainted, but only when the screen image is required.
;

method TWindow.RectChanged2      ;, .pRect

.prect  RECT
.rect   RECT
.client RECT

begin
        pushad

        mov     esi, [.self]

        mov     ecx, [.pRect]
        jecxz   .prect_ok

        lea     eax, [.prect]
        stdcall RectCopy, eax, [.pRect]
        mov     [.pRect], eax

.prect_ok:
        get     ecx, esi, TWindow:width
        get     edx, esi, TWindow:height
        xor     eax, eax

        mov     [.client.left], eax
        mov     [.client.top], eax
        mov     [.client.right], ecx
        mov     [.client.bottom], edx
        lea     edi, [.client]

        mov     eax, [.pRect]
        test    eax, eax
        jz      .rect_ok

        mov     eax, edi
        lea     edi, [.rect]
        stdcall RectIntersect, edi, eax, [.pRect]
        jc      .notify_parent

.rect_ok:
        cmp     [esi+TWindow._screen], 0
        je      .notify_parent

if defined options.Threads & options.Threads
        lea     eax, [esi+TWindow._rect_mx]
        stdcall WaitForMutex, eax, 1000
        jc      .finish
end if

        mov     edx, [esi+TWindow._rect_list]
        test    edx, edx
        jnz     .list_ok

        stdcall CreateArray, sizeof.RECT
        mov     edx, eax

.list_ok:
        stdcall RectListAdd, edx, edi

.rect_added_ok:
        mov     [esi+TWindow._rect_list], edx

if defined options.Threads & options.Threads
        lea     eax, [esi+TWindow._rect_mx]
        stdcall MutexRelease, eax
end if

; now send some redraw events if needed.

        mov     ecx, [esi+TWindow.handle]
        jecxz   .notify_parent

        mov     eax, [edi+RECT.right]  ;[.rect.right]
        mov     edx, [edi+RECT.bottom] ;[.rect.bottom]
        sub     eax, [edi+RECT.left]   ;[.rect.left]
        sub     edx, [edi+RECT.top]    ;[.rect.top]

        stdcall _RefreshWindowRect, ecx, [edi+RECT.left], [edi+RECT.top], eax, edx
        jmp     .finish  ; only root windows have handle, so no need to notify further.

; now do the same for the parent window

.notify_parent:

        get     ebx, esi, TWindow:Parent
        test    ebx, ebx
        jz      .finish

        get     ecx, esi, TWindow:x         ; shift the rectangle for the parent coordinates.
        get     edx, esi, TWindow:y

        mov     edi, [.pRect]
        lea     eax, [.client]
        test    edi, edi
        cmovz   edi, eax

        add     [edi+RECT.left], ecx
        add     [edi+RECT.top], edx
        add     [edi+RECT.right], ecx
        add     [edi+RECT.bottom], edx

        exec    ebx, TWindow:RectChanged2, edi

.finish:
        popad
        return
endp




method TWindow.RectChangedXY2
begin
        push    eax

        mov     eax, [.x]
        add     [.width], eax

        mov     eax, [.y]
        add     [.height], eax

        lea     eax, [.x]
        exec    [.self], TWindow:RectChanged2, eax

        pop     eax
        return
endp



method TWindow.ForceRefresh
begin
        pushad
        mov     esi, [.self]

        exec    esi, TWindow:RectChanged2, 0

        get     ebx, esi, TObject:Children
        test    ebx, ebx
        jz      .end_loop

        mov     ecx, [ebx+TArray.count]
        lea     ebx, [ebx+TArray.array]

.children_loop:
        dec     ecx
        js      .end_loop

        istype  [ebx+4*ecx], TWindow
        jne     .children_loop

        exec    [ebx+4*ecx], TWindow:ForceRefresh
        jmp     .children_loop

.end_loop:
        popad
        return
endp




;method TWindow.RectChangedList  ;, .pRectList
;begin
;        pushad
;        mov     esi, [.self]
;        mov     edx, [.pRectList]
;        mov     ecx, [edx+TArray.count]
;        lea     edx, [edx+TArray.array]
;
;.loop:
;        dec     ecx
;        js      .end_loop
;
;        exec    esi, TWindow:RectChanged2, edx
;        add     edx, sizeof.RECT
;        jmp     .loop
;
;.end_loop:
;        popad
;        return
;endp



;_________________________________________________________________________________________



method TWindow.SetParent
begin
        inherited [.value]
        exec    [.self], TWindow:__UpdateHandle      ; create/destroy handle of the OS window, according to the  window state.
        return
endp


;_________________________________________________________________________________________



method TWindow.__UpdateHandle
begin
        pushad

        mov     esi, [.self]

        xor     ecx, ecx

        get     eax, esi, TWindow:Visible
        test    eax, eax
        jz      .no_handle

        get     eax, esi, TWindow:Parent
        test    eax, eax
        jnz     .no_handle

; need handle

        cmp     [esi+TWindow.handle], ecx
        jne     .finish                         ; the window needs OS handle and has one.

; The window has no handle, but needs it.


        stdcall _CreateWindow, esi
        mov     [esi+TWindow.handle], eax

;        OutputValue "Window created:", eax, 16, 8

.finish:
        popad
        return


.no_handle:

        cmp     [esi+TWindow.handle], ecx
        je      .finish                         ; no need for handle and there is no such.

        stdcall _DestroyWindow, [esi+TWindow.handle]    ; destroy not needed window handle.

        mov     [esi+TWindow.handle], ecx
        jmp     .finish

endp




method TWindow.__UpdateFocus
begin
        mov     eax, [.self]
        cmp     eax, [__FocusedWindow]
        jne     .finish

        get     eax, [.self], TWindow:Visible
        test    eax, eax
        jz      .clear

        get     eax, [.self], TWindow:Enabled
        test    eax, eax
        jz      .clear

        get     eax, [.self], TWindow:WantFocus
        test    eax, eax
        jz      .clear

.finish:
        return

.clear:
        mov     [__FocusedWindow], 0
        exec    [.self], TWindow:EventFocusOut
        return
endp


SCREEN_MORE_X = 256
SCREEN_MORE_Y = 256


method TWindow.__UpdateImages
begin
        pushad

        xor     eax, eax
        mov     esi, [.self]

        mov     ebx, [esi+TWindow._visible]
        mov     edi, [esi+TWindow._screen]
        get     eax, esi, TWindow:Parent

        mov     ecx, [esi+TWindow._width]
        mov     edx, [esi+TWindow._height]

        xor     ebx, 1          ; ebx = FALSE, if the window is visible

        test    eax, eax
        cmovnz  ebx, eax        ; if the window has parent, it should have no screen image!

        test    edi, edi
        jz      .check_hidden

        test    ebx, ebx
        jnz     .destroy

; check the size

        cmp     ecx, [edi+TImage.width]
        ja      .destroy

        cmp     edx, [edi+TImage.height]
        ja      .destroy

.set_wrap:

        stdcall SetImageWrap, edi, ecx, edx

.exit:
        popad
        return


.destroy:
        stdcall DestroyImage, edi                         ; screen

        xor     eax, eax
        mov     [esi+TWindow._screen], eax

.check_hidden:

        test    ebx, ebx
        jnz     .exit


.create_new:

        lea     eax, [ecx + SCREEN_MORE_X]
        lea     ebx, [edx + SCREEN_MORE_Y]

        stdcall CreateImage, eax, ebx
        mov     edi, eax
        mov     [esi+TWindow._screen], edi

        exec    esi, TWindow:RectChanged2, 0
        jmp     .set_wrap

endp







;_________________________________________________________________________________________
;
; Geometry methods.


method TWindow.SetX
begin
        push    eax
        mov     eax, [.self]
        exec    eax, TWindow:Move, [.value], [eax+TWindow._y]
        pop     eax
        return
endp



method TWindow.SetY
begin
        push    eax
        mov     eax, [.self]
        exec    eax, TWindow:Move, [eax+TWindow._x], [.value]
        pop     eax
        return
endp



method TWindow.SetWidth
begin
        push    eax
        mov     eax, [.self]
        exec    eax, TWindow:Resize, [.value], [eax+TWindow._height]
        pop     eax
        return
endp



method TWindow.SetHeight
begin
        push    eax
        mov     eax, [.self]
        exec    eax, TWindow:Resize, [eax+TWindow._width], [.value]
        pop     eax
        return
endp




method TWindow.Move
begin
        pushad
        mov     esi, [.self]

        mov     ebx, [.newX]
        mov     ecx, [.newY]
        xchg    ebx, [esi+TWindow._x]
        xchg    ecx, [esi+TWindow._y]

        cmp     ebx, [esi+TWindow._x]
        jne     .update

        cmp     ecx, [esi+TWindow._y]
        je      .finish

.update:
        get     edx, esi, TWindow:Parent
        test    edx, edx
        jz      .finish

        exec    edx, TWindow:RectChangedXY2, ebx, ecx, [esi+TWindow._width], [esi+TWindow._height]        ; previous position.
        exec    edx, TWindow:RectChangedXY2, [esi+TWindow._x], [esi+TWindow._y], [esi+TWindow._width], [esi+TWindow._height]       ; new position.

.finish:
        exec    esi, TWindow:__UpdateHandle
        popad
        return
endp



method TWindow.Resize     ;, .newWidth, .newHeight
begin
        pushad

        mov     esi, [.self]
        mov     ecx, [.newWidth]
        mov     ebx, [.newHeight]
        xchg    ecx, [esi+TWindow._width]
        xchg    ebx, [esi+TWindow._height]

        cmp     ecx, [esi+TWindow._width]    ; oldW, newW
        jne     .refresh

        cmp     ebx, [esi+TWindow._height]   ; oldH, newH
        je      .exit

.refresh:
        exec    esi, TWindow:__UpdateImages
        exec    esi, TWindow:RectChanged2, 0    ; refresh the window itself.

;now notify the parent window, if there are parts not covered by the new window size:

        get     edx, esi, TWindow:Parent
        test    edx, edx
        jz      .exit

        sub     ecx, [esi+TWindow._width]       ; oldW, newW
        jbe     .check_height

        mov     eax, [esi+TWindow._width]
        add     eax, [esi+TWindow._x]
        exec    edx, TWindow:RectChangedXY2, eax, [esi+TWindow._y], ecx, ebx

.check_height:
        sub     ebx, [esi+TWindow._height]
        jbe     .exit

        mov     eax, [esi+TWindow._height]
        add     eax, [esi+TWindow._y]
        exec    edx, TWindow:RectChangedXY2, [esi+TWindow._x], eax, [esi+TWindow._width], ebx

.exit:
        popad
        return
endp


method TWindow.EventFocusIn
begin

        return
endp


method TWindow.EventFocusOut
begin

        return
endp


method TWindow.EventMouseEnter
begin
        push    esi edx

        mov     esi, [.self]
        get     edx, esi, TWindow:OSWindow

        stdcall GetStockCursor, [esi+TWindow._cursor]
        stdcall SetMouseCursor, [edx+TWindow.handle], eax

        pop     edx esi
        return
endp




method TWindow.EventMouseLeave
begin

        return
endp




method TWindow.EventMouseMove   ;, .x, .y, .kbdState
begin

        return
endp




method TWindow.EventButtonPress        ;, .button, .kbdState, .x, .y
begin

        return
endp



method TWindow.EventButtonRelease      ;, .button, .kbdState, .x, .y
begin

        return
endp




method TWindow.EventScroll     ;, .direction, .command, .value
begin
        mov     eax, [.self]
        cmp     [eax+TWindow._OnScroll], 0
        je      .finish

        pushad
        stdcall [eax+TWindow._OnScroll], [.self], [.direction], [.command], [.value]
        popad

.finish:
        return
endp




method TWindow.CloseRequest     ;, .reason
begin

        return
endp




method TWindow.EventKeyPress    ;, .utf8, .scancode, .kbdState
begin
        mov     eax, [.self]
        cmp     [eax+TWindow._OnKeyPressed], 0
        je      .finish

        pushad
        stdcall [eax+TWindow._OnKeyPressed], [.self], [.utf8], [.scancode], [.kbdState]
        popad

.finish:
        return
endp



method TWindow.EventKeyRelease  ;, .kbdState
begin

        return
endp


; Utility procedures.


method  TWindow.ClientToScreenXY; , .window, .x, .y
begin
        push    eax esi

        mov     ecx, [.x]
        mov     edx, [.y]
        mov     esi, [.self]

.loop:
        get     eax, esi, TWindow:Parent
        test    eax, eax
        jz      .os_win

        add     ecx, [esi+TWindow._x]
        add     edx, [esi+TWindow._y]

        mov     esi, eax
        jmp     .loop

.os_win:
        mov     eax, [esi+TWindow.handle]
        test    eax, eax
        jz      .finish


        stdcall _ClientToScreen, eax, ecx, edx

.finish:
        pop     esi eax
        return
endp



method TWindow.Autosize
begin
        mov     eax, [.self]
        mov     edx, [eax+TWindow._height]
        mov     eax, [eax+TWindow._width]
        return
endp



endmodule

