; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Text caret control library.
;
;  Target OS: Any
;
;  Dependencies: timers.asm; graphics libraries.
;
;  Notes: Text caret have always only one instance.
;_________________________________________________________________________________________
module "Text caret library"

CARET_MUTEX_WAIT = 2000 ; ms

; the caret is shared resource and can exists only in one instance
struct TCaret
  .pWindow    dd ?

  .pTopWindow dd ?
  .xOffs      dd ?      ; the offset between the caret owner window and its top level parent window
  .yOffs      dd ?      ;

  .state      dd ?       ; 1 if the caret is displayed 0 if not.
  .visible    dd ?
  .next_time  dd ?

  .Pos       TBounds
  .mxLock    TMutex
ends



uglobal
  if used CaretAttach
    Caret TCaret
  end if
endg


if used CaretAttach
initialize InitTextCaret
begin
        stdcall MutexCreate, 0, Caret.mxLock
        stdcall MutexRelease, Caret.mxLock
        return
endp

finalize FinalizeTextCaret
begin
        stdcall MutexDestroy, Caret.mxLock
        return
endp


end if




proc __CaretUpdateTopLevel, .pWindow
begin
        pushad

        stdcall WaitForMutex, Caret.mxLock, CARET_MUTEX_WAIT
        jc      .exit

        mov     esi, [.pWindow]
        cmp     esi, [Caret.pWindow]
        jne     .finish

        xor     ecx, ecx
        xor     edx, edx

        test    esi, esi
        jz      .found

.loop:
        get     ebx, esi, TWindow:Parent
        test    ebx, ebx
        jz      .found

        add     ecx, [esi+TWindow._x]
        add     edx, [esi+TWindow._y]
        mov     esi, ebx
        jmp     .loop

.found:
        mov     [Caret.xOffs], ecx
        mov     [Caret.yOffs], edx
        mov     [Caret.pTopWindow], esi

.finish:
        stdcall MutexRelease, Caret.mxLock
        clc

.exit:
        popad
        return
endp




; Attaches the caret to the window, if the window needs caret.
proc CaretAttach, .pWindow
begin
        push    ecx

        mov     eax,[.pWindow]
        cmp     eax, [Caret.pWindow]
        je      .exit

        stdcall WaitForMutex, Caret.mxLock, CARET_MUTEX_WAIT
        jc      .exit

        stdcall __ClearCaret

        mov     eax, [.pWindow]
        xchg    [Caret.pWindow], eax
        push    eax

        xor     eax, eax
        mov     [Caret.Pos.x], eax
        mov     [Caret.Pos.y], eax
        mov     [Caret.Pos.height], eax
        mov     [Caret.Pos.width], eax

        stdcall __CaretUpdateTopLevel, [Caret.pWindow]

        stdcall MutexRelease, Caret.mxLock
        pop     eax                             ; the old window.
        clc

.exit:
        pop     ecx
        return
endp




proc CaretChange, .x, .y, .width, .height, .visible
begin
        pushad

        stdcall WaitForMutex, Caret.mxLock, CARET_MUTEX_WAIT
        jc      .exit

        stdcall __ClearCaret

        mov     ecx, [.x]
        mov     edx, [.y]

        mov     [Caret.Pos.x], ecx
        mov     [Caret.Pos.y], edx

        xor     eax, eax
        mov     ecx, [.width]
        mov     edx, [.height]

        inc     eax

        test    ecx, ecx
        cmovz   ecx, eax

        test    edx, edx
        cmovz   edx, eax

        mov     [Caret.Pos.width], ecx
        mov     [Caret.Pos.height], edx

        mov     eax, [.visible]
        mov     [Caret.visible], eax

; restart the visible state from the beginning in order to make the moved
; caret move visible.

        mov     [Caret.state], 1
        stdcall GetTimestamp
        add     eax, [GUI.caretTimes+4]
        mov     [Caret.next_time], eax

        cmp     [Caret.pWindow], 0
        je      .exit

        exec    [Caret.pWindow], TWindow:RectChangedXY2, [Caret.Pos.x], [Caret.Pos.y], [Caret.Pos.width], [Caret.Pos.height]
        stdcall MutexRelease, Caret.mxLock
        clc

.exit:
        popad
        return
endp




proc __CaretTimerProc
begin
        pushad

        stdcall WaitForMutex, Caret.mxLock, CARET_MUTEX_WAIT
        jc      .exit

        stdcall GetTimestamp
        mov     ebx, eax

        cmp     ebx, [Caret.next_time]
        jb      .release

        cmp     [Caret.pWindow], 0
        je      .release

        xor     [Caret.state], 1
        stdcall __ClearCaret

        mov     eax, [Caret.state]
        and     eax, 1
        add     ebx, [GUI.caretTimes+4*eax]
        mov     [Caret.next_time], ebx

.release:
        stdcall MutexRelease, Caret.mxLock

.exit:
        popad
        return
endp


; Clears the caret at its present place.

proc __ClearCaret
begin
        stdcall WaitForMutex, Caret.mxLock, CARET_MUTEX_WAIT
        jc      .exit

        cmp     [Caret.pWindow], 0
        je      .finish

        exec    [Caret.pWindow], TWindow:RectChangedXY2, [Caret.Pos.x], [Caret.Pos.y], [Caret.Pos.width], [Caret.Pos.height]

.finish:
        stdcall MutexRelease, Caret.mxLock
.exit:
        return
endp


proc __PaintCaret, .topWindow, .screen
begin
        stdcall WaitForMutex, Caret.mxLock, CARET_MUTEX_WAIT
        jc      .exit

        push    ebx

        mov     ebx, [.topWindow]
        cmp     ebx, [Caret.pTopWindow]
        jne     .release

        mov     ebx, [.screen]
        test    ebx, ebx
        jz      .release

        cmp     [Caret.state], 0
        je      .release

        cmp     [Caret.visible], 0
        je      .release

        push    ecx edx
        mov     ecx, [Caret.Pos.x]
        mov     edx, [Caret.Pos.y]
        add     ecx, [Caret.xOffs]
        add     edx, [Caret.yOffs]
;        stdcall XorRect, ebx, ecx, edx, [Caret.Pos.width], [Caret.Pos.height], [GUI.clTextCaret]

        stdcall BlendSolidRect, ebx, ecx, edx, [Caret.Pos.width], [Caret.Pos.height], [GUI.clTextCaret]  ; ????????? does not look good this way...
        pop     edx ecx

.release:
        stdcall MutexRelease, Caret.mxLock
        pop     ebx

.exit:
        return
endp


endmodule


