; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Files manipulation library. Linux part.
;
;  Target OS: Linux
;
;  Dependencies: Uses system calls for file processing and memory.asm for memory
;                allocations.
;
;  Notes:
;_________________________________________________________________________________________

uses libc

DIR_SLASH equ '/'
DIR_SEP_PAIR equ '/\' ; for use with StrNormalizePath and StrNormalizePath2

fsFromBegin     = 0
fsFromEnd       = 2
fsFromCurrent   = 1


faReadOnly   = O_RDONLY
faWriteOnly  = O_WRONLY
faReadWrite  = O_RDWR

faOpenAlways   = O_CREAT             ; if the file does not exists it will be created.
faCreateAlways = O_CREAT or O_TRUNC  ; Creates new file always. If the file exists, truncates it.
faCreateOnly   = O_CREAT or O_EXCL   ; Creates new file only if not exists, else error.

faNonBlocking =  O_NONBLOCK

lockExclusive   = LOCK_EX
lockShared      = LOCK_SH
lockUnlock      = LOCK_UN
lockTryOnly     = LOCK_NB


if used STDIN | used STDOUT | used STDERR
  initialize InitStdHandles
  begin
    if used STDIN
          mov     dword [__std_handle_in], 0
    end if

    if used STDOUT
          mov     dword [__std_handle_out], 1
    end if

    if used STDERR
          mov     dword [__std_handle_err], 2
    end if
          return
  endp
end if



body FileOpenAccess
begin
        push    edx ecx ebx

        if defined ptrStrTable
          stdcall StrPtr, [.filename]
          mov     ebx, eax
        else
          mov   ebx, [.filename]
        end if

        mov     eax, sys_open
        mov     ecx, [.access]
        mov     edx, S_IRUSR+S_IWUSR+S_IRGRP+S_IWGRP+S_IROTH
        int     $80
        cmp     eax, -1
        jle     .error
        clc
        pop     ebx ecx edx
        return

.error:
        neg     eax

if defined options.DebugMode & options.DebugMode
        stdcall __FileErrorDisplay, eax, [.filename]
end if

        stc
        pop     ebx ecx edx
        return
endp




;--------------------------------------
; if CF = 0, the file was closed.
; if CF = 1, eax is error code.
;--------------------------------------
body FileClose
begin
        push    ebx
        mov     eax, sys_close
        mov     ebx, [.handle]
        int     $80
        cmp     eax, -1
        jle     .error
        clc
        pop     ebx
        return
.error:
        neg     eax
        stc
        pop     ebx
        return
endp



body FileLock
begin
        push    eax ebx ecx

        mov     ebx, [.handle]
        mov     eax, sys_flock
        mov     ecx, [.lock]
        int     $80

        shl     eax, 1  ; CF = 1 if eax<0
        pop     ecx ebx eax
        return
endp


; returns 32bit file size in eax
body FileSize
.stat STAT
begin
        push    edx ecx ebx

        mov     eax, sys_newfstat
        mov     ebx, [.handle]
        lea     ecx, [.stat]
        int     $80

        cmp     eax, -1
        jle     .error

        mov     eax, [.stat.st_size]
        clc
        pop     ebx ecx edx
        return

.error:
        neg     eax
        stc
        pop     ebx ecx edx
        return
endp



body FileRead
begin
        push    edx ecx ebx
        mov     eax, sys_read
        mov     ebx, [.handle]
        mov     ecx, [.buffer]
        mov     edx, [.count]
        int     $80
        cmp     eax, -1
        jle     .error
        clc
        pop     ebx ecx edx
        return

.error:
        neg     eax
        stc
        pop     ebx ecx edx
        return
endp


body FileWrite
begin
        push    edx ecx ebx
        mov     eax, sys_write
        mov     ebx, [.handle]
        mov     ecx, [.buffer]
        mov     edx, [.count]
        int     $80
        cmp     eax, -1
        jle     .error
        clc
        pop     ebx ecx edx
        return

.error:
        neg     eax
        stc
        pop     ebx ecx edx
        return
endp


;----------------------------------------------------
; if CF = 0, eax is the new possition in the file
; if CF = 1, eax is error code.
;----------------------------------------------------
body FileSeek
begin
        push    edx ecx ebx
        mov     eax,sys_lseek
        mov     ebx,[.handle]
        mov     ecx,[.dist]
        mov     edx,[.direction]
        int     $80
        cmp     eax,-1
        jle     .error
        clc
        pop     ebx ecx edx
        return

.error:
        neg     eax
        stc
        pop     ebx ecx edx
        return
endp



body FileTruncate
begin
        push    edx ecx ebx

        mov     eax, sys_lseek
        mov     ebx, [.handle]
        xor     ecx, ecx
        mov     edx, fsFromCurrent
        int     $80
        cmp     eax, -1
        jle     .error

        mov     ecx, eax
        mov     eax, sys_ftruncate
        mov     ebx, [.handle]
        int     $80
        cmp     eax,-1
        jle     .error

        clc
        pop     ebx ecx edx
        return

.error:
        neg     eax
        stc
        pop     ebx ecx edx
        return
endp


body FileTruncateTo     ; , .handle, .length
begin
        push    edx ecx ebx

        mov     eax, sys_ftruncate
        mov     ecx, [.length]
        mov     ebx, [.handle]
        int     $80
        cmp     eax,-1
        jle     .error

        clc
        pop     ebx ecx edx
        return

.error:
        neg     eax
        stc
        pop     ebx ecx edx
        return
endp



;--------------------------------------
; if CF = 0, the file was deleted.
; if CF = 1, eax is error code.
;--------------------------------------
body FileDelete
begin
        push    ebx

        if defined ptrStrTable
          stdcall StrPtr, [.filename]
          mov     ebx, eax
        else
          mov   ebx, [.filename]
        end if

        mov     eax,sys_unlink
        int     $80
        cmp     eax,-1
        jle     .error
        clc
        pop     ebx
        return

.error:
        neg     eax
        stc
        pop     ebx
        return
endp




body GetCurrentDir
.buffer rb 1024
begin
        push    ebx ecx

        mov     eax, sys_getcwd
        lea     ebx, [.buffer]
        mov     ecx, 1024
        int     $80
        cmp     eax, -1
        jle     .error

        stdcall StrDupMem, ebx
        clc
        pop     ecx ebx
        return

.error:
        neg     eax
        stc
        pop     ecx ebx
        return
endp




body SetCurrentDir
begin
        push    ebx eax

        stdcall StrPtr, [.hDirectory]
        mov     ebx, eax
        mov     eax, sys_chdir
        int     $80

        clc
        test    eax, eax
        jns     .finish

        neg     eax
        stc

.finish:
        pop     eax ebx
        return
endp



body GetErrorString
.buffer rb 1024
begin
        push    ecx edx

        lea     eax, [.buffer]
        cinvoke strerror_r, [.code], eax, 1024
        test    eax, eax
        jnz     @f
        lea     eax, [.buffer]
@@:
        stdcall StrDupMem, eax

        pop     edx ecx
        clc
        return
endp




body FreeErrorString
begin
        stdcall StrDel, [.ptrString]
        return
endp




ftDirectory  =  DT_DIR
ftUnknown    =  DT_UNKNOWN
ftFile       =  DT_REG
ftLinkMask   =  $80000000


; reads all items of the directory and fills them in the TArray of type TDirItem.
; arguments:
;   dir - string with the path to the directory.
;
; returns:
;   if not error: CF=0; eax = pointer to TArray of TDirItem;
;   if error:     CF=1; eax = error code;

body DirectoryRead
.size   dd ?
.stat STAT64
begin
        push    ebx ecx edx esi edi

        stdcall GetMem, 1024
        mov     esi, eax

        stdcall CreateArray, sizeof.TDirItem
        mov     edi, eax

; open the file
        if defined ptrStrTable
          stdcall StrPtr, [.dir]
          mov     ebx, eax
        else
          mov   ebx, [.dir]
        end if

        mov     eax, sys_open
        mov     ecx, O_RDONLY or O_DIRECTORY
        mov     edx, S_IWUSR+S_IWGRP+S_IWOTH
        int     $80
        test    eax, eax
        js      .error

        mov     ebx, eax

.read_loop:

        mov     eax, sys_getdents64
        mov     ecx, esi
        mov     edx, 1024
        int     $80

; process read information

        test    eax, eax
        jz      .end_of_directory
        js      .error

        mov     [.size], eax

        xor     ecx, ecx

.buffer_parse:
        lea     eax, [esi+ecx+linux_dirent64.d_name]
        stdcall StrDup, eax
        push    eax

        stdcall AddArrayItems, edi, 1
        mov     edi, edx

        pop     [eax+TDirItem.hFilename]

        movzx   edx, byte [esi+ecx+linux_dirent64.d_type]

        call    .get_file_details

        movzx   edx, word [esi+ecx+linux_dirent64.d_reclen]
        add     ecx, edx

        cmp     ecx, [.size]
        jb      .buffer_parse

        jmp     .read_loop

.end_of_directory:

        mov     eax, sys_close
        int     $80

        stdcall FreeMem, esi
        mov     eax, edi
        pop     edi esi edx ecx ebx
        clc
        return

.error:
        stdcall FreeMem, esi
        neg     eax
        pop     edi esi edx ecx ebx
        stc
        return

.get_file_details:
        push    ecx edx esi edi
        mov     edi, eax        ; TDirItem

        stdcall StrPtr, [eax+TDirItem.hFilename]
        mov     ecx, eax

        mov     eax, sys_fstatat64
        lea     edx, [.stat]
        xor     esi, esi
        int     $80

        test    eax, eax
        jnz     .end_file

        mov     edx, [esp+8]
        cmp     edx, DT_LNK
        jne     .type_ok

        mov     edx, [.stat.st_mode]
        shr     edx, 12
        or      edx, ftLinkMask

.type_ok:
        mov     [edi+TDirItem.Type], edx

        mov     eax, dword [.stat.st_size]
        mov     edx, dword [.stat.st_size+4]

        mov     dword [edi+TDirItem.Size], eax
        mov     dword [edi+TDirItem.Size+4], edx

        mov     eax, [.stat.st_mtime]
        xor     edx, edx
        mov     dword [edi+TDirItem.mTime], eax
        mov     dword [edi+TDirItem.mTime+4], edx

.end_file:
        pop     edi esi edx ecx
        retn
endp





body CreateDirectory
begin
        pushad

        if defined ptrStrTable
          stdcall StrPtr, [.DirName]
          mov     ebx, eax
        else
          mov   ebx, [.DirName]
        end if

        mov     eax, sys_mkdir
        mov     ecx, 755o
        int     $80

        cmp     eax, -1
        jle     .error
        clc
        popad
        return

.error:
        neg     eax
        mov     [esp+4*regEAX], eax
        stc
        popad
        return
endp








body GetFileInfo
.fstat STAT
begin
        push    edi ebx ecx eax

        lea     edi, [.fstat]
        mov     ecx, sizeof.STAT / 4
        xor     eax, eax
        rep stosd

        mov     eax, sys_newfstat
        mov     ebx, [.hFile]
        lea     ecx, [.fstat]
        int     $80

        test    eax, eax
        jnz     .error

        lea     ecx, [.fstat]
        mov     ebx, [.ptrFileInfo]

        mov     eax, [ecx+STAT.st_atime]
        cdq
        mov     dword [ebx+TFileInfo.timeAccessed], eax
        mov     dword [ebx+TFileInfo.timeAccessed+4], edx

        mov     eax, [ecx+STAT.st_mtime]
        cdq
        mov     dword [ebx+TFileInfo.timeModified], eax
        mov     dword [ebx+TFileInfo.timeModified+4], edx
; linux does not support creation time, so we set it as the modified time.
        mov     dword [ebx+TFileInfo.timeCreated], eax
        mov     dword [ebx+TFileInfo.timeCreated+4], edx

        pop     eax ecx ebx edi
        clc
        return

.error:
        neg     eax
        mov     [esp], eax

        pop     eax ecx ebx edi
        stc
        return

endp



; Returns:
; CF=1 if the file NOT EXISTS
; CF=0 if the file EXISTS
body FileExists
.fstat STAT
begin
        pushad

        lea     edi, [.fstat]
        mov     ecx, sizeof.STAT / 4
        xor     eax, eax
        rep stosd

        if defined ptrStrTable
          stdcall StrPtr, [.hFileName]
          mov     ebx, eax
        else
          mov   ebx, [.hFileName]
        end if

        mov     eax, sys_newstat
        lea     ecx, [.fstat]
        int     80h
        test    eax, eax
        jz      .exists

; error
        stc
        popad
        return

.exists:
        clc
        popad
        return
endp




body CreatePipe
.hRead  dd ?
.hWrite dd ?
begin
        push    ebx

        mov     eax, sys_pipe
        lea     ebx, [.hRead]
        int     $80
        neg     eax
        jnz     .err

        clc
        mov     eax, [.hRead]
        mov     edx, [.hWrite]
        pop     ebx
        return

.err:
        stc
        pop     ebx
        return
endp






flushInput = 0
flushOutput = 1
flushInputOutput = 2

TCFLSH  = $540B         ; Linux constant.


proc TerminalFlush, .hFile, .flush_what
begin
        pushad
        mov     eax, sys_ioctl
        mov     ebx, [.hFile]
        mov     ecx, TCFLSH
        mov     edx, [.flush_what]
        int     $80

        mov     [esp+4*regEAX], eax
        bt      eax, 31         ; if negative - error
        popad
        return
endp





body IsRelativePath     ;, .hPath
begin
        push    eax
        stdcall StrLen, [.hPath]
        test    eax, eax
        jz      .relative

        stdcall StrPtr, [.hPath]
        cmp     byte [eax], DIR_SLASH
        clc
        je      .finish

.relative:
        stc     ; relative!

.finish:
        pop     eax
        return
endp



body GetCurrentPrefix
begin
        stdcall StrNew
        return
endp